#include <utility.h>
#include <ansi_c.h>
#include <userint.h>
#include "toolbox.h"
#include "custsupp.h"


static char gIniFilePath[MAX_PATHNAME_LEN];

#define INI_FILE_NAME       "custctrl.ini"


static int  GetSectionList(char *prefix, ListType *sectionList);
static int  AddSectionListToListBox(int panel, int ctrl, char *prefix, int grayOutItems);
static int  CVICALLBACK ConfigStringCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  FindSettingInListCtrl(int panel, int list, char *settingName, int *indexWithValue);

/*****************************************************************************************/

char * CVIFUNC GetCustCtrlIniFilePath(void)
{
    static int pathInited = FALSE;

    if (!pathInited)
        {
        if (GetProjectDir (gIniFilePath) != 0)
            gIniFilePath[0] = 0;

        MakePathname (gIniFilePath, INI_FILE_NAME, gIniFilePath);
        pathInited = TRUE;
        }

    return gIniFilePath;
}

/*****************************************************************************************/

int CVIFUNC ReadConfigIni(IniText *iniText)
{
    int error = 0;
    int fileExists;

    nullChk( *iniText = Ini_New(1));

    errChk( fileExists = FileExists(GetCustCtrlIniFilePath(), 0));

    if (!fileExists)
        errChk(WriteConfigIni(*iniText));

    errChk( Ini_ReadFromFile (*iniText, GetCustCtrlIniFilePath()));

Error:
    if (error < 0)
        {
        Ini_Dispose (*iniText);
        *iniText = 0;
        }

    return error;
}

/*****************************************************************************************/

int CVIFUNC WriteConfigIni(IniText iniText)
{
    int error = 0;

    errChk( Ini_WriteToFile (iniText, GetCustCtrlIniFilePath()));

Error:
    return error;
}

/*****************************************************************************************/

static int GetSectionList(char *prefix, ListType *sectionList)
{
    int     error = 0;
    int     numSections;
    int     index;
    char    *sectionName;
    char    sectionNameBuf[SETTING_NAME_BUF_SIZE+256];
    int     prefixLength;
    int     addSectionToList;
    IniText iniText = 0;

    *sectionList = 0;

    nullChk( *sectionList = ListCreate(SETTING_NAME_BUF_SIZE));

    errChk( ReadConfigIni(&iniText));

    if (prefix)
        prefixLength = strlen(prefix);

    numSections = Ini_NumberOfSections (iniText);

    for (index = 1; index <= numSections; index++)
        {
        Ini_NthSectionName (iniText, index, &sectionName);
        addSectionToList = TRUE;

            /* strip off the prefix */
        if (prefix && (prefixLength > 0))
            {
            if (strncmp(sectionName, prefix, prefixLength) == 0)
                sectionName += prefixLength;
            else
                addSectionToList = FALSE;   /* there was a prefix, but it did not match */
            }

        if (addSectionToList)
            {
            StringCopyMax(sectionNameBuf, sectionName, SETTING_NAME_BUF_SIZE);
            nullChk( ListInsertItem(*sectionList, sectionNameBuf, END_OF_LIST));
            }
        }

Error:
    Ini_Dispose (iniText);
    if (error < 0)
        {
        ListDispose(*sectionList);
        *sectionList = 0;
        }

    return error;
}

/****************************************************************************************/

int CVIFUNC DeleteSettings(ListType settingsToDelete, char *prefix)
{
    int     error = 0;
    IniText iniText = 0;
    int     index, numSettings;
    char    section[MAX_SETTING_NAME_LENGTH + 256];     /* prefix should never be anywhere near 256 characters */

    numSettings = ListNumItems(settingsToDelete);

    if (numSettings > 0)
        {
        if (!prefix)
            prefix = "";

        errChk( ReadConfigIni(&iniText));

        for (index = 1; index <= numSettings; index++)
            {
            sprintf(section, "%s%s", prefix, (char *)ListGetPtrToItem(settingsToDelete, index));
            Ini_RemoveSection (iniText, section);
            }

        errChk( WriteConfigIni(iniText));
        }

Error:
    Ini_Dispose (iniText);
    return error;
}

/*****************************************************************************************/

static int AddSectionListToListBox(int panel, int ctrl, char *prefix, int grayOutItems)
{
    int         error = 0;
    ListType    sectionList = 0;
    int         index, numItems;
    char        *settingName;
    char        settingNameBuf[SETTING_NAME_BUF_SIZE+256];
    char        fgCode[20];

    errChk(GetSectionList(prefix, &sectionList));
    numItems = ListNumItems(sectionList);

    if (grayOutItems)
        sprintf(fgCode, "\033fg%6.6x", VAL_DK_GRAY);
    else
        fgCode[0] = 0;

    for (index = 1; index <= numItems; index++)
        {
        settingName = (char *)ListGetPtrToItem(sectionList, index);
        sprintf(settingNameBuf, "%s%s", fgCode, settingName);
        errChk( InsertListItem (panel, ctrl, -1, settingNameBuf, settingName));
        }

Error:
    ListDispose(sectionList);
    return error;
}

/*****************************************************************************************/

    /*  returns indexWithValue == -1 if the setting is not in the list */
static int FindSettingInListCtrl(int panel, int list, char *settingName, int *indexWithValue)
{
    int     error = 0;
    char    settingNameBuf[SETTING_NAME_BUF_SIZE];
    int     numItems;
    int     index;

    *indexWithValue = -1;
    index = 0;
    errChk( GetNumListItems (panel, list, &numItems));
    for (index = 0; index < numItems; index++)
        {
        errChk( GetValueFromIndex (panel, list, index, settingNameBuf));
        if (StrICmpWithoutSurrWhiteSpace(settingNameBuf, settingName) == 0)
            {
            *indexWithValue = index;
            break;
            }
        }

Error:
    return error;
}

/*****************************************************************************************/

#define DLG_WIDTH               300
#define CONFIG_STRING_WIDTH     170

    /*  This dialog serves as the Load, Save, and Delete Settings dialog.
        Returns CONFIG_DLG_OK if OK was pressed, CONFIG_DLG_CANCEL if cancel was pressed, or a negative error code
        if showSettingsName == FALSE, then the value of settingsName will be the selected value in the list box.
        Otherwise, the configString will be initialized with the value of settingsName and
        updated by list box commits (or direct editing). Then when the dialog exits, the value in
        the configString will be copied back to settingsName.

        If the promptOnOverwrite parameter is TRUE, a confirmation dialog to appear is the user
        selects an existing setting (this parameter is ignored if showSettingsName == FALSE).

    */
int CVIFUNC ConfigDialog(char *dialogTitle, char *ctrlPrefix, char *settingsName, int showSettingsName, int grayList,  int listPickEqualsOk, int checkMode, ListType *checkedItems, int promptOnOverwrite)
{
    int     error = 0;
    int     panel = 0;
    int     list, ok, cancel, configString;
    int     stringBoundingWidth;
    int     labelWidth;
    int     ctrlLeft;
    int     contentWidth = DLG_WIDTH - 2 * DLG_MARGIN;
    int     top = DLG_MARGIN;
    int     height, width;
    int     done = FALSE;
    int     result;
    int     evtPanel, evtCtrl;
    int     ctrlIndex;
    char    settingNameBuf[SETTING_NAME_BUF_SIZE];
    int     waitCursorState = GetWaitCursorState();
    int     indexWithValue;

    SetWaitCursor(TRUE);
    Assert((checkMode != 0) == (checkedItems != 0));    /* can't have one without the other */

    if (checkedItems)
        *checkedItems = 0;

    errChk( panel =         NewPanel(0, dialogTitle, 10, 10, 10, 10));
    SetPanelAttribute(panel, ATTR_CONFORM_TO_SYSTEM, TRUE);
    errChk( configString =  NewCtrl (panel, CTRL_STRING, "Configuration Name:", 0, 0));
    errChk( ok =            NewCtrl (panel, CTRL_SQUARE_COMMAND_BUTTON, "__OK", 0, 0));
    errChk( cancel =        NewCtrl (panel, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 0, 0));
    errChk( list =          NewCtrl (panel, CTRL_LIST, "", 0, DLG_MARGIN));

    errChk( InstallCtrlCallback(panel, configString, ConfigStringCallback, 0));

    errChk( SetCtrlAttribute(panel, list, ATTR_DATA_TYPE, VAL_STRING));
    errChk( SetCtrlAttribute(panel, configString, ATTR_MAX_ENTRY_LENGTH, MAX_SETTING_NAME_LENGTH));

    if (settingsName)
        errChk( SetCtrlVal(panel, configString, settingsName));

    GetCtrlVal(panel, configString, settingNameBuf);

    if (checkMode)
        errChk( SetCtrlAttribute(panel, list, ATTR_CHECK_MODE, TRUE));

    SetCtrlAttribute(panel, configString, ATTR_WIDTH, CONFIG_STRING_WIDTH);
    errChk( SetCommonDialogShortcutKeys(panel, ok, cancel, cancel));
    errChk( SetCtrlsToAttributeExtreme (panel, ATTR_WIDTH, VAL_TB_MAX, 0, ok, cancel, 0));

    errChk( PutLabelOnLeft(panel, configString, 0));

    GetCtrlAttribute(panel, configString, ATTR_LABEL_WIDTH, &labelWidth);

    ctrlLeft = DLG_MARGIN + labelWidth + DLG_LABEL_TO_CTRL_MIN_H_GAP;
    SetCtrlAttribute(panel, configString, ATTR_LEFT, ctrlLeft);
    SetCtrlAttribute(panel, configString, ATTR_LABEL_LEFT, DLG_MARGIN);
    GetCtrlBoundingRect(panel, configString, 0, 0, 0, &stringBoundingWidth);
    contentWidth = Max(contentWidth, stringBoundingWidth);
    errChk( SetCtrlAttribute(panel, list, ATTR_WIDTH, contentWidth));

    errChk( SetCtrlAttribute(panel, configString, ATTR_VISIBLE, showSettingsName));

    if (showSettingsName)
        {
        SetCtrlAttribute(panel, configString, ATTR_TOP, top);
        GetCtrlAttribute(panel, configString, ATTR_HEIGHT, &height);
        top += height +  DLG_CTRL_V_GAP;
        }

    SetCtrlAttribute(panel, list, ATTR_TOP, top);
    GetCtrlAttribute(panel, list, ATTR_HEIGHT, &height);
    top += height +  DLG_CTRL_V_GAP;

    errChk( SetAttributeForCtrls (panel, ATTR_TOP, top, 0, ok, cancel, 0));
    GetCtrlAttribute(panel, ok, ATTR_HEIGHT, &height);
    top += height;

    GetCtrlAttribute(panel, ok, ATTR_WIDTH, &width);
    SetCtrlAttribute(panel, cancel, ATTR_LEFT, DLG_MARGIN + contentWidth - width);
    SetCtrlAttribute(panel, ok, ATTR_LEFT, DLG_MARGIN + contentWidth - 2 * width - DLG_BTN_H_GAP);

    SetPanelAttribute(panel, ATTR_WIDTH, contentWidth + 2 * DLG_MARGIN);
    SetPanelAttribute(panel, ATTR_HEIGHT, top + DLG_MARGIN);

    SetPanelPos(panel, VAL_AUTO_CENTER, VAL_AUTO_CENTER);

    errChk( AddSectionListToListBox(panel, list, ctrlPrefix, grayList));

    if (showSettingsName)       /* select the item in the list box that matches the inital config name */
        {
        errChk( FindSettingInListCtrl(panel, list, settingsName, &indexWithValue))

        if (indexWithValue >= 0)
            errChk( SetCtrlIndex(panel, list, indexWithValue));
        }

    InstallPopup(panel);
    SetWaitCursor(FALSE);

    while (!done)
        {
        GetUserEvent(1, &evtPanel, &evtCtrl);

        if (evtPanel == panel)
            {
            if (evtCtrl == ok)
                {
FakeOk:
                if (showSettingsName && settingNameBuf[0] == 0)     /* must enter a non-blank name */
                    {
                    errChk( SetActiveCtrl(panel, configString));
                    Beep();
                    }
                else
                    {
                    done = TRUE;
                    result = CONFIG_DLG_OK;

                    if (showSettingsName && promptOnOverwrite)
                        {
                        char    msg[128 + SETTING_NAME_BUF_SIZE];

                        errChk( FindSettingInListCtrl(panel, list, settingNameBuf, &indexWithValue))

                        if (indexWithValue >= 0)
                            {
                            sprintf(msg, "A configuration named \"%s\" already exists.\nWould you like to replace it?", settingNameBuf);
                            done = ConfirmPopup("Confirmation", msg);
                            }
                        }
                    }
                }
            else
            if (evtCtrl == cancel)
                {
                done = TRUE;
                result = CONFIG_DLG_CANCEL;
                }
            else
            if (evtCtrl == list)
                {
                GetCtrlIndex (panel, list, &ctrlIndex);
                if (ctrlIndex >= 0)
                    {
                    GetCtrlVal(panel, list, settingNameBuf);
                    errChk( SetCtrlVal(panel, configString,settingNameBuf));
                    }
                if (listPickEqualsOk)
                    goto FakeOk;
                }
            else
            if (evtCtrl == configString)
                {
                GetCtrlVal(panel, configString, settingNameBuf);
                }
            }
        }


        /*  If the list was in checkMode, build a list of checked items to return */
    if (checkMode && checkedItems)
        {
        int index, numItems;

        nullChk( *checkedItems = ListCreate(SETTING_NAME_BUF_SIZE));

        errChk( GetNumListItems (panel, list, &numItems));

        for (index = 1; index <= numItems; index++)
            {
            int checked;

            errChk( IsListItemChecked (panel, list, index - 1, &checked));

            if (checked)
                {
                errChk( GetLabelFromIndex (panel, list, index - 1, settingNameBuf));
                nullChk( ListInsertItem(*checkedItems, settingNameBuf, END_OF_LIST));
                }
            }
        }


        /*  Update the settingsName parameter to reflect the user's selection */
    if (settingsName)
        {
        if (showSettingsName)
            GetCtrlVal(panel, configString, settingsName);
        else
            {
            GetCtrlIndex (panel, list, &ctrlIndex);
            if (ctrlIndex >= 0)
                GetCtrlVal(panel, list, settingsName);
            else
                settingsName[0] = 0;
            }
        RemoveSurroundingWhiteSpace(settingsName);
        }

Error:
    SetWaitCursor(waitCursorState);

    if (panel > 0)
        DiscardPanel(panel);

    if (error < 0)
        {
        if (checkedItems)
            {
            ListDispose(*checkedItems);
            *checkedItems = 0;
            }

        return error;
        }
    else
        return result;
}

/*****************************************************************************************/

    /* disallow '[' and ']' in config name string control */
static int  CVICALLBACK ConfigStringCallback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int swallow = 0;

    UNUSED(panel);
    UNUSED(control);
    UNUSED(callbackData);
    UNUSED(eventData2);

    switch (event)
        {
        case EVENT_KEYPRESS:
            eventData1 &= VAL_ASCII_KEY_MASK;
            if (eventData1 == '[' || eventData1 == ']')
                {
                Beep();
                swallow = 1;
                }
            break;
        }

    return swallow;
}

/*****************************************************************************************/

    /*  pass 0 to maxSize and it will be ignored */
void CVIFUNC ExpandCtrlToMeetLabel(int panel, int ctrl, int gap, int maxSize)
{
    int labelLeft, left, width, labelWidth, labelRight, newLeft, oldRight, newWidth;

    GetCtrlAttribute(panel, ctrl, ATTR_LABEL_LEFT, &labelLeft);
    GetCtrlAttribute(panel, ctrl, ATTR_LABEL_WIDTH, &labelWidth);
    labelRight = labelLeft + labelWidth;
    newLeft = labelRight + gap;
    GetCtrlAttribute(panel, ctrl, ATTR_LEFT, &left);
    GetCtrlAttribute(panel, ctrl, ATTR_WIDTH, &width);
    oldRight = left + width;
    newWidth = oldRight - newLeft;
    if (maxSize > 0 && newWidth > maxSize)
        {
        newWidth = maxSize;
        newLeft = left - (newWidth - width);
        }

    SetCtrlAttribute(panel, ctrl, ATTR_WIDTH, newWidth);
    SetCtrlAttribute(panel, ctrl, ATTR_LEFT, newLeft);
    SetCtrlAttribute(panel, ctrl, ATTR_LABEL_LEFT, labelLeft);
}

/*****************************************************************************************/

int CVIFUNC MakeCtrlGrayRecessedIndicator(int panel, int ctrl)
{
    int error = 0;
    int conform;

    errChk( SetCtrlAttribute(panel, ctrl, ATTR_CTRL_MODE, VAL_INDICATOR));
    
    errChk( GetPanelAttribute(panel, ATTR_CONFORM_TO_SYSTEM, &conform));
    errChk( SetCtrlAttribute (panel, ctrl, ATTR_FRAME_COLOR, VAL_TRANSPARENT));
    if (!conform)
        errChk( SetCtrlAttribute (panel, ctrl, ATTR_TEXT_BGCOLOR, VAL_PANEL_GRAY));

Error:
    return error;
}

/*****************************************************************************************/

    /*  The function is used to arrange a rectangular outlined decoration and a textMsg so
        that the text message is placed over the decoration to create a labeled grouping effect:

        ---Text-Message -------------
        |                           |
        |                           |
        -----------------------------

        The textMsg control is assumed to be in its correct vertical position.
        left is the left position of the decoration
        *bottom is a pointer to the position to place the bottom of the decoration at.
    */
int CVIFUNC FormatGroupDecoration(int dialog, int textMsg, int deco, int left, int bottom)
{
    int error = 0;
    int textHeight, textTop;
    int decoTop;

    errChk( GetCtrlAttribute(dialog, textMsg, ATTR_HEIGHT, &textHeight));
    errChk( GetCtrlAttribute(dialog, textMsg, ATTR_TOP, &textTop));
    errChk( SetCtrlAttribute(dialog, deco, ATTR_LEFT, left));
    errChk( SetCtrlAttribute(dialog, textMsg, ATTR_LEFT, left + 10));

    decoTop =  textTop + textHeight / 2 - 1;
    errChk( SetCtrlAttribute(dialog, deco, ATTR_TOP, decoTop));

    errChk( MoveBehind(dialog, textMsg, deco));
    errChk( SetCtrlAttribute(dialog, deco, ATTR_HEIGHT, bottom - decoTop));

Error:
    return error;
}

/*****************************************************************************************/
