/*============================================================================*/
/*                        L a b W i n d o w s / C V I                         */
/*----------------------------------------------------------------------------*/
/*    Copyright (c) National Instruments 1987-1996.  All Rights Reserved.     */
/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Title:       daqctrl.c                                                     */
/* Purpose:     contains code shared by the DAQ Controls                      */
/*                                                                            */
/*============================================================================*/

#include <utility.h>
#include "daqctrl.h"
#include "easytab.h"

#define ERROR_BOX_MARGIN    6
#define MAX_DEVICE_NUMBER   16

static void SetDisplayedError(DaqCtrl daqCtrl, int errorCode);
static void DisposeSheetLists(DaqCtrl daqCtrl);
static void DisposeNonSheetLists(DaqCtrl daqCtrl);
static int  CVICALLBACK EditConfigDlgCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static void SetTabOrderForList(int panel, ListType ctrlList);
static void GetErrorStringAndTitle(DaqCtrl daqCtrl, int errorCode, char **errorMsg, char errorTitle[]);
static int  HandleRevert(DaqCtrl daqCtrl);
static int  BuildDialogIfNeeded(DaqCtrl daqCtrl);
static int  SetDialogTitle(DaqCtrl daqCtrl, char *newTitle);
static void GetDialogTitle(DaqCtrl daqCtrl, char title[]);
static int  SetDialogDisplayedState(DaqCtrl daqCtrl, int display);
static int  GetHelpFilePath(char helpFilePath[]);
static int  CVICALLBACK HelpCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static int  CVICALLBACK ErrorTextBoxCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2);
static void DialogErrorDisplay(DaqCtrl daqCtrl);
static void CreateErrorDecoration(DaqCtrl daqCtrl);
static void CVICALLBACK ErrorDisplayExpirationCallback(void *callbackData);

/**************************************************************************/
#pragma warning(disable: 4761)
int CVIFUNC InitDaqCtrl(DaqCtrl daqCtrl)
{
    int error = 0;

    nullChk( RegisterExistence(daqCtrl));
    daqCtrl->minDialogWidth = 300;
    daqCtrl->channelStringIsRevertable = TRUE;

Error:
    return error;
}

/**************************************************************************/

void CVIFUNC DaqCtrl_DisposeDaqCtrl(DaqCtrl daqCtrl)
{
    DeRegisterExistence(daqCtrl);
    if (daqCtrl->dialog > 0)
        DiscardPanelNoBOLE(daqCtrl->dialog, TRUE);

    if (daqCtrl->errorBoxDecoration > 0)
        DiscardCtrlNoBOLE(daqCtrl->panel, daqCtrl->errorBoxDecoration, TRUE);

    DisposeSheetLists(daqCtrl);
    DisposeNonSheetLists(daqCtrl);
    free(daqCtrl->channelString);
    free(daqCtrl);
}

/*****************************************************************************************/

static void DisposeSheetLists(DaqCtrl daqCtrl)
{
    ListDispose(daqCtrl->daqTabCtrls);
    ListDispose(daqCtrl->configTabCtrls);
    ListDispose(daqCtrl->generalTabCtrls);
    ListDispose(daqCtrl->generalTabCtrlsToDistribute);
}

/*****************************************************************************************/

static void DisposeNonSheetLists(DaqCtrl daqCtrl)
{
    ListDispose(daqCtrl->ctrlsWithLabelsOnLeft);
    ListDispose(daqCtrl->ctrlsOnLeft);
    ListDispose(daqCtrl->ctrlListsToDistribute);
}

/*****************************************************************************************/

static int CreateNonSheetLists(DaqCtrl daqCtrl)
{
    int error = 0;

    nullChk(daqCtrl->daqTabCtrls = ListCreate(sizeof(int)));
    nullChk(daqCtrl->configTabCtrls = ListCreate(sizeof(int)));
    nullChk(daqCtrl->generalTabCtrls = ListCreate(sizeof(int)));
    nullChk(daqCtrl->generalTabCtrlsToDistribute = ListCreate(sizeof(int)));

Error:
    return 0;
}

/*****************************************************************************************/

static int CreateSheetLists(DaqCtrl daqCtrl)
{
    int error = 0;

    nullChk(daqCtrl->ctrlsWithLabelsOnLeft = ListCreate(sizeof(int)));
    nullChk(daqCtrl->ctrlsOnLeft = ListCreate(sizeof(int)));
    nullChk(daqCtrl->ctrlListsToDistribute = ListCreate(sizeof(int)));

Error:
    return 0;
}

/*****************************************************************************************/

int CVIFUNC GetDAQCtrlInfo(int panel, int daqCtrlId, char *type, DaqCtrl *daqCtrl)
{
    int error = 0;

    errChk( GetChainedCallbackData(panel, daqCtrlId, type, (void **)daqCtrl));

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC SetCtrlModeToIODirection(int panel, int control, int ioDirection)
{
    int error = 0;
    int mode;

    if(ioDirection == ANALOG_OUTPUT)
        mode = VAL_HOT;
    else
        mode = VAL_INDICATOR;

    errChk( SetCtrlAttribute (panel, control, ATTR_CTRL_MODE, mode));

Error:
    return error;
}

/*****************************************************************************************/

static int BuildDialogIfNeeded(DaqCtrl daqCtrl)
{
    int error = 0;

    if (daqCtrl->dialog <= 0)
        errChk( daqCtrl->buildDialog(daqCtrl));
    DialogErrorDisplay(daqCtrl);        /* in case an error would have already been displayed if the dialog had been created earlier */

Error:
    return error;
}

/**************************************************************************/

int CVIFUNC DaqCtrl_RememberAttributes(DaqCtrl daqCtrl)
{
    int error = 0;
    int panel = daqCtrl->panel;
    int ctrl  = daqCtrl->baseCtrl;

    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, &daqCtrl->savedDevice));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, daqCtrl->savedChannelString));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, daqCtrl->savedTitleString));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, &daqCtrl->savedHighLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, &daqCtrl->savedLowLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, &daqCtrl->savedIODirection));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, &daqCtrl->savedEnabled));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, &daqCtrl->savedAllowRightClickEdit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, &daqCtrl->savedVisible));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, &daqCtrl->savedTieLimits));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, &daqCtrl->savedModal));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, &daqCtrl->savedErrorNotification));

    strcpy(daqCtrl->savedSettingsName, daqCtrl->settingsName);

Error:
    return error;
}

/**************************************************************************/

    /* copy the saved values of the attributes back to the control */
int CVIFUNC DaqCtrl_RevertAttributes(DaqCtrl daqCtrl)
{
    int error = 0;
    int panel = daqCtrl->panel;
    int ctrl = daqCtrl->baseCtrl;
    int savedUpdateState = daqCtrl->updatingCtrlFromDialog;

    daqCtrl->updatingCtrlFromDialog = TRUE;
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, daqCtrl->savedDevice));

    if (daqCtrl->channelStringIsRevertable)
        errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, daqCtrl->savedChannelString));

    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, daqCtrl->savedTitleString));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, daqCtrl->savedHighLimit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, daqCtrl->savedLowLimit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, daqCtrl->savedIODirection));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, daqCtrl->savedEnabled));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, daqCtrl->savedAllowRightClickEdit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, daqCtrl->savedVisible));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, daqCtrl->savedTieLimits));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, daqCtrl->savedModal));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, daqCtrl->savedErrorNotification));
    strcpy(daqCtrl->settingsName, daqCtrl->savedSettingsName);

Error:
    daqCtrl->updatingCtrlFromDialog = savedUpdateState;
    return error;
}

/**************************************************************************/

    /*  copy the control's attributes to the dialog fields */
int CVIFUNC DaqCtrl_CopyAttributesToDialog(DaqCtrl daqCtrl)
{
    int     error = 0;
    int     panel, ctrl, dialog;
    int     allowRightClickEdit;
    int     visible;
    int     tieLimits;
    int     device;
    char    channelString[MAX_CHANNEL_STRING_LEN + 1];
    int     enabled;
    double  highLimit;
    int     ioDirection;
    double  lowLimit;
    char    title[MAX_EDIT_DIALOG_TITLE_LEN + 1];
    int     modal;
    int     errorNotification;

    dialog = daqCtrl->dialog;
    panel = daqCtrl->panel;
    ctrl = daqCtrl->baseCtrl;

    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, &device));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, &channelString));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, &highLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, &lowLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, &allowRightClickEdit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, &visible));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, &tieLimits));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, &ioDirection));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, &enabled));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, title));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, &modal));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, &errorNotification));

    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->deviceNumeric, device));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->channelStringCtrl, channelString));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->highLimitNumeric, highLimit));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->lowLimitNumeric, lowLimit));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->modalCheckBox, modal));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->allowRightClickEditCheckBox, allowRightClickEdit));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->visibleCheckBox, visible));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->tieLimitsCheckBox, tieLimits));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->ioDirectionRing, ioDirection));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->enabledCheckBox, enabled));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->titleString, title));
    errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->errorNotificationRing, errorNotification));


Error:
    return error;
}

/*****************************************************************************************/

    /* copy current dialog control settings to the control's attributes */
int CVIFUNC DaqCtrl_ApplyDialogSettings(DaqCtrl daqCtrl)
{
    int     error = 0;
    int     dialog;
    int     panel;
    int     ctrl;
    int     allowRightClickEdit;
    int     visible;
    int     tieLimits;
    int     device;
    char    channelString[MAX_CHANNEL_STRING_LEN+1];
    int     enabled;
    double  highLimit;
    int     ioDirection;
    double  lowLimit;
    char    title[MAX_EDIT_DIALOG_TITLE_LEN + 1];
    int     modal;
    int     errorNotification;
    int     savedUpdateState = daqCtrl->updatingCtrlFromDialog;

    dialog = daqCtrl->dialog;
    panel = daqCtrl->panel;
    ctrl = daqCtrl->baseCtrl;

    daqCtrl->updatingCtrlFromDialog = TRUE;
    errChk( GetCtrlVal(dialog, daqCtrl->modalCheckBox, &modal));
    errChk( GetCtrlVal(dialog, daqCtrl->errorNotificationRing, &errorNotification));
    errChk( GetCtrlVal(dialog, daqCtrl->allowRightClickEditCheckBox, &allowRightClickEdit));
    errChk( GetCtrlVal(dialog, daqCtrl->visibleCheckBox, &visible));
    errChk( GetCtrlVal(dialog, daqCtrl->tieLimitsCheckBox, &tieLimits));
    errChk( GetCtrlVal(dialog, daqCtrl->deviceNumeric, &device));
    errChk( GetCtrlVal(dialog, daqCtrl->channelStringCtrl, channelString));
    errChk( GetCtrlVal(dialog, daqCtrl->enabledCheckBox, &enabled));
    errChk( GetCtrlVal(dialog, daqCtrl->highLimitNumeric, &highLimit));
    errChk( GetCtrlVal(dialog, daqCtrl->ioDirectionRing, &ioDirection));
    errChk( GetCtrlVal(dialog, daqCtrl->lowLimitNumeric, &lowLimit));
    errChk( GetCtrlVal(dialog, daqCtrl->titleString, title));

    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, errorNotification));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, modal));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, allowRightClickEdit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, visible));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, tieLimits));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, device));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, channelString));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, enabled));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, highLimit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, ioDirection));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, lowLimit));
    errChk( daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, title));

Error:
    daqCtrl->updatingCtrlFromDialog = savedUpdateState;
    return error;
}

/*****************************************************************************************/

int CVIFUNC DaqCtrl_UpdateConfigDlgCtrl(DaqCtrl daqCtrl)
{
    int error = 0;

    if (daqCtrl->dialog)
        errChk( SetCtrlVal(daqCtrl->dialog, daqCtrl->configNameString, daqCtrl->settingsName));

Error:
    return error;
}

/*****************************************************************************************/

#define ERROR_ELABORATION_BUF_SIZE  512

static char errorElaborationMsg[ERROR_ELABORATION_BUF_SIZE];

    /*  specifies extra text to display with the next error message */
void CVIFUNC DaqCtrl_SetErrorElaboration(DaqCtrl daqCtrl, char *elaboration)
{
    UNUSED(daqCtrl);
    StringCopyMax (errorElaborationMsg, elaboration, ERROR_ELABORATION_BUF_SIZE);
}

/*****************************************************************************************/

    /*  doesn't return an error, if it fails, the red box doesn't show up, but nothing will break */
static void CreateErrorDecoration(DaqCtrl daqCtrl)
{
    daqCtrl->errorBoxDecoration = NewCtrl (daqCtrl->panel, CTRL_FLAT_FRAME, "", 0, 0);
    if (daqCtrl->errorBoxDecoration > 0)
        {
        SetCtrlAttribute (daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_VISIBLE, FALSE);
        SetCtrlAttribute (daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_FRAME_COLOR, VAL_RED);
        }
}

/*****************************************************************************************/

static void CVICALLBACK ErrorDisplayExpirationCallback(void *callbackData)
{
    if (ObjectExists(callbackData) && ((DaqCtrl)callbackData)->autoErrorClearingNeeded)
        DaqCtrl_DisplayError((DaqCtrl)callbackData, 0, 0.0, 0.0);
}

/*****************************************************************************************/

void CVIFUNC DaqCtrl_DisplayError(DaqCtrl daqCtrl, int errorCode, double timeTillReplaceable, double timeTillClearable)
{
    double  time;

    if (daqCtrl->currentErrorCode == 0 && errorCode == 0)
        return;     /* return as fast as possible in this case */

    time = Timer();
    if (errorCode == 0)
        {
        if (daqCtrl->errorClearableTime <= time)
            DaqCtrl_ClearDisplayedError(daqCtrl);
        }
    else
        {
        if (daqCtrl->errorReplaceableTime <= time || (daqCtrl->currentErrorCode > 0 && errorCode < 0))
            {
            daqCtrl->errorReplaceableTime = time + timeTillReplaceable;
            daqCtrl->errorClearableTime = time + timeTillClearable;

            if (daqCtrl->autoErrorClearingNeeded)
                PostDelayedCall (ErrorDisplayExpirationCallback, (void *)daqCtrl, timeTillClearable + 1.0);

            if (daqCtrl->errorNotifyFunc)
                {
                char    *errorMsg;
                GetErrorStringAndTitle(daqCtrl, errorCode, &errorMsg, 0);

                daqCtrl->errorNotifyFunc(daqCtrl, errorCode, errorMsg);
                }

            SetDisplayedError(daqCtrl, errorCode);
            }

        errorElaborationMsg[0] = 0;
        }
}

/*****************************************************************************************/

void CVIFUNC DaqCtrl_ClearDisplayedError(DaqCtrl daqCtrl)
{
    if (daqCtrl->errorBoxDecoration > 0)
        SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_VISIBLE, FALSE);

        /* don't clear the error display if the error text is the active control */
    if (daqCtrl->dialog <= 0 || (GetActiveCtrl(daqCtrl->dialog) != daqCtrl->errorTextBox))
        {
        daqCtrl->currentErrorCode = 0;
        DialogErrorDisplay(daqCtrl);
        }
}

/*****************************************************************************************/

    /*  pass 0 for errorTitle if you do not need it */
static void GetErrorStringAndTitle(DaqCtrl daqCtrl, int errorCode, char **errorMsg, char errorTitle[])
{
    char    *errorTitlePtr;
    int     absError = ABS_VAL(errorCode);

    if (errorTitle)
        strcpy (errorTitle, "???");
    *errorMsg = '\0';
    
    if (daqCtrl->getErrorString && daqCtrl->getErrorString(daqCtrl, errorCode, &errorTitlePtr, errorMsg))
        {
        if (errorTitle)
            strcpy(errorTitle, errorTitlePtr);
        }
    else
    if (absError >= 10000 && absError < 20000)
        {
        if (errorTitle)
            sprintf(errorTitle, "Easy I/O %s = %d",  errorCode < 0 ? "Error" : "Warning", errorCode);
        *errorMsg = GetDAQErrorString(errorCode);
        }
    else
        {
        if (errorTitle)
            sprintf(errorTitle, "%s = %d", errorCode < 0 ? "Error" : "Warning", errorCode);
        *errorMsg = GetGeneralErrorString(errorCode);
        }
}

/*****************************************************************************************/

static void DialogErrorDisplay(DaqCtrl daqCtrl)
{
    char    errorTitle[400];
    char    *errorMsg;

    if (daqCtrl->dialog > 0)
        {
        if (daqCtrl->currentErrorCode != 0)
            {
            if (!daqCtrl->errorDisplayVisible || (daqCtrl->displayedErrorCode != daqCtrl->currentErrorCode))
                {
                daqCtrl->errorDisplayVisible = TRUE;
                daqCtrl->displayedErrorCode = daqCtrl->currentErrorCode;

                GetErrorStringAndTitle(daqCtrl, daqCtrl->currentErrorCode, &errorMsg, errorTitle);

                ResetTextBox(daqCtrl->dialog, daqCtrl->errorTextBox, errorMsg);
                SetCtrlVal(daqCtrl->dialog, daqCtrl->errorTextBox, "\n");
                SetCtrlVal(daqCtrl->dialog, daqCtrl->errorTextBox, errorElaborationMsg);
                SetCtrlAttribute (daqCtrl->dialog, daqCtrl->errorTextBox, ATTR_FIRST_VISIBLE_LINE, 0);
                SetCtrlAttribute(daqCtrl->dialog, daqCtrl->errorTextMsg, ATTR_CTRL_VAL, errorTitle);

                SetPanelAttribute(daqCtrl->dialog, ATTR_HEIGHT, daqCtrl->heightWithErrorExplanation);
                SetCtrlAttribute (daqCtrl->dialog, daqCtrl->errorTextBox, ATTR_VISIBLE, TRUE);
                }
            }
        else
            {
            if (daqCtrl->errorDisplayVisible)
                {
                daqCtrl->errorDisplayVisible = FALSE;
                SetPanelAttribute(daqCtrl->dialog, ATTR_HEIGHT, daqCtrl->heightWithoutErrorExplanation);
                SetCtrlAttribute (daqCtrl->dialog, daqCtrl->errorTextBox, ATTR_VISIBLE, FALSE);
                SetCtrlVal(daqCtrl->dialog, daqCtrl->errorTextMsg, "");
                }
            }
        }
}

/*****************************************************************************************/

static void SetDisplayedError(DaqCtrl daqCtrl, int errorCode)
{
    if (daqCtrl->currentErrorCode != errorCode)
        {
        int     top, left, width, height;

        switch (daqCtrl->errorNotification)
            {
            case ERR_NOTIFY_BEEP_AND_BOX:
            case ERR_NOTIFY_BEEP_ONLY:
                Beep();
                break;
            default:
                break;
            }

        if (daqCtrl->errorBoxDecoration <= 0)   /* didn't want to create it until we knew we needed it */
            CreateErrorDecoration(daqCtrl);

        if (daqCtrl->errorBoxDecoration > 0)
            {
            switch (daqCtrl->errorNotification)
                {
                case ERR_NOTIFY_BEEP_AND_BOX:
                case ERR_NOTIFY_BOX_ONLY:
                    GetCtrlBoundingRect (daqCtrl->panel, daqCtrl->baseCtrl, &top, &left, &height, &width);
                    top -= ERROR_BOX_MARGIN;
                    left -= ERROR_BOX_MARGIN;
                    width += 2 * ERROR_BOX_MARGIN;
                    height += 2 * ERROR_BOX_MARGIN;
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_TOP, top);
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_LEFT, left);
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_HEIGHT, height);
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_WIDTH, width);
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_ZPLANE_POSITION, 0);
                    SetCtrlAttribute(daqCtrl->panel, daqCtrl->errorBoxDecoration, ATTR_VISIBLE, TRUE);
                    break;
                default:
                    break;
                }
            }

        daqCtrl->currentErrorCode = errorCode;
        DialogErrorDisplay(daqCtrl);
        }
}

/*****************************************************************************************/

int CVIFUNC DaqCtrl_EditPopup(DaqCtrl daqCtrl, char *dialogTitle, int modal)
{
    int     error = 0;
    int     wasModal;
    int     panel = daqCtrl->panel;
    int     daqCtrlId = daqCtrl->baseCtrl;
    int     isDisplayed;

    if (dialogTitle)
        errChk( daqCtrl->setAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, dialogTitle));

    if (modal)
        {
        errChk( daqCtrl->getAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, &wasModal));
        errChk( daqCtrl->setAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, TRUE));
        errChk( daqCtrl->getAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED, &isDisplayed));

        if (!wasModal && isDisplayed)       /* must hide and reshow for modal change to take effect */
            errChk( daqCtrl->setAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED, 0));
        }

    errChk( daqCtrl->setAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED, 1));

    if (modal)
        errChk( daqCtrl->setAttribute(panel, daqCtrlId, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, wasModal));

    daqCtrl->closeTheDialog = FALSE;

    while (ObjectExists(daqCtrl) && !daqCtrl->closeTheDialog)
        ProcessSystemEvents();

Error:
    return error;
}

/*****************************************************************************************/

static void SetTabOrderForList(int panel, ListType ctrlList)
{
    int numCtrls, index;
    int tabPos = 0;
    int savedBole = SetBOLE(FALSE);
    int ctrl;
    int error;

    numCtrls = ListNumItems(ctrlList);

    for (index = 1; index <= numCtrls; index++)
        {
        ListGetItem(ctrlList, &ctrl, index);
        error = SetCtrlAttribute(panel, ctrl, ATTR_CTRL_TAB_POSITION, tabPos);

            /* will return error for indicator only controls */
        if (error >= 0)
            tabPos++;
        }

    SetBOLE(savedBole);
}

/*****************************************************************************************/

    /*  increases the value of daqCtrl->minDialogWidth as needed */
int CVIFUNC DaqCtrl_CalcMinDialogWidth(DaqCtrl daqCtrl)
{
    int         error = 0;
    int         dialog;
    int         hMargin;
    int         maxLabelWidth;
    int         ctrlWidth;
    int         contentWidth;
    int         sheetMargin;
    int         ctrlHMargin;
    int         ctrlLeft;
    int         minDialogWidth;

    dialog = daqCtrl->dialog;

    hMargin = daqCtrl->hMargin;
    sheetMargin = hMargin;
    ctrlHMargin = hMargin + sheetMargin;

        /*** Horizontally position numerics and their labels ***/

        /* put the control labels on the left */
    errChk( PutLabelsOnLeftForList(dialog, daqCtrl->ctrlsWithLabelsOnLeft, 0));

        /* find the largest label */
    errChk( AttrMaxAndMinForList(dialog, daqCtrl->ctrlsWithLabelsOnLeft, ATTR_LABEL_WIDTH, &maxLabelWidth, 0, 0, 0));

        /* calculate the left position of the numeric ctrls based on the margin width, the largest label size, and the label-ctrl gap */
    ctrlLeft = ctrlHMargin + maxLabelWidth + DLG_LABEL_TO_CTRL_MIN_H_GAP;

#ifdef RESIZE_FOR_TITLEBAR
        /* make an effort to accomodate the title width */
    errChk( GetPanelAttribute (dialog, ATTR_TITLE, title));
    errChk( GetTextDisplaySize (title, VAL_MENU_META_FONT, 0, &titleWidth));
    daqCtrl->minDialogWidth = Max(titleWidth+70, daqCtrl->minDialogWidth); /* +70 to account for room taken by the close box + minimize box */
#endif

    errChk( GetCtrlAttribute(dialog, daqCtrl->deviceNumeric, ATTR_WIDTH, &ctrlWidth));
    contentWidth = maxLabelWidth + DLG_LABEL_TO_CTRL_MIN_H_GAP + ctrlWidth;
    minDialogWidth = contentWidth + 2*hMargin + 2*sheetMargin;
    daqCtrl->minDialogWidth = Max(daqCtrl->minDialogWidth, minDialogWidth);

Error:
    return error;
}

/******************************************************************************/

int CVIFUNC DaqCtrl_FormatDialog(DaqCtrl daqCtrl)
{
    int         error = 0;
    int         dialog;
    int         vMargin;
    int         hMargin;
    int         ctrlWidth;
    int         contentWidth;
    int         sheetMargin;
    int         index;
    int         numLists;
    ListType    listToDistribute;
    int         ctrlHMargin;
    int         ctrlLeft;
    int         sheetTop;
    int         top;
    Rect        r;
    
    dialog = daqCtrl->dialog;

    vMargin = daqCtrl->vMargin;  /* make sure the error message has room */
    hMargin = daqCtrl->hMargin;
    sheetMargin = hMargin;
    ctrlHMargin = hMargin + sheetMargin;

    errChk( EasyTab_SetBounds (dialog, daqCtrl->sheetGroupCtrl, VAL_EASY_TAB_EXTERIOR_BOUNDS, MakeRect(daqCtrl->vMargin, daqCtrl->hMargin, 1, 1))); 


    errChk( GetCtrlAttribute(dialog, daqCtrl->deviceNumeric, ATTR_WIDTH, &ctrlWidth));
    ctrlLeft = daqCtrl->minDialogWidth- hMargin - sheetMargin - ctrlWidth;

    contentWidth = daqCtrl->minDialogWidth - 2 * sheetMargin;

        /* finally move the ctrls to their appropriate horizontal position */
    errChk( SetAttributeForList(dialog, daqCtrl->ctrlsWithLabelsOnLeft, ATTR_LEFT, ctrlLeft));

        /* align the labels flush with the left margin */
    errChk( SetAttributeForList(dialog, daqCtrl->ctrlsWithLabelsOnLeft, ATTR_LABEL_LEFT, ctrlHMargin));

        /* scoot error notificatation ring back to compensate for its extra width */
    SetCtrlRightPosition(dialog, daqCtrl->errorNotificationRing, daqCtrl->minDialogWidth- hMargin - sheetMargin);
    errChk( SetCtrlAttribute(dialog, daqCtrl->errorNotificationRing, ATTR_LABEL_LEFT, ctrlHMargin)); /* must be reset after moving the control */

            /*  size the config buttons */
    errChk( SetCtrlsToAttributeExtreme (dialog, ATTR_WIDTH, VAL_TB_MAX, 0, daqCtrl->saveButton, daqCtrl->recallButton, daqCtrl->deleteConfigButton, 0));

        /* make the ok and cancel buttons the same width */
    errChk( SetCtrlsToAttributeExtreme(dialog, ATTR_WIDTH, VAL_TB_MAX, 2, daqCtrl->okButton, daqCtrl->cancelButton));

        /* position the list boxes, check boxes and buttons flush to the left margin */
    errChk( SetAttributeForList (dialog, daqCtrl->ctrlsOnLeft, ATTR_LEFT, ctrlHMargin));

        /* distribute the ctrls vertically */
    errChk( EasyTab_GetBounds (dialog, daqCtrl->sheetGroupCtrl, VAL_EASY_TAB_INTERIOR_BOUNDS, &r));
    sheetTop = r.top;
    top = sheetTop + vMargin;

    numLists = ListNumItems(daqCtrl->ctrlListsToDistribute);
    for (index = 1; index <= numLists; index++)
        {
        top = sheetTop + vMargin;
        ListGetItem(daqCtrl->ctrlListsToDistribute, &listToDistribute, index);

        errChk( DistributeCtrlsInList(dialog, listToDistribute, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP));
        SetTabOrderForList(dialog, listToDistribute);
        }

Error:
    return error;
}

/******************************************************************************/

int CVIFUNC DaqCtrl_FinishDialogSizing(DaqCtrl daqCtrl)
{
    int     error = 0;
    int     sheetTop;
    int     sheetLeft;
    int     sheetHeight;
    int     sheetWidth;
    int     dialog;
    int     top;
    int     ctrlHeight;
    int     btnWidth;
    int     hMargin;
    int     screenHeight;
    int     sheetBottom;
    int     okLeft;
    Rect    r;

    dialog = daqCtrl->dialog;
    hMargin = daqCtrl->hMargin;

    errChk( EasyTab_AutoSize(dialog, daqCtrl->sheetGroupCtrl));
    errChk( EasyTab_GetBounds(dialog, daqCtrl->sheetGroupCtrl, VAL_EASY_TAB_INTERIOR_BOUNDS, &r));
    sheetTop = r.top;
    sheetHeight = r.height;
    sheetLeft = r.left;
    sheetWidth = r.width;

    top = sheetBottom = sheetTop + sheetHeight;
    top += (int)(DLG_CTRL_V_GAP * 1.1);  /* a little extra space between buttons and the other controls */

        /* position ok and cancel buttons vertically */
    errChk( SetAttributeForCtrls(dialog, ATTR_TOP, top, 2, daqCtrl->okButton, daqCtrl->cancelButton));
    errChk( GetCtrlAttribute(dialog, daqCtrl->okButton, ATTR_HEIGHT, &ctrlHeight));
    top += ctrlHeight + DLG_CTRL_V_GAP;

        /* position the ok and cancel buttons horizontally */
    errChk( GetCtrlAttribute(dialog, daqCtrl->okButton, ATTR_WIDTH, &btnWidth));
    errChk( SetCtrlAttribute(dialog, daqCtrl->cancelButton, ATTR_LEFT, sheetWidth + hMargin - btnWidth));
    okLeft = sheetWidth + hMargin - 2*btnWidth - DLG_BTN_H_GAP;
    errChk( SetCtrlAttribute(dialog, daqCtrl->okButton, ATTR_LEFT, okLeft));

        /* make certain strings longer */
    ExpandCtrlToMeetLabel(dialog, daqCtrl->titleString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);
    ExpandCtrlToMeetLabel(dialog, daqCtrl->configFileString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);
    ExpandCtrlToMeetLabel(dialog, daqCtrl->configNameString, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0);

        /*** Set the panel size */
    errChk( SetPanelAttribute(dialog, ATTR_WIDTH, sheetWidth + 2 * hMargin));
    errChk( SetPanelAttribute(dialog, ATTR_HEIGHT, top));
    daqCtrl->heightWithoutErrorExplanation = top;

        /*** Set daq error explanation box size ***/

    errChk( GetScreenSize(&screenHeight, 0));
    errChk( GetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_HEIGHT, &ctrlHeight));
    daqCtrl->heightWithErrorExplanation = top + ctrlHeight;

    if (daqCtrl->heightWithErrorExplanation > screenHeight - 30) /* -30 gives consideration for a litte title bar and window frame room as well */
        {
        int numVisLines;
        int lineHeight;
        int height;
        int desiredHeight;
        int numLinesThatWillFit;
        int actualHeight;

        errChk( GetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_VISIBLE_LINES, &numVisLines));
        errChk( GetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_HEIGHT, &height));
        lineHeight = (height - 2 * TEXT_BOX_FRAME_WIDTH) / numVisLines;

        desiredHeight = top - sheetBottom;
        numLinesThatWillFit = desiredHeight / lineHeight;
        errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_VISIBLE_LINES, numLinesThatWillFit));
        errChk( GetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_HEIGHT, &actualHeight));

        daqCtrl->heightWithErrorExplanation = top;
        errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_TOP, sheetBottom + (desiredHeight - actualHeight)/ 2));
        errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_WIDTH, okLeft-3));
        }
    else
        {
        errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_TOP, top));
        errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_WIDTH, sheetWidth + 2 * hMargin));
        }



Error:
    return error;
}

/*********************************************************************/

int CVIFUNC CreateDaqDlg(DaqCtrl daqCtrl)
{
    int error = 0;

    errChk( daqCtrl->dialog = NewPanel(0, daqCtrl->dialogTitle, 10,10,10,10));
    errChk( SetPanelAttribute(daqCtrl->dialog, ATTR_CAN_MAXIMIZE, 0));
    errChk( SetPanelAttribute (daqCtrl->dialog, ATTR_SIZABLE, 0));
    errChk( SetPanelAttribute (daqCtrl->dialog, ATTR_CONFORM_TO_SYSTEM, 1));

Error:
    return error;
}

/*********************************************************************/

    /*  call before calling the functions that create controls */
int CVIFUNC CreateDlgLists(DaqCtrl daqCtrl)
{
    int error = 0;

    errChk( CreateNonSheetLists(daqCtrl));
    errChk( CreateSheetLists(daqCtrl));

Error:
    return 0;
}

/*********************************************************************/

    /* must be called after CreateErrorControls() because it needs the margins */
int CVIFUNC CreateDlgSheetGroup(DaqCtrl daqCtrl)
{
    int error = 0;

    errChk( daqCtrl->sheetGroupCtrl = EasyTab_Create(daqCtrl->dialog, daqCtrl->vMargin, daqCtrl->hMargin, 0, 0));
    /* set size, since we will be inquiring about the about bounds before setting them again, and that would trigger an autosize */
    errChk( EasyTab_SetBounds (daqCtrl->dialog, daqCtrl->sheetGroupCtrl, VAL_EASY_TAB_EXTERIOR_BOUNDS, MakeRect(daqCtrl->vMargin, daqCtrl->hMargin, 1, 1))); 

Error:
    return error;
}

/*********************************************************************/

    /*  also initializes daqCtrl->vMargin and daqCtrl->hMargin */
int CVIFUNC CreateErrorControls(DaqCtrl daqCtrl)
{
    int error = 0;
    int dialog = daqCtrl->dialog;
    int errorTextMsgHeight;

        /* Create the DAQ error message */
    errChk( daqCtrl->errorTextMsg = NewCtrl (dialog, CTRL_TEXT_MSG, "", 0, 0));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextMsg, ATTR_TEXT_COLOR, VAL_RED));
    errChk( GetCtrlAttribute (dialog, daqCtrl->errorTextMsg, ATTR_HEIGHT, &errorTextMsgHeight));
    daqCtrl->vMargin = Max(DLG_MARGIN, errorTextMsgHeight); /* make sure the error message has room */
    daqCtrl->hMargin = DLG_MARGIN;

        /* Create the DAQ error explanation box */
    errChk( daqCtrl->errorTextBox = NewCtrl (dialog, CTRL_TEXT_BOX, "", 0, 0));
    errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_VISIBLE, FALSE));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_FRAME_COLOR, VAL_TRANSPARENT));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_TEXT_COLOR, VAL_RED));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_WRAP_MODE, VAL_WORD_WRAP));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_CTRL_MODE, VAL_NORMAL));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_NO_EDIT_TEXT, TRUE));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_SCROLL_BARS, VAL_VERT_SCROLL_BAR));
    errChk( SetCtrlAttribute (dialog, daqCtrl->errorTextBox, ATTR_SCROLL_BAR_SIZE, VAL_SMALL_SCROLL_BARS));
    errChk( SetCtrlAttribute(dialog, daqCtrl->errorTextBox, ATTR_VISIBLE_LINES, 4));

    errChk( InstallCtrlCallback (dialog, daqCtrl->errorTextBox, ErrorTextBoxCallback, 0));

Error:
    return error;
}

/*********************************************************************/

int CVIFUNC CreateConfigCtrls(DaqCtrl daqCtrl)
{
    int error = 0;
    int dialog = daqCtrl->dialog;

    errChk( daqCtrl->configFileString =     NewCtrl (dialog, CTRL_STRING, "Configuration File:", 0, 0));
    errChk( daqCtrl->configNameString =     NewCtrl (dialog, CTRL_STRING, "Configuration Name:", 0, 0));
    errChk( daqCtrl->saveButton =           NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Save...", 0, 0));
    errChk( daqCtrl->recallButton =         NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Recall...", 0, 0));
    errChk( daqCtrl->deleteConfigButton =   NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Delete...", 0, 0));

    errChk( SetCtrlVal(dialog, daqCtrl->configFileString, GetCustCtrlIniFilePath()));
    errChk( SetCtrlVal(dialog, daqCtrl->configNameString, daqCtrl->settingsName));

    errChk( MakeCtrlGrayRecessedIndicator(dialog, daqCtrl->configFileString));
    errChk( MakeCtrlGrayRecessedIndicator(dialog, daqCtrl->configNameString));

    errChk( InstallCtrlCallback (dialog, daqCtrl->saveButton,             EditConfigDlgCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->recallButton,           EditConfigDlgCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->deleteConfigButton,     EditConfigDlgCallback, (void *)daqCtrl));

    nullChk( AppendToIntegerList(daqCtrl->ctrlsWithLabelsOnLeft, 0, daqCtrl->configFileString, daqCtrl->configNameString, 0));

    nullChk( AppendToIntegerList(daqCtrl->ctrlsOnLeft, 0, daqCtrl->saveButton, daqCtrl->recallButton, daqCtrl->deleteConfigButton, 0));

    nullChk( AppendToIntegerList(daqCtrl->configTabCtrls, 0, daqCtrl->configFileString, daqCtrl->configNameString,
        daqCtrl->saveButton, daqCtrl->recallButton, daqCtrl->deleteConfigButton, 0));

Error:
    return error;
}

/*********************************************************************/

int CVIFUNC CreateGeneralCtrls(DaqCtrl daqCtrl)
{
    int error = 0;
    int dialog = daqCtrl->dialog;

    errChk( daqCtrl->titleString =      NewCtrl (dialog, CTRL_STRING, "Dialog Title:", 0, 0));
    errChk( daqCtrl->errorNotificationRing = NewCtrl (dialog, CTRL_RING, "Error Notification:", 0, 0));
    errChk( daqCtrl->modalCheckBox =    NewCtrl (dialog, CTRL_CHECK_BOX, "Edit Dialog Pops Up Modally", 0, 0));
    errChk( daqCtrl->allowRightClickEditCheckBox = NewCtrl (dialog, CTRL_CHECK_BOX, "Allow Right Click Editing", 0, 0));
    errChk( daqCtrl->visibleCheckBox =  NewCtrl (dialog, CTRL_CHECK_BOX, "Control Is Visible", 0, 0));
    errChk( daqCtrl->tieLimitsCheckBox = NewCtrl (dialog, CTRL_CHECK_BOX, "Tie Control Range To Input Limits", 0, 0));

    if (GetHelpFilePath(0))
        {
        errChk( daqCtrl->helpButton =   NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "  __Help...  ", 0, 0));
        SetCtrlAttribute (dialog, daqCtrl->helpButton, ATTR_SHORTCUT_KEY, VAL_F1_VKEY);
        errChk( InstallCtrlCallback (dialog, daqCtrl->helpButton, HelpCallback, (void *)daqCtrl));
        }
    else
        daqCtrl->helpButton = 0;    /* 0 will terminate the argument list in the AppendToIntegerList() calls if the help button isn't present */

    errChk( InstallCtrlCallback (dialog, daqCtrl->errorNotificationRing,    DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->modalCheckBox,            DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->allowRightClickEditCheckBox, DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->visibleCheckBox,          DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->tieLimitsCheckBox,        DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->titleString,              DAQ_CtrlDialogCallback, (void *)daqCtrl));

    errChk( InsertListItem(dialog, daqCtrl->errorNotificationRing, -1, "Beep And Box", ERR_NOTIFY_BEEP_AND_BOX));
    errChk( InsertListItem(dialog, daqCtrl->errorNotificationRing, -1, "Beep Only", ERR_NOTIFY_BEEP_ONLY));
    errChk( InsertListItem(dialog, daqCtrl->errorNotificationRing, -1, "Box Only", ERR_NOTIFY_BOX_ONLY));
    errChk( InsertListItem(dialog, daqCtrl->errorNotificationRing, -1, "No Beep Or Box", ERR_NOTIFY_NONE));
    errChk( SizeRingCtrlToText(dialog, daqCtrl->errorNotificationRing));

    nullChk( AppendToIntegerList(daqCtrl->ctrlsWithLabelsOnLeft, 0, daqCtrl->titleString, daqCtrl->errorNotificationRing, 0));

        /*  Help button must be last in these calls since it might be zero */
    nullChk( AppendToIntegerList(daqCtrl->ctrlsOnLeft, 0, daqCtrl->modalCheckBox, daqCtrl->allowRightClickEditCheckBox, daqCtrl->visibleCheckBox, daqCtrl->tieLimitsCheckBox, daqCtrl->helpButton, 0));

    nullChk( AppendToIntegerList(daqCtrl->generalTabCtrls, 0, daqCtrl->titleString, daqCtrl->errorNotificationRing, daqCtrl->modalCheckBox,
        daqCtrl->allowRightClickEditCheckBox, daqCtrl->visibleCheckBox, daqCtrl->tieLimitsCheckBox, daqCtrl->helpButton, 0));

    nullChk( AppendToIntegerList(daqCtrl->generalTabCtrlsToDistribute, 0, daqCtrl->titleString, daqCtrl->errorNotificationRing, daqCtrl->modalCheckBox,
        daqCtrl->allowRightClickEditCheckBox, daqCtrl->visibleCheckBox, daqCtrl->tieLimitsCheckBox, daqCtrl->helpButton, 0));

Error:
    return error;
}

/*********************************************************************/

int CVIFUNC CreateDAQCtrls(DaqCtrl daqCtrl)
{
    int error = 0;
    int dialog = daqCtrl->dialog;

    errChk( daqCtrl->deviceNumeric =        NewCtrl (dialog, CTRL_NUMERIC, "Device:", 0, 0));
    errChk( daqCtrl->channelStringCtrl =    NewCtrl (dialog, CTRL_STRING, "Channel String:", 0, 0));
    errChk( daqCtrl->highLimitNumeric =     NewCtrl (dialog, CTRL_NUMERIC, "High Limit (Volts):", 0, 0));
    errChk( daqCtrl->lowLimitNumeric =      NewCtrl (dialog, CTRL_NUMERIC, "Low Limit (Volts):", 0, 0));
    errChk( daqCtrl->ioDirectionRing =      NewCtrl (dialog, CTRL_RING, "Input/Output Direction:", 0, 0));
    errChk( daqCtrl->enabledCheckBox =      NewCtrl (dialog, CTRL_CHECK_BOX, "Enabled", 0, 0));

    errChk( SetCtrlAttribute(dialog, daqCtrl->channelStringCtrl, ATTR_MAX_ENTRY_LENGTH, MAX_CHANNEL_STRING_LEN));

    errChk( InsertListItem (dialog, daqCtrl->ioDirectionRing, -1, "Input", ANALOG_INPUT));
    errChk( InsertListItem (dialog, daqCtrl->ioDirectionRing, -1, "Output", ANALOG_OUTPUT));

    errChk( SetCtrlAttribute(dialog, daqCtrl->deviceNumeric, ATTR_DATA_TYPE, VAL_INTEGER));
    errChk( SetCtrlAttribute(dialog, daqCtrl->highLimitNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));
    errChk( SetCtrlAttribute(dialog, daqCtrl->lowLimitNumeric, ATTR_DATA_TYPE, VAL_DOUBLE));

    errChk( SetCtrlAttribute(dialog, daqCtrl->highLimitNumeric, ATTR_PRECISION, 5));    /* as per car# B0BK6B5 */
    errChk( SetCtrlAttribute(dialog, daqCtrl->lowLimitNumeric, ATTR_PRECISION, 5));     /* as per car# B0BK6B5 */

    errChk( SetCtrlAttribute (dialog, daqCtrl->deviceNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute(dialog, daqCtrl->deviceNumeric, ATTR_MIN_VALUE, MIN_DEVICE));
    errChk( SetCtrlAttribute(dialog, daqCtrl->deviceNumeric, ATTR_MAX_VALUE, MAX_DEVICE));

    errChk( InstallCtrlCallback (dialog, daqCtrl->deviceNumeric,            DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->channelStringCtrl,    DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->highLimitNumeric,     DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->lowLimitNumeric,      DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->ioDirectionRing,      DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->enabledCheckBox,    DAQ_CtrlDialogCallback, (void *)daqCtrl));

    nullChk( AppendToIntegerList(daqCtrl->ctrlsWithLabelsOnLeft, 0, daqCtrl->deviceNumeric, daqCtrl->channelStringCtrl,
        daqCtrl->highLimitNumeric, daqCtrl->lowLimitNumeric, daqCtrl->ioDirectionRing, 0));

    nullChk( AppendToIntegerList(daqCtrl->ctrlsOnLeft, 0, daqCtrl->enabledCheckBox, 0));

    nullChk( AppendToIntegerList(daqCtrl->daqTabCtrls, 0, daqCtrl->deviceNumeric, daqCtrl->channelStringCtrl,
        daqCtrl->highLimitNumeric, daqCtrl->lowLimitNumeric, daqCtrl->ioDirectionRing, daqCtrl->enabledCheckBox, 0));



Error:
    return error;
}

/*********************************************************************/

int CVIFUNC CreateStdButtons(DaqCtrl daqCtrl)
{
    int error = 0;
    int dialog = daqCtrl->dialog;

        /* Create the ok and cancel buttons */
    errChk( daqCtrl->okButton =        NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "__Done", 0, 0));
    errChk( daqCtrl->cancelButton =    NewCtrl (dialog, CTRL_SQUARE_COMMAND_BUTTON, "  __Revert  ", 0, 0));
    errChk( SetCommonDialogShortcutKeys(dialog, daqCtrl->okButton, daqCtrl->cancelButton, daqCtrl->okButton));

    errChk( InstallCtrlCallback (dialog, daqCtrl->okButton,           DAQ_CtrlDialogCallback, (void *)daqCtrl));
    errChk( InstallCtrlCallback (dialog, daqCtrl->cancelButton,       DAQ_CtrlDialogCallback, (void *)daqCtrl));

Error:
    return error;
}

/****************************************************************************************/

static int CVICALLBACK EditConfigDlgCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int         error = 0;
    int         swallowedEvent = 0;
    DaqCtrl     daqCtrl;
    int         result;
    ListType    settingsToDelete = 0;
    char        settingsName[SETTING_NAME_BUF_SIZE];
    char        errorDescription[512];
    int         warning = 0;

    UNUSED(eventData1);
    UNUSED(eventData2);

    daqCtrl = (DaqCtrl)callbackData;
    *errorDescription = 0;

    strcpy(settingsName, daqCtrl->settingsName);

    switch (event)
        {
        case EVENT_COMMIT:
            swallowedEvent = 1;   /* this dialog isn't known to the user's code, so we don't want them to get its events if they use GetUserEvent() */
            if (control == daqCtrl->saveButton)
                {
                errChk( result = ConfigDialog("Save Configuration", daqCtrl->settingsPrefix, settingsName, TRUE, TRUE, TRUE, FALSE, 0, TRUE));
                if (result == CONFIG_DLG_OK && ObjectExists(daqCtrl))
                    {
                    errChk( daqCtrl->saveSettings(daqCtrl, settingsName));
                    }
                }
            else
            if (control == daqCtrl->recallButton)
                {
                errChk( result = ConfigDialog("Recall Configuration", daqCtrl->settingsPrefix, settingsName, FALSE, FALSE, TRUE, FALSE, 0, FALSE));
                if (result == CONFIG_DLG_OK && ObjectExists(daqCtrl))
                    {
                    errChk( warning = daqCtrl->recallSettings(daqCtrl, settingsName, errorDescription));
                    errChk( DaqCtrl_SetChangedState(daqCtrl, TRUE));
                    }
                }
            else
            if (control == daqCtrl->deleteConfigButton)
                {
                errChk( result = ConfigDialog("Delete Configurations", daqCtrl->settingsPrefix, settingsName, FALSE, FALSE, FALSE, TRUE, &settingsToDelete, FALSE));
                if (result == CONFIG_DLG_OK && ObjectExists(daqCtrl))
                    errChk( DeleteSettings(settingsToDelete, daqCtrl->settingsPrefix));
                }
            break;
        }

    if (ObjectExists(daqCtrl))
        errChk( SetCtrlVal(dialog, daqCtrl->configNameString, daqCtrl->settingsName));  /* in case it changed, update the dialog */

Error:
    if (error >= 0)
        error = 0;

    if (!error && warning)
        error = warning;

    if (error != 0)
        {
        char    *action = "";
        char    errMsg[512];

        if (control == daqCtrl->saveButton)
            action = "saving a configuration";
        else
        if (control == daqCtrl->recallButton)
            action = "recalling a configuration";
        else
        if (control == daqCtrl->deleteConfigButton)
            action = "deleting configuration(s)";

        sprintf(errMsg, "%s %d occurred while %s.\n\n%s\n\n%s", error > 0 ? "Warning: " : "Error: ", -ABS_VAL(error), action, GetGeneralErrorString(error), errorDescription);
        if (error < 0)
            MessagePopup("Configuration Error", errMsg);
        else
            MessagePopup("Configuration Warning", errMsg);
        }

    ListDispose(settingsToDelete);
    return swallowedEvent;
}

/*****************************************************************************************/

static int CVICALLBACK HelpCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     error = 0;
    DaqCtrl daqCtrl;
    int     swallowEvent = 0;
    char    helpFilePath[MAX_PATHNAME_LEN];
    char    errMsg[512];

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    daqCtrl = (DaqCtrl)callbackData;
    daqCtrl->dialogEvent = event;

    switch (event)
        {
        case EVENT_COMMIT:
            swallowEvent = 1;   /* this dialog isn't known to the user's code, so we don't want him to get its events if he uses GetUserEvent() */
            if (GetHelpFilePath(helpFilePath))
                {
                if (SystemHelp (helpFilePath, HELP_CONTENTS, 0, 0) == 0)
                    MessagePopup("Error Starting Help", "The help application could not be started");;
                }
            else
                {
                sprintf(errMsg, "Could not find help file '%s'", DAQ_CTRL_HELP_FILE);
                MessagePopup("Error Finding Help File", errMsg);
                }
            break;
        }

/*Error:*/
    if (ObjectExists(daqCtrl))
        {
        daqCtrl->dialogEvent = 0;

        if (error < 0)
            DaqCtrl_DisplayError((DaqCtrl)daqCtrl, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);
        }
    return swallowEvent;
}

/*****************************************************************************************/

static int CVICALLBACK ErrorTextBoxCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    UNUSED(dialog);
    UNUSED(control);
    UNUSED(callbackData);
    UNUSED(eventData1);
    UNUSED(eventData2);

    switch (event)
        {
        case EVENT_LEFT_CLICK:
            SetActiveCtrl(dialog, control);
            break;
        }

    return 0;
}

/*****************************************************************************************/

int CVICALLBACK DAQ_CtrlDialogCallback(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     error = 0;
    DaqCtrl daqCtrl;
    int     swallowEvent = 0;
    int     updateDlg = FALSE;

    UNUSED(eventData1);
    UNUSED(eventData2);

    daqCtrl = (DaqCtrl)callbackData;
    daqCtrl->dialogEvent = event;

    switch (event)
        {
        case EVENT_VAL_CHANGED:
            errChk( DaqCtrl_SetChangedState(daqCtrl, TRUE));

            if (control == daqCtrl->allowRightClickEditCheckBox)    /* don't apply until confirmed during the COMMIT EVENT */
                break;
            if (control == daqCtrl->visibleCheckBox)                /* don't apply until confirmed during the COMMIT EVENT */ 
                break;

            if (control != daqCtrl->channelStringCtrl)
                updateDlg = TRUE;           /* allow channel string to have an unaccepted channel string while typing */
            errChk( daqCtrl->applyDialogSettings(daqCtrl));
            errChk( daqCtrl->updateDialogDimming(daqCtrl));
            break;
        case EVENT_COMMIT:
            swallowEvent = 1;   /* this dialog isn't known to the user's code, so we don't want him to get its events if he uses GetUserEvent() */
            if (control == daqCtrl->cancelButton)
                {
                errChk( HandleRevert(daqCtrl));
                }
            else
            if (control == daqCtrl->okButton)
                {
                errChk( daqCtrl->setAttribute(daqCtrl->panel, daqCtrl->baseCtrl, ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED, 0));
                }
            else
            if (control == daqCtrl->channelStringCtrl)
                {
                /*  change channel string on commit, so we only complain about 'finished' channel 
                    strings (special code in SetDaqChartChannelString() ignores syntax errors during EVENT_VAL_CHANGED) */
                errChk( daqCtrl->applyDialogSettings(daqCtrl));
                errChk( daqCtrl->updateDialogDimming(daqCtrl));
                updateDlg = TRUE;
                }
            else
            if (control == daqCtrl->allowRightClickEditCheckBox || control == daqCtrl->visibleCheckBox)
                {
                int confirmed, value;

                errChk( GetCtrlVal(dialog, control, &value));

                if (!value)
                    {
                    char    msg[200];

                    if (control == daqCtrl->allowRightClickEditCheckBox)
                        sprintf(msg, "Are you sure you want to disable right click editing for this %s?", daqCtrl->ctrlTypeName);
                    else
                        sprintf(msg, "Are you sure you want to make this %s invisible?", daqCtrl->ctrlTypeName);

                    errChk( confirmed = ConfirmPopup("Confirmation", msg));
                    if (ObjectExists(daqCtrl))
                        {
                        if (!confirmed)
                            errChk( SetCtrlVal(dialog, control, TRUE));
                        else
                            errChk( DaqCtrl_SetChangedState(daqCtrl, TRUE));
                        }
                    }
                else
                    errChk( DaqCtrl_SetChangedState(daqCtrl, TRUE));

                errChk( daqCtrl->applyDialogSettings(daqCtrl));
                errChk( daqCtrl->updateDialogDimming(daqCtrl));
                updateDlg = TRUE;
                }
            break;
        }

Error:
    daqCtrl->dialogEvent = 0;
    if (updateDlg)
        daqCtrl->copyAttributesToDialog(daqCtrl);

    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)daqCtrl, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return swallowEvent;
}

/*****************************************************************************************/

    /*  This callback is used instead of DAQ_CtrlDialogCallback() by controls on which a value change does not imply
        that a revertable change has been made to the object  (examples: the Start/Stop Logging button and Pause button
        on the DAQ Chart).
    */
int CVICALLBACK DAQ_CtrlDialogCallbackNoChange(int dialog, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    int     error = 0;
    DaqCtrl daqCtrl;
    int     swallowEvent = 0;

    UNUSED(dialog);
    UNUSED(control);
    UNUSED(eventData1);
    UNUSED(eventData2);

    daqCtrl = (DaqCtrl)callbackData;
    daqCtrl->dialogEvent = event;

    switch (event)
        {
        case EVENT_VAL_CHANGED:
            errChk( daqCtrl->applyDialogSettings(daqCtrl));
            errChk( daqCtrl->updateDialogDimming(daqCtrl));
            break;
        case EVENT_COMMIT:
            swallowEvent = 1;   /* this dialog isn't known to the user's code, so we don't want him to get its events if he uses GetUserEvent() */
            break;
        }

Error:
    daqCtrl->dialogEvent = 0;

    if (error < 0)
        DaqCtrl_DisplayError((DaqCtrl)daqCtrl, error, MIN_ERROR_DISPLAY_TIME, MIN_CTRL_ERROR_DISPLAY_TIME_TILL_CLEARED);

    return swallowEvent;
}

/*****************************************************************************************/

static int HandleRevert(DaqCtrl daqCtrl)
{
    int         error = 0;
    RevertReply reply;
    int         confirmed;

    if (daqCtrl->changed)
        {
        if (daqCtrl->confirmRevertFunc)
            {
            errChk( daqCtrl->confirmRevertFunc(daqCtrl, &reply));
            }
        else
            {
            char    msg[128];

            sprintf(msg,"Revert Changes To %s Configuration?", daqCtrl->ctrlTypeName);
            errChk( confirmed = ConfirmPopup ("Confirmation", msg));
            if (confirmed)
                reply = ALL_CHANGES_WILL_REVERT;
            }

        if (reply != DONT_REVERT && ObjectExists(daqCtrl))
            {
            errChk( daqCtrl->revertAttributes(daqCtrl));
            errChk( daqCtrl->copyAttributesToDialog(daqCtrl));
            if (reply == ALL_CHANGES_WILL_REVERT)
                errChk( DaqCtrl_SetChangedState(daqCtrl, FALSE));
            }
        }

Error:
    return error;
}

/*****************************************************************************************/

    /*  returns 0 if no error, positive error code if warning, negative error code if fatal error */
int CVIFUNC DaqCtrl_ReadSettings(DaqCtrl daqCtrl, IniText iniText, char *section, char *errorDescription, int *doingGets)
{
    int     error = 0;
    int     panel = daqCtrl->panel;
    int     ctrl = daqCtrl->baseCtrl;
    int     device;
    char    channelString[256];
    char    dialogTitle[256];
    double  highLimit, lowLimit;
    int     enabled, ioDirection, visible, allowRightClickEdit, modal;
    int     tieToLimits;
    int     recalledVersion;
    int     errorNotification;
    int     foundDevice, foundChannel, foundTieToLimits, foundHighLimit;
    int     foundLowLimit, foundEnabled;
    int     foundIODirection, foundVisible, foundAllowRightClickEdit, foundDialogTitle;
    int     foundModal, foundVersion, foundErrorNotification;
    int     result, warning = 0;

    UNUSED(errorDescription);

    *doingGets = TRUE;
    errChk( foundDevice             = Ini_GetInt (iniText, section, DEVICE_TAG, &device));
    if (!foundDevice)
        errChk( foundDevice         = Ini_GetInt (iniText, section, BETA_DEVICE_TAG, &device));
    errChk( foundChannel            = Ini_GetStringIntoBuffer (iniText, section, CHANNEL_TAG, channelString, 256));
    errChk( foundHighLimit          = Ini_GetDouble (iniText, section, HIGH_LIMIT_TAG, &highLimit));
    errChk( foundLowLimit           = Ini_GetDouble (iniText, section, LOW_LIMIT_TAG, &lowLimit));
    errChk( foundEnabled            = Ini_GetBoolean (iniText, section, ENABLED_TAG, &enabled));
    errChk( foundIODirection        = Ini_GetInt (iniText, section, IO_DIRECTION_TAG, &ioDirection));
    errChk( foundVisible            = Ini_GetBoolean (iniText, section, VISIBLE_TAG, &visible));
    errChk( foundAllowRightClickEdit= Ini_GetBoolean (iniText, section, ALLOW_RIGHT_CLICK_EDIT_TAG, &allowRightClickEdit));
    errChk( foundDialogTitle        = Ini_GetStringIntoBuffer (iniText, section, DIALOG_TITLE_TAG, dialogTitle, 256));
    errChk( foundModal              = Ini_GetBoolean (iniText, section, DIALOG_MODAL_TAG, &modal));
    errChk( foundTieToLimits        = Ini_GetBoolean (iniText, section, TIE_TO_LIMITS_TAG, &tieToLimits));
    errChk( foundVersion            = Ini_GetInt (iniText, section, VERSION_TAG, &recalledVersion));
    errChk( foundErrorNotification  = Ini_GetInt (iniText, section, ERROR_NOTIFICATION_TAG, &errorNotification));
    *doingGets = FALSE;

    if (foundDevice)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, device);
    if (!warning && result < 0) warning = -result;

    if (foundTieToLimits)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, tieToLimits);
    if (!warning && result < 0) warning = -result;

    if (foundHighLimit)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, (double)highLimit);
    if (!warning && result < 0) warning = -result;

    if (foundLowLimit)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, (double)lowLimit);
    if (!warning && result < 0) warning = -result;

    if (foundEnabled)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, enabled);
    if (!warning && result < 0) warning = -result;

    if (foundIODirection)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, ioDirection);
    if (!warning && result < 0) warning = -result;

    if (foundVisible)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, visible);
    if (!warning && result < 0) warning = -result;

    if (foundAllowRightClickEdit)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, allowRightClickEdit);
    if (!warning && result < 0) warning = -result;

    if (foundDialogTitle)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, dialogTitle);
    if (!warning && result < 0) warning = -result;

    if (foundModal)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, modal);
    if (!warning && result < 0) warning = -result;

    if (foundErrorNotification)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, errorNotification);
    if (!warning && result < 0) warning = -result;

    if (foundChannel)
        result = daqCtrl->setAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, channelString);
    if (!warning && result < 0) warning = -result;

    if (foundVersion)
        daqCtrl->lastRecalledVersion = recalledVersion;
    else
        daqCtrl->lastRecalledVersion = -1;

Error:
    if (error < 0)
        return error;
    else
        return warning; /*  bad set attribute call is a warning condition */
}

/****************************************************************************************/

int CVIFUNC DaqCtrl_WriteSettings(DaqCtrl daqCtrl, IniText iniText, char *section)
{
    int     error = 0;
    int     panel = daqCtrl->panel;
    int     ctrl = daqCtrl->baseCtrl;
    int     device;
    char    channelString[MAX_CHANNEL_STRING_LEN + 1];
    char    dialogTitle[MAX_EDIT_DIALOG_TITLE_LEN + 1];
    double  highLimit, lowLimit;
    int     enabled, ioDirection, visible, allowRightClickEdit, modal, tieToLimits;
    int     version, errorNotification;

    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_DEVICE, &device));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_CHANNEL_STRING, channelString));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS, &tieToLimits));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_HIGH_LIMIT, &highLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_LOW_LIMIT, &lowLimit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ENABLED, &enabled));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_IO_DIRECTION, &ioDirection));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_VISIBLE, &visible));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT, &allowRightClickEdit));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE, dialogTitle));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL, &modal));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_VERSION, &version));
    errChk( daqCtrl->getAttribute(panel, ctrl, ATTR_DAQ_CTRL_ERROR_NOTIFICATION, &errorNotification));

    errChk( Ini_PutInt      (iniText, section, DEVICE_TAG, device));
    errChk( Ini_PutString   (iniText, section, CHANNEL_TAG, channelString));
    errChk( Ini_PutBoolean  (iniText, section, TIE_TO_LIMITS_TAG, tieToLimits));
    errChk( Ini_PutDouble   (iniText, section, HIGH_LIMIT_TAG, highLimit));
    errChk( Ini_PutDouble   (iniText, section, LOW_LIMIT_TAG, lowLimit));
    errChk( Ini_PutBoolean  (iniText, section, ENABLED_TAG, enabled));
    errChk( Ini_PutInt      (iniText, section, IO_DIRECTION_TAG, ioDirection));
    errChk( Ini_PutBoolean  (iniText, section, VISIBLE_TAG, visible));
    errChk( Ini_PutBoolean  (iniText, section, ALLOW_RIGHT_CLICK_EDIT_TAG, allowRightClickEdit));
    errChk( Ini_PutString   (iniText, section, DIALOG_TITLE_TAG, dialogTitle));
    errChk( Ini_PutBoolean  (iniText, section, DIALOG_MODAL_TAG, modal));
    errChk( Ini_PutInt      (iniText, section, VERSION_TAG, version));
    errChk( Ini_PutInt      (iniText, section, ERROR_NOTIFICATION_TAG, errorNotification));

Error:
    return error;
}

/*******************************************************************************/

static int SetDialogTitle(DaqCtrl daqCtrl, char *newTitle)
{
    int error = 0;

    if (!newTitle)
        newTitle = "";

    StringCopyMax (daqCtrl->dialogTitle, newTitle, MAX_EDIT_DIALOG_TITLE_LEN);

    if (daqCtrl->dialog > 0)
        errChk( SetPanelAttribute(daqCtrl->dialog, ATTR_TITLE, daqCtrl->dialogTitle));

Error:
    return error;
}

/*******************************************************************************/

static void GetDialogTitle(DaqCtrl daqCtrl, char title[])
{
    StringCopyMax (title, daqCtrl->dialogTitle, MAX_EDIT_DIALOG_TITLE_LEN);
}

/*****************************************************************************************/

static int SetDialogDisplayedState(DaqCtrl daqCtrl, int display)
{
    int error = 0;

    if (display)
        {
        errChk( BuildDialogIfNeeded(daqCtrl));

        if (!daqCtrl->dialogDisplayed)
            {
            errChk( daqCtrl->rememberAttributes(daqCtrl));
            errChk( daqCtrl->copyAttributesToDialog(daqCtrl));

            errChk( DaqCtrl_SetChangedState(daqCtrl, FALSE));

            if (daqCtrl->shouldBeDisplayedModally)
                {
                errChk( InstallPopup(daqCtrl->dialog));
                daqCtrl->isDisplayedModally = 1;
                }
            else
                {
                errChk( DisplayPanel(daqCtrl->dialog));
                daqCtrl->isDisplayedModally = 0;
                }

            daqCtrl->dialogDisplayed = 1;
            }
        else
            errChk( DisplayPanel(daqCtrl->dialog)); /* already displayed, just bring to front */
        }
    else
        {
        if (daqCtrl->dialog && daqCtrl->dialogDisplayed)
            {
            daqCtrl->closeTheDialog = TRUE; /* in case the dialog is in a call to DaqCtrl_EditPopup */

            if (daqCtrl->isDisplayedModally)
                RemovePopup(daqCtrl->dialog);
            else
                HidePanel(daqCtrl->dialog);

            daqCtrl->errorReplaceableTime = Timer();
            daqCtrl->errorClearableTime = Timer();
            }

        daqCtrl->dialogDisplayed = 0;
        daqCtrl->isDisplayedModally = 0;
        }

Error:
    return 0;
}

/************************************************************************************/

int CVIFUNC DaqCtrl_UpdateDialogDimming(DaqCtrl daqCtrl)
{
    int error = 0;

    errChk( SetCtrlAttribute(daqCtrl->dialog, daqCtrl->cancelButton, ATTR_DIMMED, !daqCtrl->changed));

Error:
    return error;
}

/************************************************************************************/

    /*  This function is kind of a kludge to get around the lack of a property object (which could have been the target of this operation) */
void CVIFUNC DaqCtrl_ChannelStringIsRevertable(DaqCtrl daqCtrl, int isRevertable)
{
    daqCtrl->channelStringIsRevertable = isRevertable;
}

/************************************************************************************/

int CVIFUNC DaqCtrl_SetChangedState(DaqCtrl daqCtrl, int changed)
{
    int error = 0;

    daqCtrl->changed = changed;
    errChk( daqCtrl->updateDialogDimming(daqCtrl));

Error:
    return error;
}

/************************************************************************************/

int CVIFUNC DaqCtrl_SetAttributeFromParmInfo(DaqCtrl daqCtrl, int attribute, va_list parmInfo)
{
    int     error = 0;
    int     newIntValue;
    int     panel;
    int     baseCtrl;

    panel = daqCtrl->panel;
    baseCtrl = daqCtrl->baseCtrl;

    switch (attribute)
        {
        case ATTR_DAQ_CTRL_DEVICE:
            newIntValue = va_arg(parmInfo, int);
            if (newIntValue != Pin(newIntValue, MIN_DEVICE, MAX_DEVICE))
                errChk( UIEValueIsInvalidOrOutOfRange);
            daqCtrl->device = newIntValue;
            break;
        case ATTR_DAQ_CTRL_CHANNEL_STRING:
            {
            char    *newChannelString;
            char    *value;
            int     valueLength;
            unsigned long groupSize = 0;

            value = va_arg(parmInfo, char *);

            if (!value)
                errChk( UIENullPointerPassed);


            if (!daqCtrl->channelString || (strcmp(daqCtrl->channelString, value) != 0))
                {
                error = GetNumChannelsNoBOLE (daqCtrl->device, value, daqCtrl->ioDirection, &groupSize);
                    /*  don't complain about string syntax in dialog as the user types in the 
                        channel string */
                if (error >= 0 && groupSize > 1)
                    error = badChanErr;

                if ((error < 0 || groupSize != 1) && daqCtrl->dialog && daqCtrl->dialogDisplayed && daqCtrl->updatingCtrlFromDialog && daqCtrl->dialogEvent != EVENT_COMMIT)
                    {
                    error = 0;
                    goto Error; /* channel string is bad, but its not finished being typed, so give the user another chance before complaining */
                    }

                valueLength = Min(strlen(value), MAX_CHANNEL_STRING_LEN);

                nullChk( newChannelString = calloc(valueLength+1, 1));
                StringCopyMax(newChannelString, value, valueLength + 1);
                free(daqCtrl->channelString);
                daqCtrl->channelString = newChannelString;
                }
            }
           break;
        case ATTR_DAQ_CTRL_ENABLED:
            /* must be implemented by derived class */
            break;
        case ATTR_DAQ_CTRL_VISIBLE:
            errChk( SetCtrlAttribute(panel, baseCtrl, ATTR_VISIBLE, va_arg(parmInfo, int)));
            break;
        case ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT:
            daqCtrl->allowRightClickEdit = va_arg(parmInfo, int);
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE:
            errChk( SetDialogTitle(daqCtrl, va_arg(parmInfo, char *)));
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED:
            errChk( SetDialogDisplayedState(daqCtrl, va_arg(parmInfo, int)));
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL:
            daqCtrl->shouldBeDisplayedModally = va_arg(parmInfo, int);
            break;
        case ATTR_DAQ_CTRL_IO_DIRECTION:
            if (va_arg(parmInfo, int) == ANALOG_OUTPUT)
                daqCtrl->ioDirection = ANALOG_OUTPUT;
            else
                daqCtrl->ioDirection = ANALOG_INPUT;

            errChk( SetCtrlModeToIODirection(panel, baseCtrl, daqCtrl->ioDirection));
            break;
        case ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS:
        case ATTR_DAQ_CTRL_HIGH_LIMIT:
        case ATTR_DAQ_CTRL_LOW_LIMIT:
            /*  must be implemented by derived class */
            break;
        case ATTR_DAQ_CTRL_ERROR_NOTIFICATION:
            newIntValue = va_arg(parmInfo, int);
            switch (newIntValue)
                {
                case ERR_NOTIFY_BEEP_AND_BOX:
                case ERR_NOTIFY_BEEP_ONLY:
                case ERR_NOTIFY_BOX_ONLY:
                case ERR_NOTIFY_NONE:
                    daqCtrl->errorNotification =  newIntValue;
                    break;
                default:
                    errChk( UIEValueIsInvalidOrOutOfRange);
                }
            break;
        case ATTR_DAQ_CTRL_VERSION:
        case ATTR_DAQ_CTRL_CHANNEL_STRING_LENGTH:
        case ATTR_DAQ_CTRL_CHANNEL_STRING_COPY:
                errChk( UIEAttributeNotSettable);
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }
Error:
    return error;
}

/************************************************************************************/

int CVIFUNC DaqCtrl_GetAttribute(DaqCtrl daqCtrl, int attribute, void *value)
{
    int error = 0;

    if (!value)
        errChk( UIENullPointerPassed);

    switch (attribute)
        {
        case ATTR_DAQ_CTRL_DEVICE:
            *(int *)value = daqCtrl->device;
            break;
        case ATTR_DAQ_CTRL_CHANNEL_STRING:
            if (daqCtrl->channelString)
                strcpy((char *)value, daqCtrl->channelString);
            else
                strcpy((char *)value, "");
           break;
        case ATTR_DAQ_CTRL_CHANNEL_STRING_LENGTH:
            if (daqCtrl->channelString)
                *(int *)value = strlen(daqCtrl->channelString);
            else
                *(int *)value = 0;
            break;
        case ATTR_DAQ_CTRL_CHANNEL_STRING_COPY:
            if (daqCtrl->channelString)
                nullChk( *(char **)value = StrDup(daqCtrl->channelString));
            else
                nullChk( *(char **)value = StrDup(""));
            break;
        case ATTR_DAQ_CTRL_ENABLED:
            /* must be implemented by derived class */
            break;
        case ATTR_DAQ_CTRL_VISIBLE:
            errChk( GetCtrlAttribute(daqCtrl->panel, daqCtrl->baseCtrl, ATTR_VISIBLE, value));
            break;
        case ATTR_DAQ_CTRL_ALLOW_RCLICK_EDIT:
            *(int *)value = daqCtrl->allowRightClickEdit;
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE:
            GetDialogTitle((DaqCtrl)daqCtrl, (char *)value);
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE_LENGTH:
            if (daqCtrl->dialogTitle)
                *(int *)value = strlen(daqCtrl->dialogTitle);
            else
                *(int *)value = 0;
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_TITLE_COPY:
            if (daqCtrl->dialogTitle)
                nullChk( *(char **)value = StrDup(daqCtrl->dialogTitle));
            else
                nullChk( *(char **)value = StrDup(""));
            break;
        case ATTR_DAQ_CTRL_IO_DIRECTION:
            *(int *)value = daqCtrl->ioDirection;
            break;
        case ATTR_DAQ_CTRL_HIGH_LIMIT:
            *(double *)value = daqCtrl->highLimit;
            break;
        case ATTR_DAQ_CTRL_LOW_LIMIT:
            *(double *)value = daqCtrl->lowLimit;
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_DISPLAYED:
            *(int *)value = daqCtrl->dialogDisplayed;
            break;
        case ATTR_DAQ_CTRL_EDIT_DIALOG_MODAL:
            *(int *)value = daqCtrl->shouldBeDisplayedModally;
            break;
        case ATTR_DAQ_CTRL_TIE_CTRL_RANGE_TO_LIMITS:
            *(int *)value = daqCtrl->tieCtrlRangeToLimits;
            break;
        case ATTR_DAQ_CTRL_VERSION:
            *(int *)value = daqCtrl->version;
            break;
        case ATTR_DAQ_CTRL_ERROR_NOTIFICATION:
            *(int *)value = daqCtrl->errorNotification;
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

     /* Returns TRUE if found, false otherwise.
        helpFilePath must be of size MAX_PATHNAME_LEN
     */
static int  GetHelpFilePath(char helpFilePath[])
{
    char    path[MAX_PATHNAME_LEN];
    extern void CVIFUNC_C __GetResDirPathString(char *);    /* undocument function to get the CVI environments bin directory path */
    int     found = FALSE;

    if (GetProjectDir(path) == 0)
        {
        MakePathname (path, DAQ_CTRL_HELP_FILE, path);
        if (FileExists(path, 0))
            found = TRUE;
        else
            /*  If running from the environment, look for CVI's copy of the help file */
        if (!InStandaloneExecutable())
            {
            __GetResDirPathString(path);
            strcat(path, DAQ_CTRL_HELP_FILE);
            if (FileExists(path, 0))
                found = TRUE;
            }
        }

    if (helpFilePath)
        {
        if (found)
            strcpy(helpFilePath, path);
        else
            helpFilePath[0] = 0;
        }
    return found;
}

/*****************************************************************************************/

int CVIFUNC GetNumChannelsNoBOLE (int device, char *channel, int ioDirection, unsigned long *numChannels)
{
    int savedBOLE = SetBOLE(FALSE);     /* disable break on library errors */
    int error;

    error = GetNumChannels (device, channel, ioDirection, numChannels);

    SetBOLE(savedBOLE);
    return error;
}

/*****************************************************************************************/

    /*  returns TRUE if properties is selected, FALSE if not, or a negative error code */
int CVIFUNC DaqCtrl_RunRightClickPopupMenu(int panel, Point p)
{
    int         error = 0;
    static int  menuBar = 0;
    static int  menu = 0;
    static int  item  = 0;
    int         selection;

    if (menuBar <= 0)
        errChk( menuBar = NewMenuBar(0));
        
    if (menu <= 0)
        errChk( menu = NewMenu (menuBar, "RightClickPopup", -1));
        
    if (item <= 0)
        errChk( item = NewMenuItem (menuBar, menu, "__Properties", -1, 0, 0, 0));
        
    errChk( selection = RunPopupMenu (menuBar, menu, panel, p.y, p.x, p.y, p.x, 4, 4));

Error:
    if (error < 0)
        return error;
    else
        return selection != 0;
}

/*****************************************************************************************/

    /*  returns the first device, or ATTR_DAQ_CTRL_DEVICE_DEFAULT if no device was found */
int CVIFUNC FindDefaultDevice(int channelType, int *found)
{
    int device = ATTR_DAQ_CTRL_DEVICE_DEFAULT;
    int foundDevice = FALSE;
    int error;
    int index;
    int numChannels;
    int savedBole = SetBOLE(FALSE);
    
    for (index = 1; index <= MAX_DEVICE_NUMBER; index++)
        {
        error = GetNumChannels(index, "0", channelType, &numChannels);
        if (error >= 0 && numChannels == 1)
            {
            foundDevice = TRUE;
            device = index;
            break;
            }
        }
        
    SetBOLE(savedBole);
    
    if (found)
        *found = foundDevice;
        
    return device;
}

