/*============================================================================*/
/*                        L a b W i n d o w s / C V I                         */
/*----------------------------------------------------------------------------*/
/*    Copyright (c) National Instruments 1987-1996.  All Rights Reserved.     */
/*----------------------------------------------------------------------------*/
/*                                                                            */
/* Title:       datalog.c                                                     */
/* Purpose:     provides datalogging functions                                */
/*                                                                            */
/*============================================================================*/

#include <userint.h>
#include <utility.h>
#include "datalog.h"
#include "custsupp.h"

#define LOGGING_PAUSED_TAG                  "Logging Paused"
#define LOG_FILE_TAG                        "Log File"
#define LOG_EVENTS_TAG                      "Log Events"
#define LOG_FILE_FORMAT_TAG                 "Log File Format"
#define START_LOGGING_ON_LOAD_TAG           "Logging Started When Loaded"
#define LOG_CONDITIONALLY_TAG               "Log Conditionally"
#define PROMPT_ON_LOG_OVERWRITE_TAG         "Log File - Prompt Before Overwrite"
#define LOG_FILE_FIELD_WIDTH_TAG            "Log File Field Width"
#define WRITE_LOG_FILE_HEADER_TAG           "Log File - Write Header"
#define LOG_FILE_CHANNEL_NAMES_TAG          "Log File Column Names"
#define LOG_FILE_COMMENT_TAG                "Log File Comment"
#define LOG_MODE_TAG                        "Log Mode"
#define NUM_SCANS_TO_LOG_TAG                "Log File - Number Of Scans To Log Or Keep"

#define TAB_DELIMITED_TEXT_TAG_VAL          "Tab Delimited Text"
#define COMMA_DELIMITED_TEXT_TAG_VAL        "Comma Delimited Text"
#define CONSTANT_FIELD_WIDTH_TEXT_TAG_VAL   "Constant Field Width Text"
#define BINARY_TAG_VAL                      "Binary"

#define DONT_STOP_TAG_VAL                   "Log Indefinitely"
#define OVERWRITE_OLDEST_TAG_VAL            "Log Indefinitely, Overwriting Oldest"
#define STOP_AT_SCAN_LIMIT_TAG_VAL          "Stop When All Scans Are Acquired"

#define DATALOG_TEXT_EXTENSION              "txt"
#define DATALOG_BINARY_EXTENSION            "dat"

#define DATALOG_BINARY_FILE_FORMAT_VERSION  1
#define DATALOG_BINARY_FILE_SIGNATURE       "DLOG"  /* don't change this once we release */
#define DATALOG_BINARY_FILE_SIGNATURE_LEN   4       /* don't change this once we release */

#define DATALOG_LINE_BUF_SIZE                   (DATALOG_MAX_FIELD_WIDTH * 5)
#define DATALOG_LINE_BUF_MAX_CHARS_BEFORE_WRITE (DATALOG_MAX_FIELD_WIDTH * 3.5)
#define DATALOG_ITEM_BUF_SIZE                   DATALOG_MAX_FIELD_WIDTH + 1
#define DATALOG_SCIENTIFIC_NOTATION_CUTOFF_VAL  1.0e12  /* switch to scientific notation if it is needed to avoid overwriting the itemBuf */

#define ATTR_DATALOG_LOGGING_ON_DEFAULT                 FALSE
#define ATTR_DATALOG_LOGGING_PAUSED_DEFAULT             FALSE
#define ATTR_DATALOG_DATA_LOG_FILE_DEFAULT              "datalog." DATALOG_TEXT_EXTENSION
#define ATTR_DATALOG_LOG_EVENTS_DEFAULT                 TRUE
#define ATTR_DATALOG_LOG_FILE_FORMAT_DEFAULT            DATA_LOG_TAB_DELIMITED_TEXT
#define ATTR_DATALOG_START_LOGGING_ON_LOAD_DEFAULT      FALSE
#define ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE_DEFAULT    TRUE
#define ATTR_DATALOG_LOG_FIELD_WIDTH_DEFAULT            16
#define ATTR_DATALOG_WRITE_LOG_HEADER_DEFAULT           TRUE
#define ATTR_DATALOG_COLUMN_HEADINGS_DEFAULT            ""
#define ATTR_DATALOG_COMMENT_DEFAULT                    ""
#define ATTR_DATALOG_LOG_MODE_DEFAULT                   DATA_LOG_DONT_STOP
#define ATTR_DATALOG_NUM_SCANS_TO_LOG_DEFAULT           10000
#define ATTR_DATALOG_LOG_CONDITIONALLY_DEFAULT          FALSE
#define ATTR_DATALOG_NUM_CHANNELS_DEFAULT               0
#define ATTR_DATALOG_PAUSE_FOR_CONDITION_DEFAULT        FALSE

#define MAX_DATE_OR_TIME_STRING_LENGTH  40
#define MAX_DATE_AND_TIME_STRING_LENGTH (MAX_DATE_OR_TIME_STRING_LENGTH * 2 + 2)  /* +2 for ", " */

#define LOG_FILE_DATA_START_TAG         "START DATA"    /* doesn't include \n because we explicitly search for this string in the file */
#define LOG_FILE_DATA_START_TAG_LEN     10              /* must match LOG_FILE_DATA_START_TAG */

#define FIND_FIRST_SCAN_PERCENTAGE      33      /* the estimate percent of the total reordering time that finding the first scan will take */
#define REMAINING_REORDER_PERCENTAGE    (100 - FIND_FIRST_SCAN_PERCENTAGE)

typedef int (*ConversionProgressCallback)(int percentComplete, void *callbackData);

typedef struct DataLogStructTag
    {
    char            dataLogFilePath[MAX_PATHNAME_LEN];
    char            eventLogFilePath[MAX_PATHNAME_LEN];
    FILE            *dataLogFile;
    FILE            *eventLogFile;
    int             dataOffsetInFile;   /* offset to start of data in the log file, valid only when logging is active */
    int             promptOnOverwrite;
    int             fieldWidth;
    int             numChannels;
    int             active;
    int             paused;
    int             pausedForCondition;
    int             writeHeader;
    int             logEvents;
    int             logOnLoad;
    int             logConditionally;
    unsigned        numScansToLogOrKeep;
    unsigned        numScansLogged;
    unsigned        numScansLoggedSinceFileWrap;
    unsigned        numScansNotLogged;
    DataLogMode     mode;
    char            *comment;
    char            *columnHeadings;
    char            *extraHeaderString;
    DataLogFormat   fileFormat;
    int             lastUsedFileFormat;
    double          normalizedTimeOfLastScan;   /* negative if no scans yet received */
    double          normalizedTimeOfLastLoggedData;
    double          timeStampOfFirstScanWhileLogging;
    int             firstScanReceived;      /* set to TRUE after the first scan is received, even if it isn't logged because logging was paused */
    int             rewindCount;
    void            *changeNotifyCallbackData;
    DataLogNotifyFunc notifyFunc;
    } DataLogStruct;



#define DATALOG_VERSION     1.0


/* private datalog functions */
static char *DataLog_EventExtension(DataLog dataLog);
static char *DataLog_DataExtension(DataLog dataLog);
static int  DataLog_OpenFiles(DataLog dataLog);
static void DataLog_CloseFiles(DataLog dataLog);
static void DataLog_NotifyOfChange(DataLog dataLog);
static void DataLog_NotifyOfChangeFailure(DataLog dataLog);
static void DataLog_NotifyOfErrorElaboration(DataLog dataLog, char *elaboration);
static int  DataLog_SetActive(DataLog dataLog, int start);
static void DataLog_SetPause(DataLog dataLog, int paused);
static void DataLog_GetActive(DataLog dataLog, int *active);
static void DataLog_SetPauseForCondition(DataLog dataLog, int pausedForCondition);
static void DataLog_GetPauseForCondition(DataLog dataLog, int *pausedForCondition);
static void DataLog_GetPause(DataLog dataLog, int *paused);
static int  DataLog_WriteHeader(DataLog dataLog);
static int  DataLog_WriteBinaryHeader(DataLog dataLog, char dateAndTimeString[], char columnHeadings[]);
static int  DataLog_ReadBinaryHeader(FILE *file, int *isDataLogFile, int *version, int *numChannels, int *headerSize,
                    char **dateAndTime, char **headings, char **comment, int *writeHeader, int *fieldWidth);
static int  DataLog_WriteTextHeader(FILE *file, char *comment, char *dataAndTimeLabel, char *time, char *extraHeaderString,
                    int numChannels, int format, int fieldWidth, char *columnHeadings);
static int  DataLog_WriteColumnHeadings(FILE *file, int fileFormat, int fieldWidth, int numChannels, char *columnHeadings);
static void DataLog_FixFileExtensions(DataLog dataLog);
static int  DataLog_SetFile(DataLog dataLog, char *fileName);
static void DataLog_GetWriteHeader(DataLog dataLog, int *writeHeader);
static void DataLog_SetWriteHeader(DataLog dataLog, int writeHeader);
static void DataLog_GetState(DataLog dataLog, int *state);
static void DataLog_GetDataFile(DataLog dataLog, char fileName[]);
static void DataLog_GetDataFileLength(DataLog dataLog, int *length);
static int  DataLog_GetDataFileCopy(DataLog dataLog, char **fileName);
static void DataLog_GetEventFile(DataLog dataLog, char fileName[]);
static void DataLog_SetLogEvents(DataLog dataLog, int logEvents);
static void DataLog_GetLogEvents(DataLog dataLog, int *logEvents);
static int  DataLog_SetFileFormat(DataLog dataLog, int fileFormat);
static void DataLog_GetFileFormat(DataLog dataLog, int *fileFormat);
static void DataLog_SetLogOnLoad(DataLog dataLog, int logOnLoad);
static void DataLog_GetLogOnLoad(DataLog dataLog, int *logOnLoad);
static void DataLog_SetLogConditionally(DataLog dataLog, int logConditionally);
static void DataLog_GetLogConditionally(DataLog dataLog, int *logConditionally);
static void DataLog_SetOverwritePrompt(DataLog dataLog, int promptOnOverwrite);
static void DataLog_GetOverwritePrompt(DataLog dataLog, int *promptOnOverwrite);
static void DataLog_SetFieldWidth(DataLog dataLog, int fieldWidth);
static void DataLog_GetFieldWidth(DataLog dataLog, int *fieldWidth);
static int  DataLog_SetComment(DataLog dataLog, char *comment);
static void DataLog_GetComment(DataLog dataLog, char *comment);
static void DataLog_GetCommentLength(DataLog dataLog, int *length);
static int  DataLog_GetCommentCopy(DataLog dataLog, char **copy);
static int  DataLog_SetColumnHeadings(DataLog dataLog, char *columnHeadings);
static void DataLog_GetColumnHeadings(DataLog dataLog, char *columnHeadings);
static void DataLog_GetColumnHeadingsLength(DataLog dataLog, int *length);
static int  DataLog_GetColumnHeadingsCopy(DataLog dataLog, char **columnHeadingsCopy);
static int  DataLog_SetExtraHeaderInfo(DataLog dataLog, char *headerString);
static void DataLog_GetExtraHeaderInfo(DataLog dataLog, char *headerString);
static void DataLog_GetExtraHeaderInfoLength(DataLog dataLog, int *length);
static int  DataLog_GetExtraHeaderInfoCopy(DataLog dataLog, char **infoCopy);
static int  DataLog_SetMode(DataLog dataLog, int mode);
static void DataLog_GetMode(DataLog dataLog, int *mode);
static void DataLog_SetNumScansToLogOrKeep(DataLog dataLog, unsigned numScans);
static void DataLog_GetNumScansToLogOrKeep(DataLog dataLog, unsigned *numScans);
static void DataLog_GetNumScansLogged(DataLog dataLog, unsigned *numScans);
static void DataLog_Flush(DataLog dataLog);
static int  DataLog_WriteTextScan(FILE *file, double timeZero, double *time, double *data, int numRows, int numChannels, int fileFormat, int fieldWidth);
static int  DataLog_WriteBinaryScan(FILE *file, double timeZero, double *time, void *data, int numRows, int numChannels);
static void DataLog_FormatValue(double value, int fieldWidth, char buf[]);
static int  DataLog_ConvertBinaryFile(DataLog dataLog, int fileFormat, char binFilePath[], char textFilePath[], char *columnHeadings, char *comment, int writeHeader, int fieldWidth,
                        ConversionProgressCallback progressProc, void *callbackData);
static int  DataLog_ConversionProgressCallback(int percentComplete, void *callbackData);
static void DataLog_GetNumChannels(DataLog dataLog, int *numChannels);
static int  DataLog_SetNumChannels(DataLog dataLog, int numChannels);
static void DataLog_NotifyOfColumnHeaderRevision(DataLog dataLog, char **columnHeaders);
static int  DataLog_GetDefaultConvertHeadings(char **defaultHeadings, int numChannels);
static int  DataLog_DoModeSpecificPostLogAction(DataLog dataLog);
static int  DataLog_RewindDataLogFile(DataLog dataLog);
static int  DataLog_IsBinaryLogFile(char *logFilePath);
static int  DataLog_ReorderBinaryLogFile(char *binFilePath, ConversionProgressCallback progressProc, void *callbackData);

static void GetDateAndTimeString(char dateAndTimeString[MAX_DATE_AND_TIME_STRING_LENGTH]);
static int  FindEarliestScanInBinFile(FILE *file, int rowSize, int dataStartOffset, int numScansInFile, int *earliestScanOffset,
                int *earliestScanIndex, ConversionProgressCallback progressProc, void *callbackData, int maxPercent);

static int  DataLog_ReorderTextLogFile(char *logFilePath, ConversionProgressCallback progressProc, void *callbackData);
static int  ConfirmReorderingCancellation(int cancel);
static int FindEarliestScanInTextFile(FILE *file, int dataStartOffset, int *numScans, int *earliestScanOffset,
                int *earliestScanIndex, ConversionProgressCallback progressProc, void *callbackData, int maxPercent);


/************************************************************************************************/
#pragma warning(disable: 4761)
    /*  returns a ResponseCode, or negative error code */
int CVIFUNC DataLog_SelectLogFile(DataLog dataLog, int writingFileNow)
{
    int     error = 0;
    char    newPath[MAX_PATHNAME_LEN];
    int     choice;
    char    fileSpec[64];
    int     result = CANCELED;
    char    *dataFilePath = 0;

    sprintf(fileSpec, "*.%s", DataLog_DataExtension(dataLog));

    errChk( choice = FileSelectPopup ("", fileSpec, DataLog_DataExtension(dataLog),
                     "Select New Data Log File", writingFileNow ? VAL_SAVE_BUTTON : VAL_OK_BUTTON, 0, 1, 1, 1, newPath));

    if (!ObjectExists(dataLog))
        choice = VAL_NO_FILE_SELECTED;

    if (choice != VAL_NO_FILE_SELECTED)
        {
        errChk( DataLog_GetDataFileCopy(dataLog, &dataFilePath));
        if (strcmp(dataFilePath, newPath) == 0)
            result = NO_CHANGE_MADE;
        else
            {
            errChk( DataLog_SetFile(dataLog, newPath));
            result = CHANGE_MADE;
            }
        }

Error:
    free(dataFilePath);

    if (error < 0)
        return error;
    else
        return result;
}
/************************************************************************************************/

int CVIFUNC DataLog_IsAttribute(int attribute)
{
    switch (attribute)
        {
        case ATTR_DATALOG_LOGGING_ON:
        case ATTR_DATALOG_LOGGING_PAUSED:
        case ATTR_DATALOG_LOGGING_STATE:
        case ATTR_DATALOG_DATA_LOG_FILE:
        case ATTR_DATALOG_DATA_LOG_FILE_COPY:
        case ATTR_DATALOG_DATA_LOG_FILE_LENGTH:
        case ATTR_DATALOG_EVENT_LOG_FILE:
        case ATTR_DATALOG_LOG_EVENTS:
        case ATTR_DATALOG_LOG_FILE_FORMAT:
        case ATTR_DATALOG_START_LOGGING_ON_LOAD:
        case ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE:
        case ATTR_DATALOG_LOG_FIELD_WIDTH:
        case ATTR_DATALOG_WRITE_LOG_HEADER:
        case ATTR_DATALOG_COLUMN_HEADINGS:
        case ATTR_DATALOG_COLUMN_HEADINGS_COPY:
        case ATTR_DATALOG_COLUMN_HEADINGS_LENGTH:
        case ATTR_DATALOG_COMMENT:
        case ATTR_DATALOG_COMMENT_LENGTH:
        case ATTR_DATALOG_COMMENT_COPY:
        case ATTR_DATALOG_LOG_MODE:
        case ATTR_DATALOG_NUM_SCANS_TO_LOG:
        case ATTR_DATALOG_NUM_SCANS_LOGGED:
        case ATTR_DATALOG_EXTRA_HEADER_INFO:
        case ATTR_DATALOG_EXTRA_HEADER_INFO_COPY:
        case ATTR_DATALOG_EXTRA_HEADER_INFO_LENGTH:
        case ATTR_DATALOG_LOG_CONDITIONALLY:
        case ATTR_DATALOG_NUM_CHANNELS:
        case ATTR_DATALOG_PAUSE_FOR_CONDITION:
            return TRUE;
        default:
            return FALSE;
        }
}

/***********************************************************************************/

int CVIFUNC_C DataLog_SetAttribute(DataLog dataLog, int attribute, ...)
{
    int     error = 0;
    va_list parmInfo;

    va_start(parmInfo, attribute);
    errChk( DataLog_SetAttributeFromParmInfo(dataLog, attribute, parmInfo));

Error:
    va_end(parmInfo);
    return error;
}

/***********************************************************************************/

int CVIFUNC DataLog_SetAttributeFromParmInfo(DataLog dataLog, int attribute, va_list parmInfo)
{
    int error = 0;

    switch (attribute)
        {
        case ATTR_DATALOG_LOGGING_ON:
            errChk( DataLog_SetActive(dataLog, va_arg(parmInfo, int)));
            break;
        case ATTR_DATALOG_LOGGING_PAUSED:
            DataLog_SetPause(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_LOGGING_STATE:
            errChk( UIEAttributeNotSettable);
            break;
        case ATTR_DATALOG_DATA_LOG_FILE:
            errChk( DataLog_SetFile(dataLog, va_arg(parmInfo, char *)));
            break;
        case ATTR_DATALOG_EVENT_LOG_FILE:
            errChk( UIEAttributeNotSettable);
            break;
        case ATTR_DATALOG_LOG_EVENTS:
            DataLog_SetLogEvents(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_LOG_FILE_FORMAT:
            errChk( DataLog_SetFileFormat(dataLog, va_arg(parmInfo, int)));
            break;
        case ATTR_DATALOG_START_LOGGING_ON_LOAD:
            DataLog_SetLogOnLoad(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_LOG_CONDITIONALLY:
            DataLog_SetLogConditionally(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_NUM_CHANNELS:
            errChk( DataLog_SetNumChannels(dataLog, va_arg(parmInfo, int)));
            break;
        case ATTR_DATALOG_PAUSE_FOR_CONDITION:
            DataLog_SetPauseForCondition(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE:
            DataLog_SetOverwritePrompt(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_LOG_FIELD_WIDTH:
            DataLog_SetFieldWidth(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_WRITE_LOG_HEADER:
            DataLog_SetWriteHeader(dataLog, va_arg(parmInfo, int));
            break;
        case ATTR_DATALOG_COLUMN_HEADINGS:
            errChk( DataLog_SetColumnHeadings(dataLog, va_arg(parmInfo, char *)));
            break;
        case ATTR_DATALOG_COMMENT:
            errChk( DataLog_SetComment(dataLog, va_arg(parmInfo, char *)));
            break;
        case ATTR_DATALOG_LOG_MODE:
            errChk( DataLog_SetMode(dataLog, va_arg(parmInfo, int)));
            break;
        case ATTR_DATALOG_NUM_SCANS_TO_LOG:
            DataLog_SetNumScansToLogOrKeep(dataLog, va_arg(parmInfo, unsigned));
            break;
        case ATTR_DATALOG_EXTRA_HEADER_INFO:
            errChk( DataLog_SetExtraHeaderInfo(dataLog, va_arg(parmInfo, char *)));
            break;
        case ATTR_DATALOG_DATA_LOG_FILE_COPY:
        case ATTR_DATALOG_DATA_LOG_FILE_LENGTH:
        case ATTR_DATALOG_EXTRA_HEADER_INFO_COPY:
        case ATTR_DATALOG_EXTRA_HEADER_INFO_LENGTH:
        case ATTR_DATALOG_COLUMN_HEADINGS_LENGTH:
        case ATTR_DATALOG_COLUMN_HEADINGS_COPY:
        case ATTR_DATALOG_COMMENT_LENGTH:
        case ATTR_DATALOG_COMMENT_COPY:
        case ATTR_DATALOG_NUM_SCANS_LOGGED:
            errChk( UIEAttributeNotSettable);
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    if (error < 0)                          /* in case an error prevented the value of an attribute from changing or forced it to change */
        DataLog_NotifyOfChangeFailure(dataLog); /* for example, a failure starting needs to cause the to be called to pop the start button back out */

    return error;
}

/***********************************************************************************/

int CVIFUNC DataLog_GetAttribute(DataLog dataLog, int attribute, void *value)
{
    int error = 0;

    if (!value)
        errChk( UIENullPointerPassed);

    switch (attribute)
        {
        case ATTR_DATALOG_LOGGING_ON:
            DataLog_GetActive(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOGGING_PAUSED:
            DataLog_GetPause(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOGGING_STATE:
            DataLog_GetState(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_DATA_LOG_FILE:
            DataLog_GetDataFile(dataLog, (char *)value);
            break;
        case ATTR_DATALOG_DATA_LOG_FILE_COPY:
            DataLog_GetDataFileCopy(dataLog, (char **)value);
            break;
        case ATTR_DATALOG_DATA_LOG_FILE_LENGTH:
            DataLog_GetDataFileLength(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_EVENT_LOG_FILE:
            DataLog_GetEventFile(dataLog, (char *)value);
            break;
        case ATTR_DATALOG_LOG_EVENTS:
            DataLog_GetLogEvents(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOG_FILE_FORMAT:
            DataLog_GetFileFormat(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_START_LOGGING_ON_LOAD:
            DataLog_GetLogOnLoad(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOG_CONDITIONALLY:
            DataLog_GetLogConditionally(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_NUM_CHANNELS:
            DataLog_GetNumChannels(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_PAUSE_FOR_CONDITION:
            DataLog_GetPauseForCondition(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE:
            DataLog_GetOverwritePrompt(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOG_FIELD_WIDTH:
            DataLog_GetFieldWidth(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_WRITE_LOG_HEADER:
            DataLog_GetWriteHeader(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_COLUMN_HEADINGS:
            DataLog_GetColumnHeadings(dataLog, (char *)value);
            break;
        case ATTR_DATALOG_COLUMN_HEADINGS_LENGTH:
            DataLog_GetColumnHeadingsLength(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_COLUMN_HEADINGS_COPY:
            errChk( DataLog_GetColumnHeadingsCopy(dataLog, (char **)value));
            break;
        case ATTR_DATALOG_COMMENT:
            DataLog_GetComment(dataLog, (char *)value);
            break;
        case ATTR_DATALOG_COMMENT_COPY:
            errChk( DataLog_GetCommentCopy(dataLog, (char **)value));
            break;
        case ATTR_DATALOG_COMMENT_LENGTH:
            DataLog_GetCommentLength(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_LOG_MODE:
            DataLog_GetMode(dataLog, (int *)value);
            break;
        case ATTR_DATALOG_NUM_SCANS_TO_LOG:
            DataLog_GetNumScansToLogOrKeep(dataLog, (unsigned *)value);
            break;
        case ATTR_DATALOG_NUM_SCANS_LOGGED:
            DataLog_GetNumScansLogged(dataLog, (unsigned *)value);
            break;
        case ATTR_DATALOG_EXTRA_HEADER_INFO:
            DataLog_GetExtraHeaderInfo(dataLog, (char *)value);
            break;
        case ATTR_DATALOG_EXTRA_HEADER_INFO_COPY:
            errChk( DataLog_GetExtraHeaderInfoCopy(dataLog, (char **)value));
            break;
        case ATTR_DATALOG_EXTRA_HEADER_INFO_LENGTH:
            DataLog_GetExtraHeaderInfoLength(dataLog, (int *)value);
            break;
        default:
            errChk( UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/***********************************************************************************/

DataLog CVIFUNC DataLog_Create(DataLogNotifyFunc notifyFunc, void *callbackData)
{
    int     error = 0;
    DataLog newDataLog = 0;

    nullChk( newDataLog = calloc(sizeof(*newDataLog), 1));

    newDataLog->lastUsedFileFormat = -1;

    nullChk( newDataLog->extraHeaderString = StrDup(""));

    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOGGING_PAUSED,              ATTR_DATALOG_LOGGING_PAUSED_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_DATA_LOG_FILE,               ATTR_DATALOG_DATA_LOG_FILE_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOG_EVENTS,                  ATTR_DATALOG_LOG_EVENTS_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOG_FILE_FORMAT,             ATTR_DATALOG_LOG_FILE_FORMAT_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_START_LOGGING_ON_LOAD,       ATTR_DATALOG_START_LOGGING_ON_LOAD_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOG_CONDITIONALLY,           ATTR_DATALOG_LOG_CONDITIONALLY_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_NUM_CHANNELS,                ATTR_DATALOG_NUM_CHANNELS_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE,     ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOG_FIELD_WIDTH,             ATTR_DATALOG_LOG_FIELD_WIDTH_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_WRITE_LOG_HEADER,            ATTR_DATALOG_WRITE_LOG_HEADER_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_COLUMN_HEADINGS,             ATTR_DATALOG_COLUMN_HEADINGS_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_COMMENT,                     ATTR_DATALOG_COMMENT_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOG_MODE,                    ATTR_DATALOG_LOG_MODE_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_NUM_SCANS_TO_LOG,            ATTR_DATALOG_NUM_SCANS_TO_LOG_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_LOGGING_ON,                  ATTR_DATALOG_LOGGING_ON_DEFAULT));
    errChk( DataLog_SetAttribute(newDataLog,  ATTR_DATALOG_PAUSE_FOR_CONDITION,         ATTR_DATALOG_PAUSE_FOR_CONDITION_DEFAULT));

    newDataLog->notifyFunc = notifyFunc;
    newDataLog->changeNotifyCallbackData = callbackData;
    nullChk( RegisterExistence(newDataLog));
    DataLog_NotifyOfChange(newDataLog);

Error:
    if (error < 0)
        {
        DataLog_Dispose(newDataLog);
        return 0;
        }
    else
        return newDataLog;
}

/***********************************************************************************/

static void DataLog_NotifyOfColumnHeaderRevision(DataLog dataLog, char **columnHeaders)
{
    if (dataLog->notifyFunc)
        dataLog->notifyFunc(dataLog->changeNotifyCallbackData, DATALOG_REVISE_COLUMN_HEADERS, (void *)columnHeaders);
}

/***********************************************************************************/

static void DataLog_NotifyOfChange(DataLog dataLog)
{
    if (dataLog->notifyFunc)
        dataLog->notifyFunc(dataLog->changeNotifyCallbackData, DATALOG_CHANGED, 0);
}

/***********************************************************************************/

static void DataLog_NotifyOfChangeFailure(DataLog dataLog)
{
    if (dataLog->notifyFunc)
        dataLog->notifyFunc(dataLog->changeNotifyCallbackData, DATALOG_CHANGE_REJECTED, 0);
}

/***********************************************************************************/

static void DataLog_NotifyOfErrorElaboration(DataLog dataLog, char *elaboration)
{
    if (dataLog->notifyFunc)
        dataLog->notifyFunc(dataLog->changeNotifyCallbackData, DATALOG_ERROR_ELABORATION, (void *)elaboration);
}

/***********************************************************************************/

void CVIFUNC DataLog_Dispose(DataLog dataLog)
{
    if (dataLog)
        {
        DeRegisterExistence(dataLog);

        dataLog->notifyFunc = 0;            /* no notifications sent during dispose */
        DataLog_SetActive(dataLog, FALSE);  /* close log file if needed */
        free(dataLog->comment);
        free(dataLog->columnHeadings);
        free(dataLog->extraHeaderString);
        free(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_FixFileExtensions(DataLog dataLog)
{
    ChangeFileNameExtension(dataLog->dataLogFilePath, dataLog->dataLogFilePath, DataLog_DataExtension(dataLog));
    ChangeFileNameExtension(dataLog->eventLogFilePath, dataLog->eventLogFilePath, DataLog_EventExtension(dataLog));
}

/***********************************************************************************/

static void DataLog_GetDataFile(DataLog dataLog, char fileName[])
{
    strcpy(fileName, dataLog->dataLogFilePath);
}

/***********************************************************************************/

static int DataLog_GetDataFileCopy(DataLog dataLog, char **fileName)
{
    int error = 0;

    nullChk( *fileName = StrDup(dataLog->dataLogFilePath));

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_GetDataFileLength(DataLog dataLog, int *length)
{
    *length = strlen(dataLog->dataLogFilePath);
}

/***********************************************************************************/

static void DataLog_GetEventFile(DataLog dataLog, char fileName[])
{
    strcpy(fileName, dataLog->eventLogFilePath);
}

/***********************************************************************************/

static int DataLog_SetFile(DataLog dataLog, char *fileName)
{
    int     error = 0;
    char    oldDataLogPath[MAX_PATHNAME_LEN];
    char    oldEventLogPath[MAX_PATHNAME_LEN];
    char    absolutePath[MAX_PATHNAME_LEN];

    if (!fileName)
        fileName = "";

    StringCopyMax (oldDataLogPath, dataLog->dataLogFilePath, MAX_PATHNAME_LEN);
    StringCopyMax (oldEventLogPath, dataLog->eventLogFilePath, MAX_PATHNAME_LEN);

    StringCopyMax (dataLog->dataLogFilePath, fileName, MAX_PATHNAME_LEN);

        /*  if the path isn't already an absolute path, tack on the project directory to make it one */
    if (!IsAbsolutePath(dataLog->dataLogFilePath) && GetProjectDir(absolutePath) == 0)
        {
        if  (MakePathname (absolutePath, dataLog->dataLogFilePath, absolutePath) == 0)
            StringCopyMax (dataLog->dataLogFilePath, absolutePath, MAX_PATHNAME_LEN);
        }

    StringCopyMax (dataLog->eventLogFilePath, dataLog->dataLogFilePath, MAX_PATHNAME_LEN);
    DataLog_FixFileExtensions(dataLog);

    if (strcmp(oldDataLogPath, dataLog->dataLogFilePath) != 0 || strcmp(oldEventLogPath, dataLog->eventLogFilePath) != 0)
        {
        if (dataLog->active)
            {                   /* undo change */
            StringCopyMax (dataLog->dataLogFilePath, oldDataLogPath, MAX_PATHNAME_LEN);
            StringCopyMax (dataLog->eventLogFilePath, oldEventLogPath, MAX_PATHNAME_LEN);

            errChk( ToolErr_OperationInProgress);
            }

        DataLog_NotifyOfChange(dataLog);
        }

Error:
    return error;
}

/***********************************************************************************/

static int DataLog_SetNumChannels(DataLog dataLog, int numChannels)
{
    int error = 0;

    if (numChannels != dataLog->numChannels)
        {
        if (dataLog->active)
            errChk( ToolErr_OperationInProgress);

        dataLog->numChannels = numChannels;
        DataLog_NotifyOfChange(dataLog);
        }

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_GetNumChannels(DataLog dataLog, int *numChannels)
{
    *numChannels = dataLog->numChannels;
}

/***********************************************************************************/

    /*  returns TRUE if ok to write, FALSE if not ok, or a negative error code */
    /*  the user may change the datalog paths from here */
static int DataLog_ConfirmOverwrite(DataLog dataLog)
{
    int     error = 0;
    int     nonEmptyDataLogFileExists;
    int     nonEmptyEventLogFileExists;
    int     ok = FALSE;
    char    msg[MAX_PATHNAME_LEN*2+ 200];
    int     choice;
    int     size;
    char    *dlgTitle;

    nonEmptyDataLogFileExists = FileExists(dataLog->dataLogFilePath, &size) && size > 0;
    nonEmptyEventLogFileExists = FileExists(dataLog->eventLogFilePath, &size) && size > 0;

    if (nonEmptyDataLogFileExists || nonEmptyEventLogFileExists)
        {
        if (nonEmptyDataLogFileExists && nonEmptyEventLogFileExists)
            {
            sprintf(msg, "The files '%s' and\n'%s' already exist on disk.\n", dataLog->dataLogFilePath, dataLog->eventLogFilePath);
            strcat(msg, "\nDo you want to overwrite the files, or choose a new log file?");
            dlgTitle = "Warning, About to overwrite existing log files";
            }
        else
            {
            if (nonEmptyDataLogFileExists)
                sprintf(msg, "The file '%s'\nalready exists on disk.\n", dataLog->dataLogFilePath);
            else
                sprintf(msg, "The file '%s'\nalready exists on disk.\n", dataLog->eventLogFilePath);

            strcat(msg, "\nDo you want to overwrite the file, or choose a new log file?");
            dlgTitle = "Warning, About to overwrite existing log file";
            }

        errChk( choice = GenericMessagePopup (dlgTitle,
                             msg, "Overwrite", "Choose New Log File",
                             "Cancel", 0, 0, 0, VAL_GENERIC_POPUP_BTN1,
                             VAL_GENERIC_POPUP_BTN1, VAL_GENERIC_POPUP_BTN3));

        if (!ObjectExists(dataLog))
            choice = VAL_GENERIC_POPUP_BTN3;

        switch (choice)
            {
            case VAL_GENERIC_POPUP_BTN1:        /* overwrite */
                ok = TRUE;
                break;
            case VAL_GENERIC_POPUP_BTN2:        /* choose new */
                errChk( ok = DataLog_SelectLogFile(dataLog, TRUE));
                break;
            case VAL_GENERIC_POPUP_BTN3:        /* cancel */
                break;
            }
        }
    else
        ok = TRUE;

Error:
    if (error < 0)
        return error;
    else
        return ok;
}

/****************************************************************************/

static char *DataLog_DataExtension(DataLog dataLog)
{
    switch (dataLog->fileFormat)
        {
        case DATA_LOG_TAB_DELIMITED_TEXT:
        case DATA_LOG_COMMA_DELIMITED_TEXT:
        case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
            return DATALOG_TEXT_EXTENSION;

        case DATA_LOG_BINARY:
            return DATALOG_BINARY_EXTENSION;
        default:
            AssertMsg(0, "Bad File Format type detected in DataLog_DataExtension()");
            return "???";
        }
}

/****************************************************************************/

static char *DataLog_EventExtension(DataLog dataLog)
{
    UNUSED(dataLog);
    return "evt";
}

/****************************************************************************/

static void DataLog_Flush(DataLog dataLog)
{
    if (dataLog->active)
        {
        if (dataLog->dataLogFile)
            fflush(dataLog->dataLogFile);
        if (dataLog->eventLogFile)
            fflush(dataLog->eventLogFile);
        }
}

/****************************************************************************/

static int DataLog_OpenFiles(DataLog dataLog)
{
    int     error = 0;
    int     canWriteDataLog, canWriteEventLog;
    char    errorElaboration[128];

    sprintf(errorElaboration, "An error occurred opening the data log file '%s'", dataLog->dataLogFilePath);    /* this is the error message in case of an error */

    errChk( GetFileWritability(dataLog->dataLogFilePath, &canWriteDataLog));
    if (!canWriteDataLog)
        errChk( UIEErrorWritingFile);

    dataLog->dataLogFile = fopen (dataLog->dataLogFilePath, "wb+");
    if (!dataLog->dataLogFile)
        errChk( UIEUnableToOpenFile);

    sprintf(errorElaboration, "An error occurred opening the event log file '%s'", dataLog->eventLogFilePath);  /* this is the error message in case of an error */

    errChk( GetFileWritability(dataLog->eventLogFilePath, &canWriteEventLog));

    if (!canWriteEventLog)
        errChk( UIEErrorWritingFile);

    dataLog->eventLogFile = fopen (dataLog->eventLogFilePath, "wb");
    if (!dataLog->eventLogFile)
        errChk( UIEUnableToOpenFile);

Error:
    if (error < 0)
        {
        DataLog_CloseFiles(dataLog);
        DataLog_NotifyOfErrorElaboration(dataLog, errorElaboration);
        return error;
        }
    else
        return 0;
}

/****************************************************************************/

static void DataLog_CloseFiles(DataLog dataLog)
{
    if (dataLog->eventLogFile)
        fclose(dataLog->eventLogFile);

    if (dataLog->dataLogFile)
        fclose(dataLog->dataLogFile);

    dataLog->dataLogFile = 0;
    dataLog->eventLogFile = 0;
}

/***********************************************************************************/

static int  DataLog_WriteBinaryHeader(DataLog dataLog, char dateAndTimeString[], char columnHeadings[])
{
    int     error = 0;
    int     numHeaderBytes;
    int     len;
    int     numWritten;
    int     version = DATALOG_BINARY_FILE_FORMAT_VERSION;
    int     timeAndDateLength;
    int     timeAndDateSize;
    int     headingsLength;
    int     headingsSize;
    int     commentLength;
    int     commentSize;

    timeAndDateLength = strlen(dateAndTimeString);
    timeAndDateSize = timeAndDateLength + 1;    /* +1 for null terminator */

    headingsLength = strlen(columnHeadings);
    headingsSize = headingsLength + 1;  /* +1 for null terminator */

    commentLength = strlen(dataLog->comment);
    commentSize = commentLength + 1;    /* +1 for null terminator */

        /*  Write the binary log file signature */
    len = DATALOG_BINARY_FILE_SIGNATURE_LEN;
    numWritten = fwrite (DATALOG_BINARY_FILE_SIGNATURE, len, 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

    numHeaderBytes = DATALOG_NUM_BINARY_FIXED_HEADER_BYTES + timeAndDateSize + headingsSize + commentSize;

        /*  Write the size of the binary log file header */
    numWritten = fwrite (&numHeaderBytes, sizeof(int), 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the version of this file format */
    numWritten = fwrite (&version, sizeof(int), 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the number of channels */
    numWritten = fwrite (&dataLog->numChannels, sizeof(int), 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the writeHeader flag */
    numWritten = fwrite (&dataLog->writeHeader, sizeof(int), 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the field width */
    numWritten = fwrite (&dataLog->fieldWidth, sizeof(int), 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the time and date */
    numWritten = fwrite (dateAndTimeString, timeAndDateSize, 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the column headings */
    numWritten = fwrite (columnHeadings, headingsSize, 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

        /*  Write the comment */
    numWritten = fwrite (dataLog->comment, commentSize, 1, dataLog->dataLogFile);
    if (numWritten != 1)
        errChk( UIEDiskFull);

Error:
    return error;
}

/*******************************************************************************/

static int DataLog_IsBinaryLogFile(char *logFilePath)
{
    int     error = 0;
    char    signature[DATALOG_BINARY_FILE_SIGNATURE_LEN];
    FILE    *file = 0;
    int     isLogFile;
    int     numRead;

    file = fopen (logFilePath, "rb");

    if (!file)
        errChk( UIEUnableToOpenFile);

    numRead = fread (signature, DATALOG_BINARY_FILE_SIGNATURE_LEN, 1, file);
    if (numRead != 1 || ferror(file))
        errChk( UIEErrorReadingFile);

    isLogFile = memcmp(signature, DATALOG_BINARY_FILE_SIGNATURE, DATALOG_BINARY_FILE_SIGNATURE_LEN) == 0;

Error:
    if (file)
        fclose(file);

    if (error < 0)
        return error;
    else
        return isLogFile;
}

/*******************************************************************************/

    /* the file position is moved to point just beyond the header info */
static int  DataLog_ReadBinaryHeader(FILE *file, int *isDataLogFile, int *version, int *numChannels, int *headerSize, char **dateAndTime,
                char **headings, char **comment, int *writeHeader, int *fieldWidth)
{
    int     error = 0;
    int     ver, size, numChan, isLogFile;
    int     numRead;
    char    signature[DATALOG_BINARY_FILE_SIGNATURE_LEN];
    int     writeHeaderVal;
    int     fieldWidthVal;
    char    *headingsString = 0;
    char    *dateAndTimeString  = 0;
    char    *commentString = 0;

    numRead = fread (signature, DATALOG_BINARY_FILE_SIGNATURE_LEN, 1, file);
    if (numRead != 1 || ferror(file))
        errChk( UIEErrorReadingFile);

    isLogFile = memcmp(signature, DATALOG_BINARY_FILE_SIGNATURE, DATALOG_BINARY_FILE_SIGNATURE_LEN) == 0;

    if (isLogFile)
        {
        numRead = fread (&size, sizeof(int), 1, file);
        if (numRead != 1 || ferror(file))
            errChk( UIEErrorReadingFile);

        numRead = fread (&ver, sizeof(int), 1, file);
        if (numRead != 1 || ferror(file))
            errChk( UIEErrorReadingFile);

        numRead = fread (&numChan, sizeof(int), 1, file);
        if (numRead != 1 || ferror(file))
            errChk( UIEErrorReadingFile);

        numRead = fread (&writeHeaderVal, sizeof(int), 1, file);
        if (numRead != 1 || ferror(file))
            errChk( UIEErrorReadingFile);

        numRead = fread (&fieldWidthVal, sizeof(int), 1, file);
        if (numRead != 1 || ferror(file))
            errChk( UIEErrorReadingFile);

        errChk( ReadStringFromBinaryFile(file, &dateAndTimeString, -1, 0, 0));
        errChk( ReadStringFromBinaryFile(file, &headingsString, -1, 0, 0));
        errChk( ReadStringFromBinaryFile(file, &commentString, -1, 0, 0));

        if (fseek (file, size, SEEK_SET) != 0)      /* skip to the end of the header */
            errChk( UIEErrorReadingFile);

        if (version)
            *version = ver;

        if (numChannels)
            *numChannels = numChan;

        if (headerSize)
            *headerSize = size;
        }

    if (isDataLogFile)
        *isDataLogFile = isLogFile;

Error:
    if (error < 0)
        {
        free(dateAndTimeString);
        free(headingsString);
        free(commentString);
        if (dateAndTime)
            *dateAndTime = 0;
        if (headings)
            *headings = 0;
        if (comment)
            *comment = 0;

        return error;
        }
    else
        {
        if (dateAndTime)
            *dateAndTime = dateAndTimeString;
        else
            free(dateAndTimeString);

        if (headings)
            *headings = headingsString;
        else
            free(headingsString);

        if (comment)
            *comment = commentString;
        else
            free(commentString);

        if (writeHeader)
            *writeHeader = writeHeaderVal;

        if  (fieldWidth)
            *fieldWidth = fieldWidthVal;

        return 0;
        }
}

/***********************************************************************************/

static void PadField(char *field, int fieldWidth, int rightJustify)
{
    int length;
    int index;
    int padWidth;

    if (fieldWidth > 0)
        {
        length = strlen(field);
        if (length > fieldWidth)
            {
            field[fieldWidth] = 0;
            field[fieldWidth - 1] = '*';
            }
        else
        if (length < fieldWidth)
            {
            padWidth = fieldWidth - length;

            if (rightJustify)
                {
                memmove(field + padWidth, field, length + 1);   /* + 1 to move NULL byte too */
                for (; padWidth > 0; padWidth--)
                    *field++ = ' ';
                }
            else
                {
                for (index = fieldWidth - padWidth; index < fieldWidth; index++)
                    field[index] = ' ';
                field[fieldWidth] = 0;
                }
            }
        }
}

/***********************************************************************************/

static int DataLog_WriteColumnHeadings(FILE *file, int fileFormat, int fieldWidth, int numChannels, char *columnHeadings)
{
    int     error = 0;
    char    itemBuf[DATALOG_ITEM_BUF_SIZE];
    int     itemLength;
    char    *commaPtr;
    int     numColumns;

    numColumns = numChannels + 1;   /* first column is the Time column */

    while (columnHeadings && numColumns > 0)
        {
        commaPtr = strchr(columnHeadings, ','); 

        if (commaPtr)
            itemLength = commaPtr - columnHeadings;
        else
            itemLength = strlen(columnHeadings);

        StringCopyMax(itemBuf, columnHeadings, Min(itemLength + 1, DATALOG_ITEM_BUF_SIZE));
        RemoveSurroundingWhiteSpace(itemBuf);

        if (fileFormat == DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT)
            PadField(itemBuf, fieldWidth, TRUE);

        errChk( WriteStringToFile(file, itemBuf));

        if (commaPtr)       /* if not the last item */
            switch (fileFormat)
                {
                case DATA_LOG_TAB_DELIMITED_TEXT:
                    errChk( WriteStringToFile(file, "\t"));         
                    break;
                case DATA_LOG_COMMA_DELIMITED_TEXT:
                    errChk( WriteStringToFile(file, ", "));         
                    break;
                }

        if (commaPtr)
            columnHeadings = commaPtr + 1;
        else
            columnHeadings = 0;

        numColumns--;
        }

    errChk( WriteStringToFile(file, "\n"));         

Error:
    return error;
}

/***********************************************************************************/

static int  DataLog_WriteTextHeader(FILE *file, char *comment, char *dataAndTimeLabel, char *dateAndTime, char *extraHeaderString, int numChannels, int fileFormat, int fieldWidth, char *columnHeadings)
{
    int     error = 0;
    char    buf[128];

    errChk( WriteStringToFile(file, "NATIONAL INSTRUMENTS\n"));
    errChk( WriteStringToFile(file, "DAQ CHART DATALOG DATA\n"));

    sprintf(buf, "VERSION: %.2f\n", DATALOG_VERSION);
    errChk( WriteStringToFile(file, buf));

    if (comment && *comment)
        {
        errChk( WriteStringToFile(file, "\n"));
        errChk( WriteStringToFile(file, comment));
        errChk( WriteStringToFile(file, "\n\n"));
        }

    if (dateAndTime && *dateAndTime)
        {
        errChk( WriteStringToFile(file, dataAndTimeLabel));
        errChk( WriteStringToFile(file, dateAndTime));
        errChk( WriteStringToFile(file, "\n"));
        }

    if (extraHeaderString && *extraHeaderString)
        {
        errChk( WriteStringToFile(file, extraHeaderString));
        errChk( WriteStringToFile(file, "\n"));
        }

    sprintf(buf, "NUMBER OF CHANNELS: %d\n", numChannels);
    errChk( WriteStringToFile(file, buf));

    switch (fileFormat)
        {
        case DATA_LOG_TAB_DELIMITED_TEXT:
            strcpy(buf, "Tab Delimited Text\n");
            break;
        case DATA_LOG_COMMA_DELIMITED_TEXT:
            strcpy(buf, "Comma Delimited Text\n");
            break;
        case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
            strcpy(buf, "Constant Field Width Text\n");
            break;
        default:
            strcpy(buf, "???");
            break;
        }

    errChk( WriteStringToFile(file, "FILE FORMAT: "));
    errChk( WriteStringToFile(file, buf));

    errChk( WriteStringToFile(file, "\n"));

    errChk( WriteStringToFile(file, LOG_FILE_DATA_START_TAG));
    errChk( WriteStringToFile(file, "\n"));
    if (columnHeadings && *columnHeadings)
        errChk( DataLog_WriteColumnHeadings(file, fileFormat, fieldWidth, numChannels, columnHeadings));

Error:
    return error;
}

/***********************************************************************************/

static void GetDateAndTimeString(char dateAndTimeString[MAX_DATE_AND_TIME_STRING_LENGTH])
{
    char    timeStr[MAX_DATE_OR_TIME_STRING_LENGTH + 1];
    char    dateStr[MAX_DATE_OR_TIME_STRING_LENGTH + 1];

    if (!InternationalTime(timeStr, 40))
        strcpy(timeStr, "Time Not Available");

    if (!InternationalDate(dateStr, 40))
        strcpy(timeStr, "Date Not Available");

    sprintf(dateAndTimeString, "%s, %s", dateStr, timeStr);
}

/***********************************************************************************/

static int  DataLog_WriteHeader(DataLog dataLog)
{
    int     error = 0;
    char    *columnHeadings = 0;
    char    dateAndTimeStr[MAX_DATE_AND_TIME_STRING_LENGTH + 1];

    GetDateAndTimeString(dateAndTimeStr);

    errChk( DataLog_GetColumnHeadingsCopy(dataLog, &columnHeadings));
    DataLog_NotifyOfColumnHeaderRevision(dataLog, &columnHeadings);

    if (dataLog->fileFormat == DATA_LOG_BINARY)
        errChk( DataLog_WriteBinaryHeader(dataLog, dateAndTimeStr, columnHeadings));
    else
    if (dataLog->writeHeader)
        errChk( DataLog_WriteTextHeader(dataLog->dataLogFile, dataLog->comment, "DATE AND TIME: ", dateAndTimeStr, dataLog->extraHeaderString, 
                    dataLog->numChannels, dataLog->fileFormat, dataLog->fieldWidth, columnHeadings));


        /* remember where the data starts in the file so we can wrap around and start again if appropriate */
    dataLog->dataOffsetInFile = ftell(dataLog->dataLogFile);
    if (dataLog->dataOffsetInFile == -1 || errno == 100)
        {
        dataLog->dataOffsetInFile = -1;
        errChk( UIEIOError);
        }

Error:
    free(columnHeadings);
    return error;
}

/***********************************************************************************/

static int DataLog_SetActive(DataLog dataLog, int start)
{
    int error = 0;
    int okToOverwrite;
    int promptOnOverwrite;

    if (start != dataLog->active)
        {
        if (start)
            {
            DataLog_GetOverwritePrompt(dataLog, &promptOnOverwrite);

            if (promptOnOverwrite)
                errChk( okToOverwrite = DataLog_ConfirmOverwrite(dataLog));
            else
                okToOverwrite = TRUE;

            if (okToOverwrite)
                {
                dataLog->active = TRUE;
                errChk( DataLog_OpenFiles(dataLog));
                dataLog->numScansLogged = 0;
                dataLog->normalizedTimeOfLastScan = -1.0;
                dataLog->firstScanReceived = FALSE;
                dataLog->rewindCount = 0;
                dataLog->numScansLoggedSinceFileWrap = 0;
                errChk( DataLog_WriteHeader(dataLog));
                DataLog_NotifyOfChange(dataLog);
                }
            else
                DataLog_NotifyOfChangeFailure(dataLog);     /* the logging on button might have been pressed, but this is turning logging off (or preventing from going on)  */
            }
        else
            {
            DataLog_CloseFiles(dataLog);
            dataLog->active = FALSE;
            DataLog_SetPause(dataLog, FALSE);   /* so we won't confuse things by unintentionally starting out paused next time */

            DataLog_NotifyOfChange(dataLog);
            }
        }

Error:
    if (error < 0 && start)
        DataLog_SetActive(dataLog, FALSE);

    return error;
}

/***********************************************************************************/

static void DataLog_GetActive(DataLog dataLog, int *active)
{
    *active = dataLog->active;
}

/***********************************************************************************/

static void DataLog_SetPause(DataLog dataLog, int paused)
{
    if (dataLog->paused != paused)
        {
        dataLog->paused = !!paused;
        DataLog_Flush(dataLog);
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetPause(DataLog dataLog, int *paused)
{
    *paused = dataLog->paused;
}

/***********************************************************************************/

static void DataLog_SetLogEvents(DataLog dataLog, int logEvents)
{
    if (logEvents != dataLog->logEvents)
        {
        dataLog->logEvents = !!logEvents;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetLogEvents(DataLog dataLog, int *logEvents)
{
    *logEvents = dataLog->logEvents;
}

/***********************************************************************************/

static void DataLog_SetLogOnLoad(DataLog dataLog, int logOnLoad)
{
    if (logOnLoad != dataLog->logOnLoad)
        {
        dataLog->logOnLoad = !!logOnLoad;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetLogOnLoad(DataLog dataLog, int *logOnLoad)
{
    *logOnLoad = dataLog->logOnLoad;
}

/***********************************************************************************/

static void DataLog_SetLogConditionally(DataLog dataLog, int logConditionally)
{
    if (logConditionally != dataLog->logConditionally)
        {
        dataLog->logConditionally = !!logConditionally;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetLogConditionally(DataLog dataLog, int *logConditionally)
{
    *logConditionally = dataLog->logConditionally;
}

/***********************************************************************************/

static void DataLog_SetNumScansToLogOrKeep(DataLog dataLog, unsigned numScans)
{
    if (dataLog->numScansToLogOrKeep != numScans)
        {
        dataLog->numScansToLogOrKeep = (unsigned int)Pin(numScans, 1, numScans);
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetNumScansToLogOrKeep(DataLog dataLog, unsigned *numScans)
{
    *numScans = dataLog->numScansToLogOrKeep;
}

/***********************************************************************************/

static void DataLog_GetNumScansLogged(DataLog dataLog, unsigned *numScans)
{
    *numScans = dataLog->numScansLogged;
}
/***********************************************************************************/

static int DataLog_SetMode(DataLog dataLog, int mode)
{
    int error = 0;

    switch (mode)
        {
        case DATA_LOG_DONT_STOP:
        case DATA_LOG_OVERWRITE_OLDEST:
        case DATA_LOG_STOP_AT_SCAN_LIMIT:
            if (mode != dataLog->mode)
                {
                dataLog->mode = mode;
                DataLog_NotifyOfChange(dataLog);
                }
            break;
        default:
            errChk( UIEValueIsInvalidOrOutOfRange);
            break;
        }

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_GetMode(DataLog dataLog, int *mode)
{
    *mode = dataLog->mode;
}

/***********************************************************************************/

static void DataLog_SetOverwritePrompt(DataLog dataLog, int promptOnOverwrite)
{
    if (dataLog->promptOnOverwrite != promptOnOverwrite)
        {
        dataLog->promptOnOverwrite = !!promptOnOverwrite;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetOverwritePrompt(DataLog dataLog, int *promptOnOverwrite)
{
    *promptOnOverwrite = dataLog->promptOnOverwrite;
}

/***********************************************************************************/

static void DataLog_SetFieldWidth(DataLog dataLog, int fieldWidth)
{
    if (fieldWidth != dataLog->fieldWidth)
        {
        fieldWidth = (int)Pin(fieldWidth, 1, DATALOG_MAX_FIELD_WIDTH);
        dataLog->fieldWidth =  fieldWidth;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetFieldWidth(DataLog dataLog, int *fieldWidth)
{
    *fieldWidth = dataLog->fieldWidth;
}

/***********************************************************************************/

static int DataLog_SetFileFormat(DataLog dataLog, int fileFormat)
{
    int error = 0;

    switch (fileFormat)
        {
        case DATA_LOG_TAB_DELIMITED_TEXT:
        case DATA_LOG_COMMA_DELIMITED_TEXT:
        case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
        case DATA_LOG_BINARY:
            if (fileFormat != dataLog->fileFormat)
                {
                if (dataLog->active)
                    errChk( ToolErr_OperationInProgress);

                dataLog->fileFormat = fileFormat;
                DataLog_FixFileExtensions(dataLog);
                DataLog_NotifyOfChange(dataLog);
                }
            break;
        default:
            errChk( UIEValueIsInvalidOrOutOfRange);
            break;
        }

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_GetFileFormat(DataLog dataLog, int *fileFormat)
{
    *fileFormat = dataLog->fileFormat;
}

/***********************************************************************************/

static void DataLog_GetState(DataLog dataLog, int *state)
{
    if (dataLog->active)
        {
        if (dataLog->paused)
            *state = DATA_LOG_PAUSED;
        else
        if (dataLog->pausedForCondition)
            *state = DATA_LOG_WAITING_FOR_CONDITION;
        else
            *state = DATA_LOG_ACTIVE;
        }
    else
        *state = DATA_LOG_STOPPED;
}

/***********************************************************************************/

static void DataLog_SetWriteHeader(DataLog dataLog, int writeHeader)
{
    if (dataLog->writeHeader != writeHeader)
        {
        dataLog->writeHeader = !!writeHeader;
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetWriteHeader(DataLog dataLog, int *writeHeader)
{
    *writeHeader  = dataLog->writeHeader;
}

/***********************************************************************************/

static int DataLog_SetComment(DataLog dataLog, char *comment)
{
    int     error = 0;
    char    *newComment = 0;

    if (!comment)
        comment = "";

    if (!dataLog->comment || strcmp(comment, dataLog->comment) != 0)
        {
        nullChk( newComment = StrDup(comment));
        free( dataLog->comment);
        dataLog->comment = newComment;
        newComment = 0;
        DataLog_NotifyOfChange(dataLog);
        }

Error:
    free(newComment);
    return error;
}

/***********************************************************************************/

static void DataLog_GetComment(DataLog dataLog, char *comment)
{
    strcpy(comment, dataLog->comment);
}

/***********************************************************************************/

static void DataLog_GetCommentLength(DataLog dataLog, int *length)
{
    *length = strlen(dataLog->comment);
}

/***********************************************************************************/

static int  DataLog_GetCommentCopy(DataLog dataLog, char **copy)
{
    int error = 0;
    nullChk( *copy = StrDup(dataLog->comment));

Error:
    return error;
}


/***********************************************************************************/

static int DataLog_SetColumnHeadings(DataLog dataLog, char *columnHeadings)
{
    int     error = 0;
    char    *newNames = 0;

    if (!columnHeadings)
        columnHeadings = "";

    if (!dataLog->columnHeadings || strcmp(columnHeadings, dataLog->columnHeadings) != 0)
        {
        nullChk( newNames = StrDup(columnHeadings));
        free( dataLog->columnHeadings);
        dataLog->columnHeadings = newNames;
        newNames = 0;
        DataLog_NotifyOfChange(dataLog);
        }

Error:
    free(newNames);
    return error;
}

/***********************************************************************************/

static void DataLog_GetColumnHeadings(DataLog dataLog, char *columnHeadings)
{
    strcpy(columnHeadings, dataLog->columnHeadings);
}

/***********************************************************************************/

static int DataLog_GetColumnHeadingsCopy(DataLog dataLog, char **columnHeadingsCopy)
{
    int error = 0;

    nullChk( *columnHeadingsCopy = StrDup(dataLog->columnHeadings));

Error:
    return error;
}

/***********************************************************************************/

static int DataLog_SetExtraHeaderInfo(DataLog dataLog, char *extraHeaderString)
{
    int     error = 0;
    char    *newExtraHeaderInfo = 0;

    if (!extraHeaderString)
        extraHeaderString = "";

    if (!dataLog->extraHeaderString || strcmp(extraHeaderString, dataLog->extraHeaderString) != 0)
        {
        nullChk( newExtraHeaderInfo = StrDup(extraHeaderString));
        free( dataLog->extraHeaderString);
        dataLog->extraHeaderString = newExtraHeaderInfo;
        newExtraHeaderInfo = 0;
        DataLog_NotifyOfChange(dataLog);
        }

Error:
    free(newExtraHeaderInfo);
    return error;
}

/***********************************************************************************/

static void DataLog_GetExtraHeaderInfo(DataLog dataLog, char *extraHeaderString)
{
    strcpy(extraHeaderString, dataLog->extraHeaderString);
}

/***********************************************************************************/

static int DataLog_GetExtraHeaderInfoCopy(DataLog dataLog, char **infoCopy)
{
    int error = 0;

    nullChk( *infoCopy = StrDup(dataLog->extraHeaderString));

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_GetExtraHeaderInfoLength(DataLog dataLog, int *length)
{
    *length = strlen(dataLog->extraHeaderString);
}

/***********************************************************************************/

static void DataLog_GetColumnHeadingsLength(DataLog dataLog, int *length)
{
    *length = strlen(dataLog->columnHeadings);
}

/***********************************************************************************/

static void DataLog_SetPauseForCondition(DataLog dataLog, int pausedForCondition)
{
    if (dataLog->pausedForCondition != pausedForCondition)
        {
        dataLog->pausedForCondition = !!pausedForCondition;
        DataLog_Flush(dataLog);
        DataLog_NotifyOfChange(dataLog);
        }
}

/***********************************************************************************/

static void DataLog_GetPauseForCondition(DataLog dataLog, int *pausedForCondition)
{
    *pausedForCondition = dataLog->pausedForCondition;
}

/***********************************************************************************/

static int DataLog_WriteBinaryScan(FILE *file, double timeZero, double *time, void *data, int numRows, int numChannels)
{
    int     error = 0;
    double  normalizedTime;     /* make time start at zero in the log file */

    while (numRows > 0)
        {
        normalizedTime = *time - timeZero;

        if (fwrite (&normalizedTime, sizeof(double), 1, file) != 1)
            errChk( UIEErrorWritingFile);

        time = (void *)((double *)time + 1);

        if (fwrite (data, sizeof(double), numChannels, file) != (unsigned)numChannels)
            errChk( UIEErrorWritingFile);

        data = (void *)((double *)data + numChannels);
        numRows--;
        }

Error:
    return error;
}

/***********************************************************************************/

static void DataLog_FormatValue(double value, int fieldWidth, char buf[])
{
    int     useScientific;
    char    *fmtStr;

    useScientific = value > DATALOG_SCIENTIFIC_NOTATION_CUTOFF_VAL || value < -DATALOG_SCIENTIFIC_NOTATION_CUTOFF_VAL;

    if (fieldWidth > 0)
        {
        if (useScientific)
            fmtStr = "%*e";
        else
            fmtStr = "%*f";
        }
    else
        {
        if (useScientific)
            fmtStr = "%e";
        else
            fmtStr = "%f";
        }

    if (fieldWidth > 0)
        sprintf(buf, fmtStr, fieldWidth, value);
    else
        sprintf(buf, fmtStr, value);

    if (fieldWidth > 0 && strlen(buf) > (unsigned)fieldWidth)
        {
        buf[fieldWidth] = 0;
        buf[fieldWidth-1] = '*';
        }
}

/***********************************************************************************/

    /*  if time == 0, then the time is assumed to be the first element in each row of data. Also,
        each row of data will be assumed to contain numChannels + 1 values (+1 for the time).
    */
static int DataLog_WriteTextScan(FILE *file, double timeZero, double *time, double *data, int numRows, int numChannels, int fileFormat, int fieldWidth)
{
    int     error = 0;
    char    *separator;
    int     channel;
    char    lineBuf[DATALOG_LINE_BUF_SIZE];
    char    itemBuf[DATALOG_ITEM_BUF_SIZE];
    char    *lineBufPtr;
    double  normalizedTime;     /* make time start at zero in the log file */
    int     timeIsInData = !time;

    switch (fileFormat)
        {
        case DATA_LOG_TAB_DELIMITED_TEXT:
            separator = "\t";
            fieldWidth = -1;
            break;
        case DATA_LOG_COMMA_DELIMITED_TEXT:
            separator = ", ";
            fieldWidth = -1;
            break;
        case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
            separator = "";
            break;
        }

    lineBuf[0] = 0;
    lineBufPtr = lineBuf;

    while (numRows > 0)
        {
        if (timeIsInData)
            {
            time = data;
            data++;
            }

        normalizedTime = *time - timeZero;

        DataLog_FormatValue(normalizedTime, fieldWidth, itemBuf);

        lineBufPtr += sprintf(lineBufPtr, "%s%s", itemBuf, separator);
        time++;

        for (channel = 1; channel <= numChannels; channel++)
            {
            DataLog_FormatValue(*data, fieldWidth, itemBuf);

            if (channel != numChannels)
                lineBufPtr += sprintf(lineBufPtr, "%s%s", itemBuf, separator);
            else
                lineBufPtr += sprintf(lineBufPtr, "%s", itemBuf);

            data++;

            if (lineBufPtr -  lineBuf > DATALOG_LINE_BUF_MAX_CHARS_BEFORE_WRITE)    /* line got too long, flush it */
                {
                errChk( WriteStringToFile(file, lineBuf));
                lineBuf[0] = 0;
                lineBufPtr = lineBuf;
                }
            }

        errChk( WriteStringToFile(file, lineBuf));
        errChk( WriteStringToFile(file, "\n"));             
        lineBuf[0] = 0;
        lineBufPtr = lineBuf;
        numRows--;
        }

Error:
    return error;
}

/***********************************************************************************/

static int DataLog_RewindDataLogFile(DataLog dataLog)
{
    int error = 0;
    int currPos;
    int endPos;
    int numBytesToClear;
    int byteVal;

    if (dataLog->active && dataLog->dataLogFile)
        {
        fflush(dataLog->dataLogFile);
        currPos = ftell (dataLog->dataLogFile);
        if (currPos == -1 || errno == 100)
            errChk( UIEIOError);

        errChk(BinaryFileSize(dataLog->dataLogFile, &endPos));

        numBytesToClear = endPos - currPos;

        if (dataLog->fileFormat == DATA_LOG_BINARY)
            byteVal = 0;
        else
            byteVal = ' ';

        errChk( WriteByteToFile(dataLog->dataLogFile, byteVal, numBytesToClear));

        if (fseek (dataLog->dataLogFile , dataLog->dataOffsetInFile, SEEK_SET))
            {
            DataLog_SetActive(dataLog, FALSE);
            errChk( UIEIOError);
            }
        else
            dataLog->rewindCount++;
        }

Error:
    return error;
}

/***********************************************************************************/

static int DataLog_DoModeSpecificPostLogAction(DataLog dataLog)
{
    int error = 0;

    switch (dataLog->mode)
        {
        case DATA_LOG_DONT_STOP:
            break;
        case DATA_LOG_OVERWRITE_OLDEST:
            if  (dataLog->numScansLoggedSinceFileWrap >= dataLog->numScansToLogOrKeep)
                {
                errChk( DataLog_RewindDataLogFile(dataLog));
                dataLog->numScansLoggedSinceFileWrap = 0;
                }
            break;
        case DATA_LOG_STOP_AT_SCAN_LIMIT:
            if (dataLog->numScansLogged >= dataLog->numScansToLogOrKeep)
                errChk( DataLog_SetActive(dataLog, FALSE));
            break;
        }
Error:
    return error;
}

/***********************************************************************************/

    /*  This function places a blank line between the new scan and the oldest scan so that
        the discontinuity is noticably.  This also prevents a leaving partially overwritten scan
        after the new scan.
    */
static int DataLog_HandleTextFileWrap(DataLog dataLog)
{
    int     error = 0;
    long    currPos;
    int     offset;
    int     numSpacesToWrite;

    if (dataLog->rewindCount > 0)
        switch (dataLog->fileFormat)
            {
            case DATA_LOG_TAB_DELIMITED_TEXT:
            case DATA_LOG_COMMA_DELIMITED_TEXT:
            case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
                fflush(dataLog->dataLogFile);   /* must flush before reading */
                currPos = ftell (dataLog->dataLogFile);
                if (currPos == -1 || errno == 100)
                    errChk( UIEIOError);

                errChk( FindBytesInFile(dataLog->dataLogFile, LINE_TERMINATOR, LINE_TERMINATOR_LEN, &offset, 0));
                numSpacesToWrite = offset;
                if (fseek (dataLog->dataLogFile, currPos, SEEK_SET))
                    errChk( UIEIOError);
                errChk( WriteByteToFile(dataLog->dataLogFile, ' ', numSpacesToWrite));  
                if (fseek (dataLog->dataLogFile, currPos, SEEK_SET))
                    errChk( UIEIOError);
                break;
            }

Error:
    return error;
}

/***********************************************************************************/

int CVIFUNC DataLog_LogData(DataLog dataLog, double *time, void *data, int numRows)
{
    int error = 0;

    if (!dataLog->firstScanReceived)
        {
        dataLog->firstScanReceived = TRUE;
        dataLog->timeStampOfFirstScanWhileLogging = *time;
        }

    if (dataLog->active && !dataLog->paused && !dataLog->pausedForCondition)
        {
        if (dataLog->fileFormat == DATA_LOG_BINARY)
            errChk( DataLog_WriteBinaryScan(dataLog->dataLogFile, dataLog->timeStampOfFirstScanWhileLogging, time, data, numRows, dataLog->numChannels));
        else
            {
            errChk( DataLog_WriteTextScan(dataLog->dataLogFile, dataLog->timeStampOfFirstScanWhileLogging, time, (double *)data, numRows, dataLog->numChannels,
                        dataLog->fileFormat, dataLog->fieldWidth));
            errChk( DataLog_HandleTextFileWrap(dataLog));
            }

        dataLog->normalizedTimeOfLastLoggedData = time[numRows - 1] - dataLog->timeStampOfFirstScanWhileLogging;
        dataLog->numScansLogged++;
        dataLog->numScansLoggedSinceFileWrap++;

        errChk( DataLog_DoModeSpecificPostLogAction(dataLog));
        }
    else
        dataLog->numScansNotLogged++;

    dataLog->normalizedTimeOfLastScan = time[numRows - 1] - dataLog->timeStampOfFirstScanWhileLogging;

Error:
    return error;
}

/***********************************************************************************/

    /*  If the message is not associated with a scan, the time can be NULL */
int CVIFUNC DataLog_LogMessage(DataLog dataLog, char message[], double *timeOfScan)
{
    int     error = 0;
    char    info[200];
    char    scanTime[30];
    char    currentTime[30];

    if (dataLog->active && dataLog->logEvents)
        {
        InternationalTime(currentTime, 30);
        if (!timeOfScan)
            {
            if (dataLog->normalizedTimeOfLastScan < 0.0)
                sprintf(scanTime, "N/A");
            else
                DataLog_FormatValue(dataLog->normalizedTimeOfLastScan, -1, scanTime);

            sprintf(info, "Current time: %s.\nTime (in seconds) of last scan since data logging was initiated: %s.\nNumber of scans logged so far: %d.\n\n", 
                currentTime, scanTime, dataLog->numScansLogged);
            }
        else
            {
            DataLog_FormatValue(*timeOfScan - dataLog->timeStampOfFirstScanWhileLogging, -1, scanTime);
            sprintf(info, "Current time: %s.\nTime (in seconds) of scan since logging was initiated: %s.\nNumber of scans logged so far: %d.\n\n", 
                currentTime, scanTime, dataLog->numScansLogged);
            }

        errChk( WriteStringToFile (dataLog->eventLogFile, message));
        errChk( WriteStringToFile (dataLog->eventLogFile, "\n"));
        errChk( WriteStringToFile (dataLog->eventLogFile, info));
        }

Error:
    return error;
}

/***********************************************************************************/

    /*  returns 0 (success), negative error code or positive warning code */
int CVIFUNC ReadInDataLogAttributes(DataLog dataLog, IniText iniText, char *section, char *errorDescription, int *doingGets)
{
    int         error = 0;
    int         loggingPaused;
    char        *fileName = 0;
    char        *fileFormatString = 0;
    int         fileFormat;
    int         logEvents;
    int         logOnLoad;
    int         logConditionally;
    int         promptOnOverwrite;
    unsigned    fieldWidth;
    int         writeHeader;
    char        *columnHeadings = 0;
    char        *comment = 0;
    char        *logModeString = 0;
    int         logMode;
    unsigned    numScansToLog;

    int         foundLoggingPaused;
    int         foundFileName;
    int         foundLogEvents;
    int         foundFileFormat;
    int         foundLogOnLoad;
    int         foundPromptOnOverwrite;
    int         foundFieldWidth;
    int         foundWriteHeader;
    int         foundColumnHeadings;
    int         foundComment;
    int         foundLogMode;
    int         foundNumScansToLog;
    int         foundLogConditionally;

        /* some attributes can't be set while logging, logging must be terminated to recall previous settings */
    errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOGGING_ON, FALSE));

    *doingGets = TRUE;
    errChk( foundLoggingPaused = Ini_GetBoolean (iniText, section, LOGGING_PAUSED_TAG, &loggingPaused));
    errChk( foundFileName = Ini_GetStringCopy (iniText, section, LOG_FILE_TAG, &fileName));
    errChk( foundLogEvents = Ini_GetBoolean (iniText, section, LOG_EVENTS_TAG, &logEvents));
    errChk( foundFileFormat = Ini_GetStringCopy (iniText, section, LOG_FILE_FORMAT_TAG, &fileFormatString));
    if (foundFileFormat)
        {
        if (StrICmp(fileFormatString, TAB_DELIMITED_TEXT_TAG_VAL) == 0)
            fileFormat = DATA_LOG_TAB_DELIMITED_TEXT;
        else
        if (StrICmp(fileFormatString, COMMA_DELIMITED_TEXT_TAG_VAL) == 0)
            fileFormat = DATA_LOG_COMMA_DELIMITED_TEXT;
        else
        if (StrICmp(fileFormatString, CONSTANT_FIELD_WIDTH_TEXT_TAG_VAL) == 0)
            fileFormat = DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT;
        else
        if (StrICmp(fileFormatString, BINARY_TAG_VAL) == 0)
            fileFormat = DATA_LOG_BINARY;
        else
            errChk( ToolErr_UnRecognizedValue);
        }

    errChk( foundLogOnLoad = Ini_GetBoolean (iniText, section, START_LOGGING_ON_LOAD_TAG, &logOnLoad));
    errChk( foundLogConditionally = Ini_GetBoolean (iniText, section, LOG_CONDITIONALLY_TAG, &logConditionally));
    errChk( foundPromptOnOverwrite = Ini_GetBoolean (iniText, section, PROMPT_ON_LOG_OVERWRITE_TAG, &promptOnOverwrite));
    errChk( foundFieldWidth = Ini_GetUInt (iniText, section, LOG_FILE_FIELD_WIDTH_TAG, &fieldWidth));
    errChk( foundWriteHeader = Ini_GetBoolean (iniText, section, WRITE_LOG_FILE_HEADER_TAG, &writeHeader));
    errChk( foundColumnHeadings = Ini_GetStringCopy (iniText, section, LOG_FILE_CHANNEL_NAMES_TAG, &columnHeadings));
    errChk( foundComment = Ini_GetStringCopy (iniText, section, LOG_FILE_COMMENT_TAG, &comment));
    errChk( foundLogMode = Ini_GetStringCopy (iniText, section, LOG_MODE_TAG, &logModeString));
    if (foundLogMode)
        {
        if (StrICmp(logModeString, DONT_STOP_TAG_VAL) == 0)
            logMode = DATA_LOG_DONT_STOP;
        else
        if (StrICmp(logModeString, OVERWRITE_OLDEST_TAG_VAL) == 0)
            logMode = DATA_LOG_OVERWRITE_OLDEST;
        else
        if (StrICmp(logModeString, STOP_AT_SCAN_LIMIT_TAG_VAL) == 0)
            logMode = DATA_LOG_STOP_AT_SCAN_LIMIT;
        else
            errChk( ToolErr_UnRecognizedValue);
        }

    errChk( foundNumScansToLog = Ini_GetUInt (iniText, section, NUM_SCANS_TO_LOG_TAG, &numScansToLog));
    *doingGets = FALSE;

    if (foundLoggingPaused)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOGGING_PAUSED, loggingPaused));

    if (foundFileName)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_DATA_LOG_FILE, fileName));

    if (foundLogEvents)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOG_EVENTS, logEvents));

    if (foundFileFormat)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOG_FILE_FORMAT, fileFormat));

    if (foundLogOnLoad)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_START_LOGGING_ON_LOAD, logOnLoad));

    if (foundLogConditionally)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOG_CONDITIONALLY, logConditionally));

    if (foundPromptOnOverwrite)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_PROMPT_ON_LOG_OVERWRITE, promptOnOverwrite));

    if (foundFieldWidth)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOG_FIELD_WIDTH, fieldWidth));

    if (foundWriteHeader)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_WRITE_LOG_HEADER, writeHeader));

    if (foundColumnHeadings)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_COLUMN_HEADINGS, columnHeadings));

    if (foundComment)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_COMMENT, comment));

    if (foundLogMode)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_LOG_MODE, logMode));

    if (foundNumScansToLog)
        errChk( DataLog_SetAttribute(dataLog, ATTR_DATALOG_NUM_SCANS_TO_LOG, numScansToLog));

    error = 0;  /* make sure previous greater-than-zero result is not interpreted as a warning */

Error:
    free(fileName);
    free(fileFormatString);
    free(columnHeadings);
    free(comment);
    free(logModeString);

    if (error < 0)
        {
        if (doingGets && errorDescription && !*errorDescription)
            {
#if 0   /* restore this code if we decide to add 'if (!found) errChk(ToolErr_MissingItem);' after each get */
            if (error == ToolErr_MissingItem && !found)
                sprintf(errorDescription, "Could not find item '%s'", itemName);
            else
#endif
                sprintf(errorDescription, "Error reading from line %d", Ini_LineOfLastAccess (iniText));    
            }
        }

    return error;
}

/****************************************************************************************/

int CVIFUNC WriteOutDataLogAttributes(DataLog dataLog, IniText iniText, char *section)
{
    int     error = 0;
    char    *valueString;

    errChk( Ini_PutBoolean (iniText, section, LOGGING_PAUSED_TAG, dataLog->paused));
    errChk( Ini_PutString (iniText, section, LOG_FILE_TAG, dataLog->dataLogFilePath));
    errChk( Ini_PutBoolean (iniText, section, LOG_EVENTS_TAG, dataLog->logEvents));

    switch (dataLog->fileFormat)
        {
        default:
        case DATA_LOG_TAB_DELIMITED_TEXT:
            valueString = TAB_DELIMITED_TEXT_TAG_VAL;
            break;
        case DATA_LOG_COMMA_DELIMITED_TEXT:
            valueString = COMMA_DELIMITED_TEXT_TAG_VAL;
            break;
        case DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT:
            valueString = CONSTANT_FIELD_WIDTH_TEXT_TAG_VAL;
            break;
        case DATA_LOG_BINARY:
            valueString = BINARY_TAG_VAL;
        }
    errChk( Ini_PutString (iniText, section, LOG_FILE_FORMAT_TAG, valueString));

    errChk( Ini_PutBoolean (iniText, section, START_LOGGING_ON_LOAD_TAG, dataLog->logOnLoad));
    errChk( Ini_PutBoolean (iniText, section, LOG_CONDITIONALLY_TAG, dataLog->logConditionally));
    errChk( Ini_PutBoolean (iniText, section, PROMPT_ON_LOG_OVERWRITE_TAG, dataLog->promptOnOverwrite));
    errChk( Ini_PutUInt (iniText, section, LOG_FILE_FIELD_WIDTH_TAG, (unsigned)dataLog->fieldWidth));
    errChk( Ini_PutBoolean (iniText, section, WRITE_LOG_FILE_HEADER_TAG, dataLog->writeHeader));
    errChk( Ini_PutString (iniText, section, LOG_FILE_CHANNEL_NAMES_TAG, dataLog->columnHeadings));
    errChk( Ini_PutString (iniText, section, LOG_FILE_COMMENT_TAG, dataLog->comment));

    switch (dataLog->mode)
        {
        default:
        case DATA_LOG_DONT_STOP:
            valueString = DONT_STOP_TAG_VAL;
            break;
        case DATA_LOG_OVERWRITE_OLDEST:
            valueString = OVERWRITE_OLDEST_TAG_VAL;
            break;
        case DATA_LOG_STOP_AT_SCAN_LIMIT:
            valueString = STOP_AT_SCAN_LIMIT_TAG_VAL;
            break;
        }
    errChk( Ini_PutString (iniText, section, LOG_MODE_TAG, valueString));

    errChk( Ini_PutUInt (iniText, section, NUM_SCANS_TO_LOG_TAG, (unsigned)dataLog->numScansToLogOrKeep));

Error:
    return error;
}

/****************************************************************************************/

int CVIFUNC DataLog_TextFormatDlg(DataLog dataLog)
{
    int     error = 0;
    int     dlg = 0;
    int     ok, cancel, fieldWidthNumeric, writeHeaderCheckBox;
    int     columnHeadingsString, commentTextBox;
    int     changed = FALSE;
    int     hMargin = DLG_MARGIN;
    int     vMargin = DLG_MARGIN;
    int     top;
    int     done = FALSE;
    int     evtPanel, evtCtrl;
    int     btnWidth;
    int     canceled = FALSE;
    int     dialogWidth;
    int     contentWidth;
    int     writeHeader;
    int     columnHeadingsLength;
    char    *columnHeadings = 0;
    int     commentLength;
    char    *comment = 0;
    int     ctrlHeight;
    int     fieldWidth;
    int     stringBoundsWidth;

    errChk(dlg = NewPanel(0, "File Layout Options", 0,0, 10,10));
    SetPanelAttribute(dlg, ATTR_CONFORM_TO_SYSTEM, TRUE);

    errChk( fieldWidthNumeric =     NewCtrl (dlg, CTRL_NUMERIC, "Field Width:", 0, hMargin));
    errChk( writeHeaderCheckBox =   NewCtrl (dlg, CTRL_CHECK_BOX, "Write Header Information:", 0, hMargin));
    errChk( columnHeadingsString =  NewCtrl (dlg, CTRL_STRING, "Column Headings:\n(separate names with commas, leave blank to use defaults)\nExample: Time Stamp, Channel 0 (Pressure), Channel 1 (Temp)", 0, hMargin));
    errChk( commentTextBox =        NewCtrl (dlg, CTRL_TEXT_BOX, "Comment (Ctrl-Enter starts a new line):", 0, hMargin));
    errChk( ok =                    NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "__OK", 0, hMargin));
    errChk( cancel =                NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 0, hMargin));

    errChk( SetCtrlAttribute (dlg, fieldWidthNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_DATA_TYPE, VAL_INTEGER));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_MIN_VALUE, DATALOG_MIN_USER_SETTABLE_FIELD_WIDTH));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_MAX_VALUE, DATALOG_MAX_FIELD_WIDTH));

    errChk( GetCtrlBoundRectAttribute(dlg, columnHeadingsString, ATTR_TB_BOUNDS_WIDTH, &stringBoundsWidth));    /* widest control */
    dialogWidth = Max(400, stringBoundsWidth + 2 * hMargin);
    contentWidth = dialogWidth - 2 * hMargin;

    errChk( SetCtrlAttribute(dlg, commentTextBox, ATTR_VISIBLE_LINES, 5));

    top = vMargin;

    errChk( PutLabelsOnLeft (dlg, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0, fieldWidthNumeric, 0));
    errChk( SetCtrlBoundRectAttribute(dlg, fieldWidthNumeric, ATTR_LEFT, hMargin));

    errChk( SetAttributeForCtrls (dlg, ATTR_WIDTH, contentWidth, 0, columnHeadingsString, commentTextBox, 0));
    errChk( SetAttributeForCtrls (dlg, ATTR_LABEL_LEFT, hMargin, 0, columnHeadingsString, commentTextBox, 0));

    errChk( DistributeCtrls (dlg, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP, 0,
        fieldWidthNumeric, writeHeaderCheckBox, columnHeadingsString, commentTextBox, 0));

        /* make ok and cancel the same width */
    errChk( SetCtrlsToAttributeExtreme(dlg, ATTR_WIDTH, VAL_TB_MAX, 0, ok, cancel, 0));

        /* position ok and cancel vertically */
    top += DLG_CTRL_V_GAP * 2;
    errChk( SetAttributeForCtrls (dlg, ATTR_TOP, top, 0, ok, cancel, 0));

        /* position the ok and cancel horizontally */
    GetCtrlAttribute(dlg, ok, ATTR_WIDTH, &btnWidth);
    SetCtrlAttribute(dlg, cancel, ATTR_LEFT, hMargin + contentWidth - btnWidth);
    SetCtrlAttribute(dlg, ok, ATTR_LEFT, hMargin + contentWidth - 2 * btnWidth - DLG_BTN_H_GAP);

    errChk( GetCtrlAttribute(dlg, ok, ATTR_HEIGHT, &ctrlHeight));
    top += ctrlHeight + DLG_CTRL_V_GAP;

    errChk( SetPanelAttribute(dlg, ATTR_WIDTH, dialogWidth));
    errChk( SetPanelAttribute(dlg, ATTR_HEIGHT, top));

    errChk( SetCommonDialogShortcutKeys(dlg, ok, cancel, cancel));
    errChk( SetPanelPos(dlg, VAL_AUTO_CENTER, VAL_AUTO_CENTER));

    errChk( DataLog_GetAttribute(dataLog, ATTR_DATALOG_LOG_FIELD_WIDTH, &fieldWidth));
    errChk( SetCtrlVal(dlg, fieldWidthNumeric, fieldWidth));

    errChk( DataLog_GetAttribute(dataLog, ATTR_DATALOG_WRITE_LOG_HEADER, &writeHeader));
    errChk( SetCtrlVal(dlg, writeHeaderCheckBox, writeHeader));
    SetCtrlAttribute(dlg, commentTextBox, ATTR_DIMMED, !writeHeader);
    SetCtrlAttribute(dlg, columnHeadingsString, ATTR_DIMMED, !writeHeader);

    errChk( DataLog_GetColumnHeadingsCopy(dataLog, &columnHeadings));
    errChk( SetCtrlVal(dlg, columnHeadingsString, columnHeadings));

    errChk( DataLog_GetCommentCopy(dataLog, &comment));
    errChk( SetCtrlVal(dlg, commentTextBox, comment));

    InstallPopup(dlg);

    while (!done)
        {
        GetUserEvent(0, &evtPanel, &evtCtrl);

        if (!ObjectExists(dataLog))
            {
            done = TRUE;
            cancel = TRUE;
            }
        else
        if (evtPanel == dlg)
            {
            if  (evtCtrl == cancel)
                {
                done = TRUE;
                canceled = TRUE;
                }
            else
            if (evtCtrl == ok)
                {
                done = TRUE;
                }
            else
            if (evtCtrl == writeHeaderCheckBox)
                {
                errChk( GetCtrlVal(dlg, writeHeaderCheckBox, &writeHeader));
                SetCtrlAttribute(dlg, commentTextBox, ATTR_DIMMED, !writeHeader);
                SetCtrlAttribute(dlg, columnHeadingsString, ATTR_DIMMED, !writeHeader);
                changed = TRUE;
                }
            else
            if (evtCtrl == fieldWidthNumeric || evtCtrl == columnHeadingsString || evtCtrl == commentTextBox)
                {
                changed = TRUE;
                }
            }
        }

    if (!canceled && changed)
        {
        errChk( GetCtrlVal(dlg, fieldWidthNumeric, &fieldWidth));
        DataLog_SetFieldWidth(dataLog, fieldWidth);

        errChk( GetCtrlVal(dlg, writeHeaderCheckBox, &writeHeader));
        DataLog_SetWriteHeader(dataLog, writeHeader);

        errChk( GetCtrlAttribute(dlg, columnHeadingsString, ATTR_STRING_TEXT_LENGTH, &columnHeadingsLength));
        free(columnHeadings);
        nullChk( columnHeadings = malloc(columnHeadingsLength + 1));
        errChk( GetCtrlVal(dlg, columnHeadingsString, columnHeadings));
        errChk( DataLog_SetColumnHeadings(dataLog, columnHeadings));

        errChk( GetCtrlAttribute(dlg, commentTextBox, ATTR_STRING_TEXT_LENGTH, &commentLength));
        free(comment);
        nullChk( comment = malloc(commentLength + 1));
        errChk( GetCtrlVal(dlg, commentTextBox, comment));
        errChk( DataLog_SetComment(dataLog, comment));
        }

Error:
    free(comment);
    free(columnHeadings);

    if (dlg > 0)
        DiscardPanel(dlg);
    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (changed)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/*******************************************************************************/

int CVIFUNC DataLog_ConvertBinaryFileDlg(DataLog dataLog, int *fileWasConverted)
{
    int     error = 0;
    int     dlg = 0;
    int     binFileString, textFileString;
    int     ok, cancel, fieldWidthNumeric, writeHeaderCheckBox;
    int     columnHeadingsString, commentTextBox;
    int     textFormatMsg, textFormatDeco;
    int     fileFormatRing;
    int     changed = FALSE;
    int     hMargin = DLG_MARGIN;
    int     vMargin = DLG_MARGIN;
    int     top;
    int     done = FALSE;
    int     evtPanel, evtCtrl;
    int     btnWidth;
    int     canceled = FALSE;
    int     minDialogWidth = 350;
    int     dialogWidth;
    int     contentWidth;
    int     frameContentWidth;
    int     writeHeader;
    int     columnHeadingsLength;
    char    *columnHeadings = 0;
    int     commentLength;
    char    *comment = 0;
    int     ctrlHeight;
    int     fieldWidth;
    int     frameContentLeft;
    int     frameContentRight;
    int     contentLeft;
    int     contentRight;
    int     ctrlWidth;
    int     largestLabelWidth;
    int     minContentWidth;
    int     minFrameContentWidth;
    int     fileFormat;
    char    binFilePath[MAX_PATHNAME_LEN];
    char    textFilePath[MAX_PATHNAME_LEN];
    int     choice;
    char    fileSpec[64];
    int     progressDlg = 0;
    int     promptOnOverwrite;
    FILE    *binFile = 0;
    int     isDataLogFile;

    if  (fileWasConverted)
        *fileWasConverted = FALSE;

    contentLeft = hMargin;
    frameContentLeft = 2 * hMargin;

    sprintf(fileSpec, "*.%s", DATALOG_BINARY_EXTENSION);

    errChk( choice = FileSelectPopup ("", fileSpec, DATALOG_BINARY_EXTENSION,
            "Select Binary Data Log File To Convert", VAL_SELECT_BUTTON, 0, 1, 1, 0, binFilePath));

    if (!ObjectExists(dataLog)) /* our datalog was disposed out from under us */
        {
        canceled = TRUE;
        goto Error;
        }

    StringCopyMax (textFilePath, binFilePath, MAX_PATHNAME_LEN);
    ChangeFileNameExtension(textFilePath, textFilePath, DATALOG_TEXT_EXTENSION);

    if  (choice != VAL_EXISTING_FILE_SELECTED)
        goto Error;

    DataLog_GetOverwritePrompt(dataLog, &promptOnOverwrite);
    if (promptOnOverwrite && FileExists(textFilePath, 0))
        {
        char    msg[200];
        int     confirmed;

        sprintf(msg, "The file %s already exists on the disk.\n\nDo you want to overwrite it?", textFilePath);
        errChk( confirmed = ConfirmPopup("Warning", msg));
        if (!confirmed || !ObjectExists(dataLog))
            goto Error;
        }

    errChk(dlg = NewPanel(0, "Convert Binary Log File To Text", 0,0, 10,10));
    SetPanelAttribute(dlg, ATTR_CONFORM_TO_SYSTEM, TRUE);

    errChk( binFileString    =      NewCtrl (dlg, CTRL_STRING, "Binary File To Convert:", 0, 0));
    errChk( textFileString   =      NewCtrl (dlg, CTRL_STRING, "Text File To Create:", 0, 0));
    errChk( textFormatMsg =         NewCtrl (dlg, CTRL_TEXT_MSG, "Text Format Options", 0, 0));
    errChk( textFormatDeco =        NewCtrl (dlg, CTRL_RAISED_FRAME, "", 0, 0));

    errChk( fileFormatRing =        NewCtrl (dlg, CTRL_RING, "Text Format:", 0, 0));
    errChk( fieldWidthNumeric =     NewCtrl (dlg, CTRL_NUMERIC, "Field Width:", 0, 0));
    errChk( writeHeaderCheckBox =   NewCtrl (dlg, CTRL_CHECK_BOX, "Write Header Information:", 0, 0));
    errChk( columnHeadingsString =  NewCtrl (dlg, CTRL_STRING, "Column Headings (separate with commas)\n(leave blank to use defaults):", 0, 0));
    errChk( commentTextBox =        NewCtrl (dlg, CTRL_TEXT_BOX, "Comment (Ctrl-Enter starts a new line):", 0, 0));

    errChk( ok =                    NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "Con__vert", 0, hMargin));
    errChk( cancel =                NewCtrl (dlg, CTRL_SQUARE_COMMAND_BUTTON, "  __Cancel  ", 0, hMargin));

    errChk( InsertListItem(dlg, fileFormatRing, -1, "Tab Delimited Text", DATA_LOG_TAB_DELIMITED_TEXT));
    errChk( InsertListItem(dlg, fileFormatRing, -1, "Comma Delimited Text", DATA_LOG_COMMA_DELIMITED_TEXT));
    errChk( InsertListItem(dlg, fileFormatRing, -1, "Constant Field Width Text", DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT));
    errChk( SizeRingCtrlToText(dlg, fileFormatRing));

    errChk( SetCtrlAttribute (dlg, fieldWidthNumeric, ATTR_CHECK_RANGE, VAL_COERCE));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_DATA_TYPE, VAL_INTEGER));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_MIN_VALUE, DATALOG_MIN_USER_SETTABLE_FIELD_WIDTH));
    errChk( SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_MAX_VALUE, DATALOG_MAX_FIELD_WIDTH));

    errChk( SetCtrlAttribute(dlg, commentTextBox, ATTR_VISIBLE_LINES, 5));

    top = vMargin;

    errChk( PutLabelsOnLeft (dlg, DLG_LABEL_TO_CTRL_MIN_H_GAP, 0, binFileString, textFileString,
        fieldWidthNumeric, fileFormatRing, 0));


    errChk( SetAttributeForCtrls(dlg, ATTR_LEFT, frameContentLeft, 0, writeHeaderCheckBox, columnHeadingsString, commentTextBox, 0));
    errChk( SetAttributeForCtrls(dlg, ATTR_LABEL_LEFT, frameContentLeft, 0, columnHeadingsString, commentTextBox, 0));

    errChk( AttributeMaxAndMin (dlg, ATTR_TB_BOUNDS_WIDTH, 0, &minFrameContentWidth, 0, 0, 0,
        fieldWidthNumeric, writeHeaderCheckBox, columnHeadingsString, commentTextBox, 0));

    errChk( AttributeMaxAndMin (dlg, ATTR_TB_BOUNDS_WIDTH, 0, &minContentWidth, 0, 0, 0,
        binFileString, textFileString, fileFormatRing, 0));

    frameContentWidth = Max(minContentWidth - 2 * hMargin, minFrameContentWidth);
    frameContentWidth = Max(frameContentWidth, minDialogWidth - 4 * hMargin);
    frameContentRight = frameContentLeft + frameContentWidth;
    contentWidth = frameContentWidth + 2 * hMargin;
    contentRight = contentLeft + contentWidth;
    dialogWidth = contentWidth + 2 * hMargin;

    SetCtrlRightPosition(dlg, fieldWidthNumeric, frameContentRight);

    errChk( SetAttributeForCtrls(dlg, ATTR_LABEL_LEFT, frameContentLeft, 0, fieldWidthNumeric, 0));

    errChk( SetAttributeForCtrls (dlg, ATTR_WIDTH, frameContentWidth, 0, columnHeadingsString, commentTextBox, 0));

    errChk( DistributeCtrls (dlg, VAL_TB_VERTICAL_DISTRIBUTION, VAL_TB_FIXED_GAP_SPACING, &top, DLG_CTRL_V_GAP, 0,
        binFileString, textFileString, fileFormatRing, textFormatMsg, fieldWidthNumeric, writeHeaderCheckBox,
        columnHeadingsString, commentTextBox, 0));

    top += DLG_CTRL_V_GAP;

    errChk( FormatGroupDecoration(dlg, textFormatMsg, textFormatDeco, hMargin, top));
    errChk( SetCtrlAttribute(dlg, textFormatDeco, ATTR_WIDTH, contentWidth));


        /*  make the file strings as big as they can be while keeping them the same size */
    errChk( AttributeMaxAndMin (dlg, ATTR_LABEL_WIDTH, 0, &largestLabelWidth, 0, 0, 0,
        binFileString, textFileString, 0));
    ctrlWidth = contentWidth - largestLabelWidth - DLG_LABEL_TO_CTRL_MIN_H_GAP;
    errChk( SetAttributeForCtrls(dlg, ATTR_LABEL_LEFT, contentLeft, 0, textFileString, binFileString, 0));
    errChk( SetAttributeForCtrls(dlg, ATTR_WIDTH, ctrlWidth, 0, textFileString, binFileString, 0));
    SetCtrlRightPosition(dlg, binFileString, contentRight);
    SetCtrlRightPosition(dlg, textFileString, contentRight);
    errChk( SetAttributeForCtrls(dlg, ATTR_LABEL_LEFT, contentLeft, 0, textFileString, binFileString, 0));

    SetCtrlRightPosition(dlg, fileFormatRing, contentRight);
    SetCtrlAttribute(dlg, fileFormatRing, ATTR_LABEL_LEFT, contentLeft);


        /* make ok and cancel the same width */
    errChk( SetCtrlsToAttributeExtreme(dlg, ATTR_WIDTH, VAL_TB_MAX, 0, ok, cancel, 0));

        /* position ok and cancel vertically */
    top += DLG_CTRL_V_GAP;
    errChk( SetAttributeForCtrls (dlg, ATTR_TOP, top, 0, ok, cancel, 0));

        /* position the ok and cancel horizontally */
    GetCtrlAttribute(dlg, ok, ATTR_WIDTH, &btnWidth);
    SetCtrlAttribute(dlg, cancel, ATTR_LEFT, hMargin + contentWidth - btnWidth);
    SetCtrlAttribute(dlg, ok, ATTR_LEFT, hMargin + contentWidth - 2 * btnWidth - DLG_BTN_H_GAP);

    errChk( GetCtrlAttribute(dlg, ok, ATTR_HEIGHT, &ctrlHeight));
    top += ctrlHeight + DLG_CTRL_V_GAP;

    errChk( SetPanelAttribute(dlg, ATTR_WIDTH, dialogWidth));
    errChk( SetPanelAttribute(dlg, ATTR_HEIGHT, top));

    errChk( SetCommonDialogShortcutKeys(dlg, ok, cancel, cancel));
    errChk( SetPanelPos(dlg, VAL_AUTO_CENTER, VAL_AUTO_CENTER));


    errChk( SetCtrlVal(dlg, binFileString, binFilePath));
    errChk( SetCtrlVal(dlg, textFileString, textFilePath));

        /*  default to currently selected format if not binary, or use last format if this isn't
            the first time.
        */
    DataLog_GetFileFormat(dataLog, &fileFormat);
    if (fileFormat != DATA_LOG_BINARY)
        errChk( SetCtrlVal(dlg, fileFormatRing, fileFormat));
    else
    if (dataLog->lastUsedFileFormat >= 0)
        errChk( SetCtrlVal(dlg, fileFormatRing, dataLog->lastUsedFileFormat));

    SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_DIMMED, fileFormat != DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT);

    errChk( MakeCtrlGrayRecessedIndicator(dlg, binFileString));
    errChk( MakeCtrlGrayRecessedIndicator(dlg, textFileString));

    binFile = fopen (binFilePath, "rb");
    if (!binFile)
        errChk( UIEUnableToOpenFile);

    errChk( DataLog_ReadBinaryHeader(binFile, &isDataLogFile, 0, 0, 0, 0, &columnHeadings, &comment, &writeHeader, &fieldWidth));
    fclose(binFile);
    binFile = 0;

    if (!isDataLogFile)
        errChk( ToolErr_IncorrectFileFormat);

    errChk( SetCtrlVal(dlg, commentTextBox, comment));
    errChk( SetCtrlVal(dlg, columnHeadingsString, columnHeadings));
    errChk( SetCtrlVal(dlg, fieldWidthNumeric, fieldWidth));
    errChk( SetCtrlVal(dlg, writeHeaderCheckBox, writeHeader));
    SetCtrlAttribute(dlg, commentTextBox, ATTR_DIMMED, !writeHeader);
    SetCtrlAttribute(dlg, columnHeadingsString, ATTR_DIMMED, !writeHeader);

    InstallPopup(dlg);

    while (!done)
        {
        GetUserEvent(0, &evtPanel, &evtCtrl);

        if (!ObjectExists(dataLog))
            {
            done = TRUE;
            canceled = TRUE;
            }
        else
        if (evtPanel == dlg)
            {
            if  (evtCtrl == cancel)
                {
                done = TRUE;
                canceled = TRUE;
                }
            else
            if (evtCtrl == ok)
                {
                done = TRUE;
                }
            else
            if (evtCtrl == fileFormatRing)
                {
                GetCtrlVal(dlg, fileFormatRing, &fileFormat);
                SetCtrlAttribute(dlg, fieldWidthNumeric, ATTR_DIMMED, fileFormat != DATA_LOG_CONSTANT_FIELD_WIDTH_TEXT);
                }
            else
            if (evtCtrl == writeHeaderCheckBox)
                {
                GetCtrlVal(dlg, writeHeaderCheckBox, &writeHeader);
                SetCtrlAttribute(dlg, commentTextBox, ATTR_DIMMED, !writeHeader);
                SetCtrlAttribute(dlg, columnHeadingsString, ATTR_DIMMED, !writeHeader);
                changed = TRUE;
                }
            else
            if (evtCtrl == fieldWidthNumeric || evtCtrl == columnHeadingsString
                || evtCtrl == commentTextBox)
                {
                changed = TRUE;
                }
            }
        }

    if (!canceled && changed)
        {
        errChk( GetCtrlVal(dlg, fileFormatRing, &dataLog->lastUsedFileFormat));

        errChk( GetCtrlVal(dlg, fieldWidthNumeric, &fieldWidth));
        errChk( GetCtrlVal(dlg, writeHeaderCheckBox, &writeHeader));

        errChk( GetCtrlAttribute(dlg, columnHeadingsString, ATTR_STRING_TEXT_LENGTH, &columnHeadingsLength));
        free(columnHeadings);
        nullChk( columnHeadings = malloc(columnHeadingsLength + 1));
        errChk( GetCtrlVal(dlg, columnHeadingsString, columnHeadings));

        errChk( GetCtrlAttribute(dlg, commentTextBox, ATTR_STRING_TEXT_LENGTH, &commentLength));
        free(comment);
        nullChk( comment = malloc(commentLength + 1));
        errChk( GetCtrlVal(dlg, commentTextBox, comment));
        }

    if (!canceled)
        {
        int conversionCancelled;

        errChk( progressDlg = CreateProgressDialog ("Conversion Status", "Percent Complete", 1, VAL_NO_INNER_MARKERS, ""));

        errChk( GetCtrlVal(dlg, fileFormatRing, &fileFormat));
        errChk( conversionCancelled = DataLog_ConvertBinaryFile(dataLog, fileFormat, binFilePath, textFilePath, columnHeadings, comment, writeHeader, fieldWidth,
                                            DataLog_ConversionProgressCallback, (void *)progressDlg));
        }

Error:
    DiscardProgressDialog(progressDlg);

    free(comment);
    free(columnHeadings);
    if (binFile)
        fclose(binFile);

    if (dlg > 0)
        DiscardPanel(dlg);

    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (changed)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/*******************************************************************************/

    /*  returns TRUE if the user has requested to cancel the operation */
static int  DataLog_ConversionProgressCallback(int percentComplete, void *callbackData)
{
    int progressDlg = (int)callbackData;

    return UpdateProgressDialog(progressDlg, percentComplete, TRUE);
}

/*******************************************************************************/

    /*  This function returns the default column headings, someday we may store enough info in the
        binary file to use the channel string to generate the defaults, but not yet.
    */
static int DataLog_GetDefaultConvertHeadings(char **defaultHeadings, int numChannels)
{
    int     error = 0;
    int     channel;
    char    *defaults = 0;
    char    columnNameBuf[30];

    nullChk( AppendString(&defaults, "Time", -1));

    for (channel = 1; channel <= numChannels; channel++)
        {
        nullChk( AppendString(&defaults, ", ", -1));

        sprintf(columnNameBuf, "Data Column %d", channel);
        nullChk( AppendString(&defaults, columnNameBuf, -1));
        }

Error:
    if (error < 0)
        {
        free(defaults);
        *defaultHeadings = 0;
        return error;
        }
    else
        {
        *defaultHeadings =defaults;
        return 0;
        }
}

/*******************************************************************************/

#define MAX_SCANS_PER_READ  1

    /*  The dataLog passed is used to determine the text options used only to provide a
        destination for error elaboration messages.

        returns TRUE if conversion was aborted by user, FALSE if conversion completed, or
        negative error code.
    */

static int DataLog_ConvertBinaryFile(DataLog dataLog, int fileFormat, char binFilePath[], char textFilePath[], char *columnHeadings, char *comment, int writeHeader, int fieldWidth,
                ConversionProgressCallback progressProc, void *callbackData)
{
    int     error = 0;
    int     canWrite;
    char    errorElaboration[128];
    FILE    *binFile = 0;
    FILE    *textFile = 0;
    int     isDataLogFile;
    int     numChannels;
    char    *defaultHeadings = 0;
    char    *revisedHeadings = 0;
    double  *rowBuf = 0;
    int     numValuesPerRow;
    int     numRead;
    char    *timeAndDateString = 0;
    int     savedWaitCursor;
    int     headerSize;
    int     binFileSize;
    int     scanDataSize;
    int     numScansInFile;
    int     numScansLeft;
    int     numScansToRead;
    int     cancel = FALSE;
    double  lastScanTime;
    double  scanTime;
    int     firstScan;

        /*  This used to read in chunks for a neglible speed improvement, however, I had to delete that
            feature so we could insert a blank line between the oldest and newest scan
            when converting a wrapped file.  Doing this is a lot easier when processing the scans one at a time.
        */
    Assert(MAX_SCANS_PER_READ == 1);

    savedWaitCursor = GetWaitCursorState();
    SetWaitCursor(TRUE);

    sprintf(errorElaboration, "An error occurred creating the data log file '%s'", textFilePath);   /* this is the error message in case of an error */

    errChk( GetFileWritability(textFilePath, &canWrite));
    if (!canWrite)
        errChk( UIEErrorWritingFile);

    textFile = fopen (textFilePath, "wb");
    if (!textFile)
        errChk( UIEUnableToOpenFile);

    sprintf(errorElaboration, "An error occurred opening the data log file '%s'", binFilePath); /* this is the error message in case of an error */

    binFile = fopen (binFilePath, "rb");
    if (!binFile)
        errChk( UIEUnableToOpenFile);

    errChk( FileExists (binFilePath, &binFileSize));    /* used to get size */

    sprintf(errorElaboration, "");

    errChk( DataLog_ReadBinaryHeader(binFile, &isDataLogFile, 0, &numChannels, &headerSize, &timeAndDateString, 0, 0, 0, 0));
    if (!isDataLogFile)
        errChk( ToolErr_IncorrectFileFormat);

    scanDataSize = binFileSize - headerSize;

    if (writeHeader)
        {
        errChk( DataLog_GetDefaultConvertHeadings(&defaultHeadings, numChannels));
        errChk( SpliceCommaSeparatedList(columnHeadings, defaultHeadings, &revisedHeadings));

        errChk( DataLog_WriteTextHeader(textFile, comment, "DATE AND TIME: ", timeAndDateString, 0,
                numChannels, fileFormat, fieldWidth, revisedHeadings));
        }

    numValuesPerRow = numChannels + 1;  /* + 1 for the time column */

    numScansInFile = scanDataSize / (numValuesPerRow * sizeof(double));
    numScansLeft = numScansInFile;

    nullChk( rowBuf = malloc(sizeof(double) * numValuesPerRow * MAX_SCANS_PER_READ));

    firstScan = TRUE;
    if (numScansInFile > 0)
        {
        do
            {
            numScansToRead = Min(MAX_SCANS_PER_READ, numScansLeft);

            numRead = fread (rowBuf, sizeof(double) * numValuesPerRow, numScansToRead , binFile);
            if (ferror(binFile) || numRead != numScansToRead)
                errChk( UIEErrorReadingFile);

            scanTime = *(double *)rowBuf;
            if (!firstScan)
                {
                if (scanTime < lastScanTime)
                    errChk( WriteStringToFile (textFile, "\n"));
                }
            else
                firstScan = FALSE;

            lastScanTime = scanTime;

            errChk( DataLog_WriteTextScan(textFile, 0.0, 0, rowBuf, numScansToRead, numChannels,  fileFormat,  fieldWidth));
            numScansLeft -= numScansToRead;

            if (progressProc)
                errChk( cancel = progressProc((int)(100.0 - ((double)numScansLeft/numScansInFile) * 100 + 0.5), callbackData));

            if (cancel)
                {
                int choice;

                errChk( choice = GenericMessagePopup ("Abort Binary To Text Conversion?",
                                     "Do you want to abort the conversion?",  "__Abort", "__Continue", 
                                     0, 0, 0, 0, VAL_GENERIC_POPUP_BTN2, VAL_GENERIC_POPUP_BTN2,
                                    VAL_GENERIC_POPUP_BTN2));
                if (choice != VAL_GENERIC_POPUP_BTN1)
                    cancel = FALSE;
                }
            }
        while (numScansLeft && !cancel);
        }

    if (scanDataSize % numValuesPerRow != 0)    /* incomplete last row */
        errChk( ToolErr_UnexpectedEndOfFile);

Error:
    free(defaultHeadings);
    free(revisedHeadings);
    free(timeAndDateString);
    free(rowBuf);
    SetWaitCursor(savedWaitCursor);

    if (textFile)
        fclose(textFile);

    if (binFile)
        fclose(binFile);

    if (error < 0)
        {
        if (*errorElaboration)
            DataLog_NotifyOfErrorElaboration(dataLog, errorElaboration);

        return error;
        }
    else
        return cancel;
}

/****************************************************************************/

    /*  returns a ResponseCode, or a negative error code */
int CVIFUNC DataLog_ReorderLogFile(char *logFilePath, int displayProgressDialog)
{
    int     error = 0;
    int     isDataLogFile;
    int     response;
    int     progressDlg = 0;

    if (displayProgressDialog)
        errChk( progressDlg = CreateProgressDialog ("Placing Scans In Ascending Order", "Percent Complete", 1, VAL_NO_INNER_MARKERS, ""));

    errChk( isDataLogFile = DataLog_IsBinaryLogFile(logFilePath));

    if (isDataLogFile)
        {
        errChk( response = DataLog_ReorderBinaryLogFile(logFilePath, progressDlg ? DataLog_ConversionProgressCallback : 0, (void *)progressDlg));
        }
    else
        {
        errChk( response = DataLog_ReorderTextLogFile(logFilePath, progressDlg ? DataLog_ConversionProgressCallback : 0, (void *)progressDlg));
        }

Error:
    DiscardProgressDialog(progressDlg);
    if (error < 0)
        return error;
    else
        return response;
}

/****************************************************************************/

    /*  Examines the time stamp of each scan to find the earliest one.
        maxPercent is the percentage of the total reordering time this operation is estimated to take.
        This is used to update the progress box.
        Returns TRUE if the user canceled the operation, FALSE otherwise.
        Assumption: numScansInFile > 0
    */
static int FindEarliestScanInBinFile(FILE *file, int rowSize, int dataStartOffset, int numScansInFile, int *earliestScanOffset,
            int *earliestScanIndex, ConversionProgressCallback progressProc, void *callbackData, int maxPercent)
{
    int     error = 0;
    double  earliestTime;
    int     firstScan;
    double  scanTime;
    int     offset;
    int     scanIndex;
    int     numRead;
    int     canceled = FALSE;

    firstScan = TRUE;
    offset = dataStartOffset;

    for (scanIndex = 0; scanIndex < numScansInFile; scanIndex++)
        {
        if (fseek (file, offset, SEEK_SET) != 0)
            errChk( UIEErrorReadingFile);

        numRead = fread (&scanTime, sizeof(double), 1 , file);
        if (ferror(file) || numRead != 1)
            errChk( UIEErrorReadingFile);

        if (progressProc)
            {
            int percentComplete;

            percentComplete =  maxPercent * scanIndex / numScansInFile;
            errChk( canceled = ConfirmReorderingCancellation(progressProc(percentComplete, callbackData)));
            if (canceled)
                goto Error;
            }

        if (firstScan || scanTime < earliestTime)
            {
            firstScan = FALSE;
            earliestTime = scanTime;
            *earliestScanOffset = offset;
            *earliestScanIndex =  scanIndex;
            }

        offset += rowSize;
        }

Error:
    return error;
}

/****************************************************************************/

    /*  returns a ResponseCode, or a negative error code */
static int DataLog_ReorderBinaryLogFile(char *binFilePath, ConversionProgressCallback progressProc, void *callbackData)
{
    int     error = 0;
    FILE    *file = 0;
    int     isDataLogFile;
    int     earliestScanOffset;
    int     earliestScanIndex;
    int     binFileSize;
    int     numChannels;
    int     numValuesPerRow;
    int     scanDataSize;
    int     rowBufSize;
    int     headerSize;
    void    *rowBuf = 0;
    char    tempFilePath[MAX_PATHNAME_LEN];
    FILE    *tempFile = 0;
    void    *headerBuf = 0;
    int     numScansInFile;
    int     numRead;
    int     numWritten;
    int     scanIndex;
    int     reOrdered = FALSE;
    int     canceled = FALSE;
    int     percentDone;

    tempFilePath[0] = 0;

    errChk( FileExists (binFilePath, &binFileSize));    /* used to get size */

    file = fopen (binFilePath, "rb");
    if (!file)
        errChk( UIEUnableToOpenFile);

    errChk( DataLog_ReadBinaryHeader(file, &isDataLogFile, 0, &numChannels, &headerSize, 0, 0, 0, 0, 0));

    if (!isDataLogFile)
        errChk( ToolErr_IncorrectFileFormat);

    scanDataSize = binFileSize - headerSize;
    numValuesPerRow = numChannels + 1;  /* + 1 for the time column */
    rowBufSize = sizeof(double) * numValuesPerRow;
    numScansInFile = scanDataSize / rowBufSize;

    nullChk( rowBuf = malloc(rowBufSize));
    nullChk( headerBuf = malloc(headerSize));

    if (numScansInFile > 0)
        {
        errChk( canceled = FindEarliestScanInBinFile(file, rowBufSize, headerSize, numScansInFile, &earliestScanOffset,
                    &earliestScanIndex, progressProc, callbackData, FIND_FIRST_SCAN_PERCENTAGE));
        if (canceled)
            goto Error;

        if (earliestScanIndex > 0)  /* if out of order */
            {
            reOrdered = TRUE;

            errChk( CreateAndOpenTemporaryFile(binFilePath, "dtlog", "tmp", "wb", tempFilePath, &tempFile));        

                /*  copy the binary file's header */
            if (fseek (file, 0, SEEK_SET) != 0)     /* start back at beginning of the file */
                errChk( UIEErrorReadingFile);

            numRead = fread (headerBuf, headerSize, 1 , file);
            if (ferror(file) || numRead != 1)
                errChk( UIEErrorReadingFile);

            numWritten = fwrite (headerBuf, headerSize, 1, tempFile);
            if (numWritten != 1)
                errChk( UIEDiskFull);


            if (fseek (file, earliestScanOffset, SEEK_SET) != 0)        /* skip to first scan */
                errChk( UIEErrorReadingFile);

            for (scanIndex = earliestScanIndex; scanIndex < numScansInFile; scanIndex++)
                {
                numRead = fread (rowBuf, rowBufSize, 1 , file);
                if (ferror(file) || numRead != 1)
                    errChk( UIEErrorReadingFile);

                numWritten = fwrite (rowBuf, rowBufSize, 1, tempFile);
                if (numWritten != 1)
                    errChk( UIEDiskFull);

                if (progressProc)
                    {
                    percentDone = FIND_FIRST_SCAN_PERCENTAGE + REMAINING_REORDER_PERCENTAGE * (scanIndex - earliestScanIndex) / numScansInFile;
                    errChk( canceled = ConfirmReorderingCancellation(progressProc(percentDone, callbackData)));
                    if (canceled)
                        goto Error;
                    }
                }

            if (fseek (file, headerSize, SEEK_SET) != 0)        /* skip back to the start of the data */
                errChk( UIEErrorReadingFile);

            for (scanIndex = 0; scanIndex < earliestScanIndex; scanIndex++)
                {
                numRead = fread (rowBuf, rowBufSize, 1 , file);
                if (ferror(file) || numRead != 1)
                    errChk( UIEErrorReadingFile);

                numWritten = fwrite (rowBuf, rowBufSize, 1, tempFile);
                if (numWritten != 1)
                    errChk( UIEDiskFull);

                if (progressProc)
                    {
                    percentDone = FIND_FIRST_SCAN_PERCENTAGE + REMAINING_REORDER_PERCENTAGE * (numScansInFile - earliestScanIndex + scanIndex) / numScansInFile;
                    errChk( canceled = ConfirmReorderingCancellation(progressProc(percentDone, callbackData)));
                    if (canceled)
                        goto Error;
                    }
                }
            }
        }

Error:
    free(rowBuf);
    free(headerBuf);

    if (file)
        fclose(file);

    if (tempFile)
        fclose(tempFile);

    if (error >= 0 && reOrdered && FileExists(tempFilePath, 0) && !canceled)
        error = DeleteAndRename(tempFilePath, binFilePath);

    if (*tempFilePath && FileExists(tempFilePath, 0))
        DeleteFile(tempFilePath);

    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (reOrdered)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/****************************************************************************/

    /*  takes the cancel return value from a progress proc and if it is TRUE, confirms it.
        cancel may be a negative error code, in which case it will be returned unchanged
        with no prompting of the user.
    */
static int ConfirmReorderingCancellation(int cancel)
{
    int error = 0;

    if (cancel && cancel > 0)
        {
        int choice;

        errChk( choice = GenericMessagePopup ("Abort Reordering of scans in log file?",
                             "Do you want to abort the reordering?",  "__Abort", "__Continue", 
                             0, 0, 0, 0, VAL_GENERIC_POPUP_BTN2, VAL_GENERIC_POPUP_BTN2,
                            VAL_GENERIC_POPUP_BTN2));
        if (choice != VAL_GENERIC_POPUP_BTN1)
            cancel = FALSE;
        }
Error:
    if (error < 0)
        return error;
    else
        return cancel;
}

/****************************************************************************/

    /*  Examines the time stamp of each scan to find the earliest one.
        maxPercent is the percentage of the total reordering time this operation is estimated to take.
        This is used to update the progress box.
        Returns TRUE if the user canceled the operation, FALSE otherwise.
    */
static int FindEarliestScanInTextFile(FILE *file, int dataStartOffset, int *numScans, int *earliestScanOffset,
                int *earliestScanIndex, ConversionProgressCallback progressProc, void *callbackData,
                int maxPercent)
{
    int     error = 0;
    double  earliestTime;
    int     firstScan;
    double  scanTime;
    int     offset;
    char    *line = 0;
    int     eof;
    int     fileSize;
    int     scanIndex = 0;
    int     approxBytesRead = 0;
    int     canceled = FALSE;

    errChk( BinaryFileSize(file, &fileSize));   /* for progress indicator */

    if (fileSize == 0)
        errChk( ToolErr_IncorrectFileFormat);

    firstScan = TRUE;

    if (fseek (file, dataStartOffset, SEEK_SET) != 0)
        errChk( UIEErrorReadingFile);

    do
        {
        offset = ftell(file);
        if (offset == -1 || errno == 100)
            errChk( UIEIOError);

        free(line);
        line = 0;
        errChk( ReadLineFromBinaryFile(file, &line, -1, &eof, 0));

        if (progressProc)
            {
            int percentComplete;

            approxBytesRead += strlen(line);

            percentComplete =  maxPercent * approxBytesRead / fileSize;
            errChk( canceled = ConfirmReorderingCancellation(progressProc(percentComplete, callbackData)));
            if (canceled)
                goto Error;
            }

        if (!eof && HasNonWhiteSpace(line))
            {
            if (sscanf(line, "%f", &scanTime) == 1)
                {
                if (firstScan || scanTime < earliestTime)
                    {
                    firstScan = FALSE;
                    earliestTime = scanTime;
                    *earliestScanOffset = offset;
                    *earliestScanIndex = scanIndex;
                    }
                scanIndex++;
                }
            }
        } while (!eof);

    if (scanIndex == 0)
        errChk( ToolErr_IncorrectFileFormat);

Error:
    free(line);
    *numScans = scanIndex;
    if (error < 0)
        return error;
    else
        return canceled;
}

/****************************************************************************/

    /*  returns a ResponseCode, or a negative error code */
static int DataLog_ReorderTextLogFile(char *logFilePath, ConversionProgressCallback progressProc, void *callbackData)
{
    int     error = 0;
    FILE    *file = 0;
    int     earliestScanOffset;
    int     earliestScanIndex;
    int     headerSize;
    char    *line = 0;
    char    tempFilePath[MAX_PATHNAME_LEN];
    FILE    *tempFile = 0;
    void    *headerBuf = 0;
    int     numScansInFile;
    int     numRead;
    int     numWritten;
    int     scanIndex;
    int     reOrdered = FALSE;
    int     found;
    int     position;
    int     eof;
    int     canceled = FALSE;
    int     percentDone;

    tempFilePath[0] = 0;

    file = fopen (logFilePath, "rb");
    if (!file)
        errChk( UIEUnableToOpenFile);

    errChk( found = FindBytesInFile(file, LOG_FILE_DATA_START_TAG, LOG_FILE_DATA_START_TAG_LEN, 0, 0));
    if (found)
        {
            /*  skip the start tag and the column headings */
        errChk( found = FindBytesInFile(file, LINE_TERMINATOR, LINE_TERMINATOR_LEN, 0, 0));
        if (!found)
            errChk( ToolErr_IncorrectFileFormat);
        errChk( found = FindBytesInFile(file, LINE_TERMINATOR, LINE_TERMINATOR_LEN, 0, &position));
        if (!found)
            errChk( ToolErr_IncorrectFileFormat);
        headerSize = position + LINE_TERMINATOR_LEN;    /* skip the terminator on the column headings line */
        nullChk( headerBuf = malloc(headerSize));
        }
    else
        headerSize = 0;     /* assume the write header option was turned off */


    errChk(canceled = FindEarliestScanInTextFile(file, headerSize, &numScansInFile, &earliestScanOffset, &earliestScanIndex,
                        progressProc, callbackData, FIND_FIRST_SCAN_PERCENTAGE));
    if (canceled)
        goto Error;

    if (numScansInFile > 0)
        {
        if (earliestScanIndex > 0)  /* if out of order */
            {
            reOrdered = TRUE;

            errChk( CreateAndOpenTemporaryFile(logFilePath, "dtlog", "tmp", "wb", tempFilePath, &tempFile));        

            if (headerSize > 0)
                {
                    /*  copy the file's header */
                if (fseek (file, 0, SEEK_SET) != 0)     /* start back at beginning of the file */
                    errChk( UIEErrorReadingFile);

                numRead = fread (headerBuf, headerSize, 1 , file);
                if (ferror(file) || numRead != 1)
                    errChk( UIEErrorReadingFile);

                numWritten = fwrite (headerBuf, headerSize, 1, tempFile);
                if (numWritten != 1)
                    errChk( UIEDiskFull);
                }

            if (fseek (file, earliestScanOffset, SEEK_SET) != 0)        /* skip to first scan */
                errChk( UIEErrorReadingFile);

            for (scanIndex = earliestScanIndex; scanIndex < numScansInFile; scanIndex++)
                {
                free(line);
                line = 0;
                errChk( ReadLineFromBinaryFile(file, &line, -1, &eof, 0));
                if (eof || !HasNonWhiteSpace(line))
                    errChk( ToolErr_IncorrectFileFormat);

                numWritten = fwrite (line, strlen(line), 1, tempFile);
                if (numWritten != 1)
                    errChk( UIEDiskFull);

                if (progressProc)
                    {
                    percentDone = FIND_FIRST_SCAN_PERCENTAGE + REMAINING_REORDER_PERCENTAGE * (scanIndex - earliestScanIndex) / numScansInFile;
                    errChk( canceled = ConfirmReorderingCancellation(progressProc(percentDone, callbackData)));
                    if (canceled)
                        goto Error;
                    }
                }

            if (fseek (file, headerSize, SEEK_SET) != 0)        /* skip back to the start of the data */
                errChk( UIEErrorReadingFile);

            for (scanIndex = 0; scanIndex < earliestScanIndex; scanIndex++)
                {
                free(line);
                line = 0;
                errChk( ReadLineFromBinaryFile(file, &line, -1, &eof, 0));
                if (eof || !HasNonWhiteSpace(line))
                    errChk( ToolErr_IncorrectFileFormat);

                numWritten = fwrite (line, strlen(line), 1, tempFile);
                if (numWritten != 1)
                    errChk( UIEDiskFull);

                if (progressProc)
                    {
                    percentDone = FIND_FIRST_SCAN_PERCENTAGE + REMAINING_REORDER_PERCENTAGE * (numScansInFile - earliestScanIndex + scanIndex) / numScansInFile;
                    errChk( canceled = ConfirmReorderingCancellation(progressProc(percentDone, callbackData)));
                    if (canceled)
                        goto Error;
                    }
                }
            }
        }

    if (progressProc)
        progressProc(100, callbackData);

Error:
    free(headerBuf);
    free(line);

    if (file)
        fclose(file);

    if (tempFile)
        fclose(tempFile);

    if (error >= 0 && reOrdered && FileExists(tempFilePath, 0) && !canceled)
        error = DeleteAndRename(tempFilePath, logFilePath);

    if (*tempFilePath && FileExists(tempFilePath, 0))
        DeleteFile(tempFilePath);

    if (error < 0)
        return error;
    else
    if (canceled)
        return CANCELED;
    else
    if (reOrdered)
        return CHANGE_MADE;
    else
        return NO_CHANGE_MADE;
}

/*******************************************************************************/

