
#include <userint.h>
#include "toolbox.h"
#include "movectrl.h"
#include "inifile.h"
#include "custsupp.h"


/**********/
/* Macros */
/**********/

#ifndef FALSE
#define FALSE 0
#define TRUE !FALSE
#endif

#define kLabelCtrlGapWidth 3

#define min(x,y)    (((x) < (y)) ? (x) : (y))


#define MOVABLE_CTRL_SANITY     0xacaba56

#define MOVABLE_CTRL_SETTINGS_PREFIX    "MOVECTRL_"

#define CTRL_ID_TAG             "Control%d Control ID"

#define ENABLED_TAG             "Control%d Enabled"
#define CTRL_PART_TAG           "Control%d Control %s"
#define LABEL_PART_TAG          "Control%d Label %s"
#define DIG_DISP_PART_TAG       "Control%d Digital Display %s"

#define POSITION_TAG            "Pos"
#define MOVABLE_TAG             "Movable"
#define SIZABLE_TAG             "Sizable"
#define CENTERABLE_TAG          "Centerable"
#define AUTOCENTER_TAG          "Autocenter"

#define DEFAULT_CONFIG_NAME     "Default"


/*********/
/* Types */
/*********/

typedef enum _PartsEnum
{   
    kLabelPart,
    kDigDispPart,
    kCtrlPart,
    kNumParts,
    kDefaultPart = kCtrlPart
} PartsEnum;
     
typedef struct _PartStruct
{
    char *name;
    char *tagString;
    PartsEnum type;
    int topAttr;
    int leftAttr;
    int heightAttr;
    int widthAttr;
    int visibleAttr;
    int canMoveDefault;
    int canCenterDefault;
    int canSizeDefault;
    int autoCenterDefault;
} PartInfo, *PartInfoPtr;

typedef struct _CenterStruct
{
    int top;
    int bottom;
    int left;
    int right;
} CenterInfo, *CenterInfoPtr;


typedef struct _MovableCtrlInfoStruct
{
    int sanity;
    int enabled;
    int partCanMove[kNumParts];
    int partCanCenter[kNumParts];
    int partCanSize[kNumParts];
    int autoCenterPart[kNumParts];
} MovableCtrlInfo, *MovableCtrlInfoPtr;


/***********/
/* Globals */
/***********/

PartInfo gParts[kNumParts] =
{
    { 
      "Label", 
      LABEL_PART_TAG,
      kLabelPart,
      ATTR_LABEL_TOP, ATTR_LABEL_LEFT, 
      ATTR_LABEL_HEIGHT, ATTR_LABEL_WIDTH,
      ATTR_LABEL_VISIBLE,
      TRUE, TRUE, FALSE, TRUE
    },
    { 
      "Digital Display", 
      DIG_DISP_PART_TAG,
      kDigDispPart,
      ATTR_DIG_DISP_TOP, ATTR_DIG_DISP_LEFT, 
      ATTR_DIG_DISP_HEIGHT, ATTR_DIG_DISP_WIDTH,
      ATTR_SHOW_DIG_DISP,
      TRUE, TRUE, FALSE, TRUE
    },
    { 
      "Control", 
      CTRL_PART_TAG,
      kCtrlPart,
      ATTR_TOP, ATTR_LEFT, 
      ATTR_HEIGHT, ATTR_WIDTH,
      ATTR_VISIBLE,
      TRUE, TRUE, FALSE, FALSE
    }
};


/****************************************/
/*      Private function prototypes     */
/****************************************/

static MovableCtrlInfo *NewCtrlInfo (int movableParts[kNumParts],
                                     int centerableParts[kNumParts],
                                     int sizableParts[kNumParts],
                                     int autoCenterParts[kNumParts]);
static void CreatePartMoveAndSizeInfo (int movableParts[kNumParts],
                                       int centerableParts[kNumParts],
                                       int sizableParts[kNumParts],
                                       int autoCenterParts[kNumParts],
                                       int ctrlMovable, int labelMovable,
                                       int digDispMovable, int sizable);
static MovableCtrlInfo *GetCtrlInfo (int panel, int control);
static int InstallOneDynamicCtrlCallback (int panel, int control,
                                          int ctrlMovable, int labelMovable,
                                          int digDispMovable, int sizable);
                                           
static int  AttributeIsValidForControl (int panel, int control, int attribute,
                                        int *attrSettable);
static void GetClickLocOnCtrl (int panel, int control, int x, int y, 
                               int *xLoc, int *yLoc);
static int  ClickIsInPartRect (int panel, int control, int x, int y,
                               PartInfo *part);

static int  LabelIsOnLeft (int panel, int control);
static int  GetMouseCursorForLoc (int xLoc, int yLoc);

static void TrackCtrlMoveSize (int panel, int control);
static void TrackCtrlSize (int panel, int control, int startX, 
                           int startY, int xLoc, int yLoc);
static void TrackPartMove (int panel, int control, PartInfo *part);

static int  GetKeyState (void);

static PartInfo *GetPartInfo (int partType);
static PartInfo *GetPartInfoForControlCoords (int panel, int control, int x, int y);

static int  CtrlHasPart (int panel, int control, int partType);
static void GetPartCenter (int panel, int control, PartInfo *part, 
                           int *x, int *y, int *height, int *width);
static void GetPartLocation (int panel, int control, PartInfo *part, 
                             CenterInfo *dist);
static void HideExtraParts (int panel, int control, 
                            int hide, int partsToShow[kNumParts]);
static void CenterPart (int panel, int control, PartInfo *part,
                        int useBr);
static void DoCtrlCenter (int panel, int control);

static int  CVICALLBACK DynamicCtrlCallback (int panel, int control, int event,
                                             void *callbackData, int eventData1, 
                                             int eventData2);

static int  WriteControlSettings (int panel, int control, IniText iniText,
                                  char *section, int index, int ignoreNormalControls);
static int  WriteSettings (int panel, int control, IniText iniText,
                           char *section);
static int  ReadPartSettings (int panel, int control, int partType,
                              MovableCtrlInfo *ctrlInfo,
                              IniText iniText, char *section, int index);
static int  ReadControlSettings (int panel, int control, IniText iniText,
                                 char *section, int index, int ignoreNormalControls,
                                 int useCtrlIDFromFile);
static int  ReadSettings (int panel, int control, IniText iniText,
                          char *section);


/************************/
/*      Functions       */
/************************/

static MovableCtrlInfo *NewCtrlInfo (int movableParts[kNumParts],
                                     int centerableParts[kNumParts],
                                     int sizableParts[kNumParts],
                                     int autoCenterParts[kNumParts])
{
    MovableCtrlInfo *info;
    int i;
    
    info = malloc (sizeof (MovableCtrlInfo));
    if (info)
    {
        info->sanity = MOVABLE_CTRL_SANITY;
        info->enabled = TRUE;

        for (i = 0 ; i < kNumParts ; i++)
        {
            info->partCanMove[i] = movableParts[i];
            info->partCanCenter[i] = centerableParts[i];
            info->partCanSize[i] = sizableParts[i];
            info->autoCenterPart[i] = autoCenterParts[i];
        }
    }
    
    return info;
}


static void CreatePartMoveAndSizeInfo (int movableParts[kNumParts],
                                       int centerableParts[kNumParts],
                                       int sizableParts[kNumParts],
                                       int autoCenterParts[kNumParts],
                                       int ctrlMovable, int labelMovable,
                                       int digDispMovable, int sizable)
{
    int i;
    PartInfo *part;
    
    movableParts[kCtrlPart] = ctrlMovable;
    movableParts[kLabelPart] = labelMovable;
    movableParts[kDigDispPart] = digDispMovable;
    
    for (i = 0 ; i < kNumParts ; i++)
    {
        part = GetPartInfo (i);
        
        centerableParts[i] = part->canCenterDefault;
        sizableParts[i] = part->canSizeDefault;
        autoCenterParts[i] = part->autoCenterDefault;
    }
    
    sizableParts[kCtrlPart] = sizable;
}


static MovableCtrlInfo *GetCtrlInfo (int panel, int control)
{
    int error = UIENoError;
    void *cbData;
    MovableCtrlInfo *info;
    
    errChk (GetChainedCallbackData (panel, control, "Movable Control", 
                                    &cbData));
    if (cbData)
    {
        info = (MovableCtrlInfo *)cbData;
        if (info->sanity == MOVABLE_CTRL_SANITY)
            return info;
    }
    
Error:  
    return NULL;
}


static int InstallOneDynamicCtrlCallback (int panel, int control,
                                          int ctrlMovable, int labelMovable,
                                          int digDispMovable, int sizable)
{
    MovableCtrlInfo *info;
    int movableParts[kNumParts];
    int centerableParts[kNumParts];
    int sizableParts[kNumParts];
    int autoCenterParts[kNumParts];
    
    CreatePartMoveAndSizeInfo (movableParts, centerableParts, sizableParts,
                               autoCenterParts, ctrlMovable, labelMovable, 
                               digDispMovable, sizable);
    info = NewCtrlInfo (movableParts, centerableParts, sizableParts,
                        autoCenterParts);
    if (!info)
        return UIEOutOfMemory;

    return ChainCtrlCallback (panel, control, DynamicCtrlCallback, info,
                              "Movable Control");
}


int CVIFUNC_C SetMovableCtrlAttribute (int panel, int control, 
                                       int attribute, ...)
{
    int error = UIENoError;
    MovableCtrlInfo *info;
    va_list parmInfo;
    
    info = GetCtrlInfo (panel, control);
    if (!info)
        return UIEInvalidControlType;
    
    va_start (parmInfo, attribute);
    
    switch (attribute)
    {
        case MOVECTRL_ATTR_ENABLED:
            info->enabled = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_LABEL_MOVABLE:
            info->partCanMove[kLabelPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_LABEL_CENTERABLE:
            info->partCanCenter[kLabelPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_LABEL_AUTOCENTER:
            info->autoCenterPart[kLabelPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_DIG_DISP_MOVABLE:
            info->partCanMove[kDigDispPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_DIG_DISP_CENTERABLE:
            info->partCanCenter[kDigDispPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_DIG_DISP_AUTOCENTER:
            info->autoCenterPart[kDigDispPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_CTRL_MOVABLE:
            info->partCanMove[kCtrlPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_CTRL_CENTERABLE:
            info->partCanCenter[kCtrlPart] = va_arg (parmInfo, int);
            break;
        case MOVECTRL_ATTR_CTRL_SIZABLE:
            info->partCanSize[kCtrlPart] = va_arg (parmInfo, int);
            break;
        default:
            errChk (UIEInvalidAttribute);
            break;
    }
    
Error:
    return error;
}


int CVIFUNC_C GetMovableCtrlAttribute (int panel, int control, 
                                       int attribute, ...)
{
    int error = UIENoError;
    va_list parmInfo;
    int *dataPtr;
    MovableCtrlInfo *info;
    
    info = GetCtrlInfo (panel, control);
    if (!info)
        return UIEInvalidControlType;
    
    va_start (parmInfo, attribute);

    dataPtr = va_arg (parmInfo, int *);
    switch (attribute)
    {
        case MOVECTRL_ATTR_ENABLED:
            *dataPtr = info->enabled;
            break;
        case MOVECTRL_ATTR_LABEL_MOVABLE:
            *dataPtr = info->partCanMove[kLabelPart];
            break;
        case MOVECTRL_ATTR_LABEL_CENTERABLE:
            *dataPtr = info->partCanCenter[kLabelPart];
            break;
        case MOVECTRL_ATTR_LABEL_AUTOCENTER:
            *dataPtr = info->autoCenterPart[kLabelPart];
            break;
        case MOVECTRL_ATTR_DIG_DISP_MOVABLE:
            *dataPtr = info->partCanMove[kDigDispPart];
            break;
        case MOVECTRL_ATTR_DIG_DISP_CENTERABLE:
            *dataPtr = info->partCanCenter[kDigDispPart];
            break;
        case MOVECTRL_ATTR_DIG_DISP_AUTOCENTER:
            *dataPtr = info->autoCenterPart[kDigDispPart];
            break;
        case MOVECTRL_ATTR_CTRL_MOVABLE:
            *dataPtr = info->partCanMove[kCtrlPart];
            break;
        case MOVECTRL_ATTR_CTRL_CENTERABLE:
            *dataPtr = info->partCanCenter[kCtrlPart];
            break;
        case MOVECTRL_ATTR_CTRL_SIZABLE:
            *dataPtr = info->partCanSize[kCtrlPart];
            break;
        default:
            errChk (UIEInvalidAttribute);
            break;
    }
    
Error:
    return error;
}


int CVIFUNC MakeMovableCtrl (int panel, int control, char *configName, 
                             int ctrlMovable, int labelMovable, 
                             int digDispMovable, int sizable)
{
    int error = UIENoError;
    int theCtrl;
    
    if (control == -1) 
    {
        errChk (GetPanelAttribute (panel, ATTR_PANEL_FIRST_CTRL, &theCtrl));
        while (theCtrl > 0) 
        {
            errChk (InstallOneDynamicCtrlCallback (panel, theCtrl, ctrlMovable,
                                                   labelMovable, digDispMovable,
                                                   sizable));
            errChk (GetCtrlAttribute (panel, theCtrl, ATTR_NEXT_CTRL, &theCtrl));
        }
    }
    else 
    {
        errChk (InstallOneDynamicCtrlCallback (panel, control, ctrlMovable,
                                               labelMovable, digDispMovable,
                                               sizable));
    }
    
    if (configName && configName[0])
    {
        errChk (LoadMovableCtrlSettings (panel, control, configName));
    }
    
Error :
    return error;
}


/***************************************************************************/
/* NOTE:  This will fail with a run-time error if attrSettable is non-NULL */
/* and the attribute type is not int!!!  It is impossible to determine the */
/* attribute type programmatically (without making a table of all the      */
/* attributes), so we have to assume the most common case, int.            */
/***************************************************************************/

static int AttributeIsValidForControl (int panel, int control, int attribute,
                                       int *attrSettable)
{
    int breakEnabled;
    int error;
    int isValid = FALSE, isSettable = FALSE;
    char attrBuf[512];
    
    breakEnabled = SetBOLE (FALSE);
    
    error = GetCtrlAttribute (panel, control, attribute, (void *)&attrBuf);
    if ((error != UIEAttributeNotAppropriateForSpecifiedControlOrPanelOrMenu) &&
        (error != UIEInvalidAttribute))
        isValid = TRUE;

    /* Don't bother to check if there is no pointer for attrSettable */
    if (isValid && attrSettable)
    {
        /* Determine if a set attribute can work */
        error = SetCtrlAttribute (panel, control, attribute, *((int *)attrBuf));
        if (error != UIEInvalidAttribute)
            isSettable = TRUE;
    }
    
    SetBOLE (breakEnabled);
    
    if (attrSettable)
        *attrSettable = isSettable;
        
    return isValid;
}


/************************************************************************/
/* Sets xLoc and yLoc according to where the click (x,y) is relative to */
/* the control.  x and y are panel coordinates on the control's parent  */
/* panel.  This uses the control area as defined by the top, left,      */
/* width, and height, NOT the bounding rect.                            */
/*                                                                      */
/* Return values :                                                      */
/*      xLoc:                           yLoc:                           */
/*          -1: left border                 -1: top border              */
/*           0: center border                0: center                  */
/*           1: right border                 1: bottom border           */
/************************************************************************/

#define kBorderWidth 3

static void GetClickLocOnCtrl (int panel, int control, int x, int y, 
                               int *xLoc, int *yLoc)
{
    int top, left, height, width, bottom, right;
    
    GetCtrlAttribute (panel, control, ATTR_TOP, &top);
    GetCtrlAttribute (panel, control, ATTR_LEFT, &left);
    GetCtrlAttribute (panel, control, ATTR_HEIGHT, &height);
    GetCtrlAttribute (panel, control, ATTR_WIDTH, &width);
    right = left + width;
    bottom = top + height;
    
    if (x <= left + kBorderWidth)
        *xLoc = -1;
    else if (x <= right - kBorderWidth)
        *xLoc = 0;
    else
        *xLoc = 1;
        
    if (y <= top + kBorderWidth)
        *yLoc = -1;
    else if (y <= bottom - kBorderWidth)
        *yLoc = 0;
    else
        *yLoc = 1;
}


static int ClickIsInPartRect (int panel, int control, int x, int y,
                              PartInfo *part)
{
    int top, left, height, width;

    if (AttributeIsValidForControl (panel, control, part->topAttr, NULL))
    {
        GetCtrlAttribute (panel, control, part->topAttr, &top);
        GetCtrlAttribute (panel, control, part->leftAttr, &left);
        GetCtrlAttribute (panel, control, part->heightAttr, &height);
        GetCtrlAttribute (panel, control, part->widthAttr, &width);
    
        if ((y >= top) && (y <= top + height) &&
            (x >= left) && (x <= left + width))
            return TRUE;
    }
    
    return FALSE;
}


static int LabelIsOnLeft (int panel, int control)
{
    int error = UIENoError;
    int labelWidth, left, labelLeft;
    int onLeft = FALSE;
    
    if (AttributeIsValidForControl (panel, control, ATTR_LABEL_LEFT, NULL)) 
    {
        GetCtrlAttribute (panel, control, ATTR_LEFT, &left);
        GetCtrlAttribute (panel, control, ATTR_LABEL_LEFT, &labelLeft);
        GetCtrlAttribute (panel, control, ATTR_LABEL_WIDTH, &labelWidth);
        
        if ((labelLeft + labelWidth) < left)
            onLeft = TRUE;
    }
    
    return onLeft;
}


static int GetMouseCursorForLoc (int xLoc, int yLoc)
{
    if ((xLoc == 0) && (yLoc == 0))
        return VAL_OPEN_HAND_CURSOR;
    else if (xLoc == 0)
        return VAL_SIZE_NS_CURSOR;
    else if (yLoc == 0)
        return VAL_SIZE_EW_CURSOR;
    else if (xLoc == yLoc)
        return VAL_SIZE_NW_SE_CURSOR;
    else
        return VAL_SIZE_NE_SW_CURSOR;
}


static void TrackCtrlMoveSize (int panel, int control)
{
    int startX, startY, leftButton;
    int xLoc, yLoc;
    PartInfo *part;
    
    GetRelativeMouseState (panel, 0, &startX, &startY, &leftButton, NULL, NULL);

    /********************************************************************/
    /* If the click is not in the control rect, it must be on the label */
    /* or extra parts, treat this as a move action.                     */
    /********************************************************************/
    part = GetPartInfoForControlCoords (panel, control, startX, startY);
    if (part->type != kCtrlPart) 
    {
        TrackPartMove (panel, control, part);
        return;         
    }
    
    /*****************************************************************/
    /* If the click is in the center of the control (far enough away */
    /* from both borders), this is also a move action.               */
    /*****************************************************************/
    GetClickLocOnCtrl (panel, control, startX, startY, &xLoc, &yLoc);
    if ((xLoc == 0) && (yLoc == 0)) 
    {
        TrackPartMove (panel, control, part);
        return;
    }
        
    /*************************************************/
    /* Now we know it's a resize action of some kind */
    /*************************************************/
    TrackCtrlSize (panel, control, startX, startY, xLoc, yLoc);
}


static void TrackCtrlSize (int panel, int control, int startX, 
                           int startY, int xLoc, int yLoc)
{
    int x, y, leftButton;
    int newWidth, newHeight;
    int labelOnLeft, breakEnabled, cursor;
    int oldTop, oldLeft, oldWidth, oldHeight;
    int dX, dY;
    MovableCtrlInfo *info;
    int i;
    
    info = GetCtrlInfo (panel, control);
    if (!info || !info->partCanSize[kCtrlPart])
        return;
        
    /*********************************************/
    /* Set appropriate mouse cursor for location */
    /*********************************************/
    GetMouseCursor (&cursor);
    SetMouseCursor (GetMouseCursorForLoc (xLoc, yLoc));

    GetCtrlAttribute (panel, control, ATTR_TOP, &oldTop);
    GetCtrlAttribute (panel, control, ATTR_LEFT, &oldLeft);
    GetCtrlAttribute (panel, control, ATTR_HEIGHT, &oldHeight);
    GetCtrlAttribute (panel, control, ATTR_WIDTH, &oldWidth);

    labelOnLeft = LabelIsOnLeft (panel, control);
    breakEnabled = SetBOLE (FALSE);
    while (leftButton) 
    {
        GetRelativeMouseState (panel, 0, &x, &y, &leftButton, NULL, NULL);
        dX = (x - startX) * xLoc;
        dY = (y - startY) * yLoc;
        
        newWidth = oldWidth + dX;
        newHeight = oldHeight + dY;
        
        if (newWidth >= 0) 
        {
            if (xLoc == -1)
                SetCtrlAttribute (panel, control, ATTR_LEFT, oldLeft - dX);
            SetCtrlAttribute (panel, control, ATTR_WIDTH, newWidth);
        }
        if (newHeight >= 0) 
        {
            if (yLoc == -1)
                SetCtrlAttribute (panel, control, ATTR_TOP, oldTop - dY);
            SetCtrlAttribute (panel, control, ATTR_HEIGHT, newHeight);
        }

        for (i = 0 ; i < kNumParts ; i++)
        {
            if (info->autoCenterPart[i])
                CenterPart (panel, control, GetPartInfo (i), TRUE);
        }
        
        ProcessSystemEvents();      
    }
    SetBOLE (breakEnabled);
    
    SetMouseCursor (cursor);
}


static void TrackPartMove (int panel, int control, PartInfo *part)
{
    int x, y, leftButton, xOffset, yOffset;
    int partTop, partLeft, ctrlTop, ctrlLeft;
    int cursor;
    MovableCtrlInfo *info;
    
    info = GetCtrlInfo (panel, control);
    if (!info || !info->partCanMove[part->type])
        return;
        
    GetMouseCursor (&cursor);
    SetMouseCursor (VAL_OPEN_HAND_CURSOR);
                      
    GetRelativeMouseState (panel, control, &xOffset, &yOffset, 
                           &leftButton, NULL, NULL);
    
    GetCtrlAttribute (panel, control, ATTR_TOP, &ctrlTop);
    GetCtrlAttribute (panel, control, ATTR_LEFT, &ctrlLeft);
    GetCtrlAttribute (panel, control, part->topAttr, &partTop);
    GetCtrlAttribute (panel, control, part->leftAttr, &partLeft);
    xOffset += (ctrlLeft - partLeft);
    yOffset += (ctrlTop - partTop);

    while (leftButton) 
    {
        GetRelativeMouseState (panel, 0, &x, &y, &leftButton, NULL, NULL);
        SetCtrlAttribute (panel, control, part->topAttr, y - yOffset);
        SetCtrlAttribute (panel, control, part->leftAttr, x - xOffset);
        ProcessSystemEvents();      
    }
     
    SetMouseCursor (cursor);
}


static int GetKeyState (void)
{
    int keys;
    
    GetGlobalMouseState (NULL, NULL, NULL, NULL, NULL, &keys);
    return keys;
}


static PartInfo *GetPartInfo (int partType)
{
    return &gParts[partType];
}


static PartInfo *GetPartInfoForControlCoords (int panel, int control, int x, int y)
{
    int i;
    PartInfo *part;
    
    for (i = 0 ; i < kNumParts ; i++)
    {
        part = GetPartInfo (i);
        if (ClickIsInPartRect (panel, control, x, y, part))
            return part;
    }
    
    return GetPartInfo (kDefaultPart);
}


static int CtrlHasPart (int panel, int control, int partType)
{
    PartInfo *part;
    
    part = GetPartInfo (partType);
    return AttributeIsValidForControl (panel, control, part->topAttr, NULL);
}


static void GetPartCenter (int panel, int control, PartInfo *part, 
                           int *x, int *y, int *h, int *w)
{
    int top, left, height, width;
    
    GetCtrlAttribute (panel, control, part->topAttr, &top);
    GetCtrlAttribute (panel, control, part->leftAttr, &left);
    GetCtrlAttribute (panel, control, part->heightAttr, &height);
    GetCtrlAttribute (panel, control, part->widthAttr, &width);
    
    /*****************************************************************/
    /* Width of area around center which should be considered center */
    /* of control for purposes of comparing parts.                   */
    /*****************************************************************/
    *h = height / 2;
    *w = width / 2;

    /**********************/
    /* Center of the part */
    /**********************/
    *x = left + (width / 2);
    *y = top + (height / 2);
}


/**********************************************************/
/* Determine which side (top, bottom, left, right) around */
/* which to center part.  This is determined by checking  */
/* the part center with respect to the control center and */
/* the distance from the four axes.  A lower value in the */
/* dist structure means a closer match.                   */
/**********************************************************/

static void GetPartLocation (int panel, int control, PartInfo *part, 
                             CenterInfo *dist)
{
    int ctrlX, ctrlY, ctrlHeight, ctrlWidth;
    int partX, partY, partHeight, partWidth;
    PartInfo *ctrlPart;
    
    ctrlPart = GetPartInfo (kCtrlPart);
    GetPartCenter (panel, control, ctrlPart, &ctrlX, 
                   &ctrlY, &ctrlHeight, &ctrlWidth);
    GetPartCenter (panel, control, part, &partX, &partY,
                   &partHeight, &partWidth);
    
    /* Initialize all to n/a at first */
    dist->top = -1;
    dist->bottom = -1;
    dist->left = -1;
    dist->right = -1;

    /* Determine whether top or bottom axis is closer */
    if (partY <= ctrlY)
        dist->top = abs (partX - ctrlX);
    else
        dist->bottom = abs (partX - ctrlX);
        
    /* Determine whether left or right axis is closer */
    if (partX <= ctrlX)
        dist->left = abs (partY - ctrlY);
    else        
        dist->right = abs (partY - ctrlY);
}


static void CenterPartOnTop (int panel, int control, PartInfo *part,
                             int useBoundingRect)
{
    int ctrlTop, partHeight;

    if (useBoundingRect)
        GetCtrlBoundingRect (panel, control, &ctrlTop, NULL, NULL, NULL);
    else
        GetCtrlAttribute (panel, control, ATTR_TOP, &ctrlTop);
        
    GetCtrlAttribute(panel, control, part->heightAttr, &partHeight);
    SetCtrlAttribute(panel, control, part->topAttr, ctrlTop - partHeight - 1);
    SetCtrlAttribute(panel, control, part->leftAttr, VAL_AUTO_CENTER);
    SetCtrlAttribute(panel, control, part->topAttr, 
                     ctrlTop - partHeight - kLabelCtrlGapWidth);
}


static void CenterPartOnBottom (int panel, int control, PartInfo *part,
                                int useBoundingRect)
{
    int ctrlTop, ctrlHeight;

    if (useBoundingRect)
        GetCtrlBoundingRect (panel, control, &ctrlTop, NULL, &ctrlHeight,
                             NULL);
    else
    {
        GetCtrlAttribute (panel, control, ATTR_TOP, &ctrlTop);
        GetCtrlAttribute (panel, control, ATTR_HEIGHT, &ctrlHeight);
    }
    
    SetCtrlAttribute(panel, control, part->topAttr, ctrlTop + ctrlHeight + 1);
    SetCtrlAttribute(panel, control, part->leftAttr, VAL_AUTO_CENTER);
    SetCtrlAttribute(panel, control, part->topAttr, 
                     ctrlTop + ctrlHeight + kLabelCtrlGapWidth);
}


static void CenterPartOnLeft (int panel, int control, PartInfo *part,
                              int useBoundingRect)
{
    int ctrlLeft, partWidth;

    if (useBoundingRect)
        GetCtrlBoundingRect (panel, control, NULL, &ctrlLeft, NULL, NULL);
    else
        GetCtrlAttribute (panel, control, ATTR_LEFT, &ctrlLeft);
    
    GetCtrlAttribute(panel, control, part->widthAttr, &partWidth);
    SetCtrlAttribute(panel, control, part->leftAttr, ctrlLeft - partWidth - 1);
    SetCtrlAttribute(panel, control, part->topAttr, VAL_AUTO_CENTER);
    SetCtrlAttribute(panel, control, part->leftAttr, 
                     ctrlLeft - partWidth - kLabelCtrlGapWidth);
}


static void CenterPartOnRight (int panel, int control, PartInfo *part,
                               int useBoundingRect)
{
    int ctrlWidth, ctrlLeft;
    
    if (useBoundingRect)
        GetCtrlBoundingRect (panel, control, NULL, &ctrlLeft, NULL,
                             &ctrlWidth);
    else
    {
        GetCtrlAttribute (panel, control, ATTR_LEFT, &ctrlLeft);
        GetCtrlAttribute (panel, control, ATTR_WIDTH, &ctrlWidth);
    }

    SetCtrlAttribute(panel, control, part->leftAttr, ctrlLeft + ctrlWidth + 1);
    SetCtrlAttribute(panel, control, part->topAttr, VAL_AUTO_CENTER);
    SetCtrlAttribute(panel, control, part->leftAttr, 
                     ctrlLeft + ctrlWidth + kLabelCtrlGapWidth);
}


static int IsClosestAxis (CenterInfo *dist, int *valPtr)
{
    int minVal = INT_MAX;
    
    if (dist->top != -1)
        minVal = min(minVal, dist->top);
    if (dist->bottom != -1)
        minVal = min(minVal, dist->bottom);
    if (dist->left != -1)
        minVal = min(minVal, dist->left);
    if (dist->right != -1)
        minVal = min(minVal, dist->right);
        
    return (minVal == *valPtr);
}


static void HideExtraParts (int panel, int control, 
                            int hide, int partsToShow[kNumParts])
{
    int i, wasVisible;
    PartInfo *part;
    
    for (i = 0 ; i < kNumParts ; i++)
    {
        part = GetPartInfo (i);
        
        if (part->type == kCtrlPart)
            continue;

        if (!CtrlHasPart (panel, control, i))
            continue;

        if (hide)
        {
            GetCtrlAttribute (panel, control, part->visibleAttr, &wasVisible);
            partsToShow[i] = wasVisible;
            SetCtrlAttribute (panel, control, part->visibleAttr, FALSE);
        }
        else if (partsToShow[i])
            SetCtrlAttribute (panel, control, part->visibleAttr, TRUE);
    }
}


static void CenterPart (int panel, int control, PartInfo *part,
                        int useBr)
{
    CenterInfo dist;
    int partsToShow[kNumParts];
    MovableCtrlInfo *info;
    
    if (!CtrlHasPart (panel, control, part->type))
        return;

    info = GetCtrlInfo (panel, control);
    if (!info || !info->partCanCenter[part->type])
        return;
    
    GetPartLocation (panel, control, part, &dist);

    if (useBr)
        HideExtraParts (panel, control, TRUE, partsToShow);
    
    if (IsClosestAxis (&dist, &dist.top))
        CenterPartOnTop (panel, control, part, useBr);
    else if (IsClosestAxis (&dist, &dist.bottom))
        CenterPartOnBottom (panel, control, part, useBr);
    else if (IsClosestAxis (&dist, &dist.left))
        CenterPartOnLeft (panel, control, part, useBr);
    else if (IsClosestAxis (&dist, &dist.right))
        CenterPartOnRight (panel, control, part, useBr);

    if (useBr)
        HideExtraParts (panel, control, FALSE, partsToShow);
}


static void DoCtrlCenter (int panel, int control)
{
    int startX, startY;
    PartInfo *part;
    
    GetRelativeMouseState (panel, 0, &startX, &startY, NULL, NULL, NULL);
    part = GetPartInfoForControlCoords (panel, control, startX, startY);

    if (part->type == kCtrlPart)
    {
        CenterPart (panel, control, GetPartInfo (kLabelPart), TRUE);
        CenterPart (panel, control, GetPartInfo (kDigDispPart), TRUE);
    }
    else
        CenterPart (panel, control, part, TRUE);
}


static int CVICALLBACK DynamicCtrlCallback (int panel, int control, int event,
                                            void *callbackData, int eventData1, 
                                            int eventData2)
{
    MovableCtrlInfo *info;
    
    if (GetKeyState() == VAL_MENUKEY_MODIFIER) 
    {
        info = GetCtrlInfo (panel, control);
        if (!info || (info->enabled == FALSE))
            return (FALSE);
            
        switch (event) 
        {
            case EVENT_LEFT_CLICK:
                TrackCtrlMoveSize (panel, control);
                return (TRUE);
                break;
            case EVENT_RIGHT_CLICK:
                DoCtrlCenter (panel, control);
                break;
        }
    }
    
    return (FALSE);
}


static int WriteControlSettings (int panel, int control, IniText iniText,
                                 char *section, int index, int ignoreNormalControls)
{
    int     error = 0;
    char    itemName[128];
    char    geometryString[50];     /* Large enough to hold four ints */
    int     top, left, height, width;
    int     i;
    MovableCtrlInfo *info;
    PartInfo *part;
    
    info = GetCtrlInfo (panel, control);
    if (!info)
    {
        if (!ignoreNormalControls)
            error = UIEInvalidControlType;

        goto Error;
    }
    
    sprintf (itemName, CTRL_ID_TAG, index);
    errChk (Ini_PutInt (iniText, section, itemName, control));
    
    sprintf (itemName, ENABLED_TAG, index);
    errChk (Ini_PutBoolean (iniText, section, itemName, info->enabled));
    
    for (i = 0 ; i < kNumParts ; i++)
    {
        if (!CtrlHasPart (panel, control, i))
            continue;
            
        part = GetPartInfo (i);

        /* Put current position and size */
        GetCtrlAttribute (panel, control, part->topAttr, &top);
        GetCtrlAttribute (panel, control, part->leftAttr, &left);
        GetCtrlAttribute (panel, control, part->heightAttr, &height);
        GetCtrlAttribute (panel, control, part->widthAttr, &width);
        
        sprintf (geometryString, "%d %d %d %d", top, left, height, width);
        sprintf (itemName, part->tagString, index, POSITION_TAG);
        errChk (Ini_PutString (iniText, section, itemName, geometryString));
        
        /* Put movable attribute */
        sprintf (itemName, part->tagString, index, MOVABLE_TAG);
        errChk (Ini_PutBoolean (iniText, section, itemName, info->partCanMove[i]));
        
        /* Put sizable attribute */         
        sprintf (itemName, part->tagString, index, SIZABLE_TAG);
        errChk (Ini_PutBoolean (iniText, section, itemName, info->partCanSize[i]));

        /* Put centerable attribute */
        sprintf (itemName, part->tagString, index, CENTERABLE_TAG);
        errChk (Ini_PutBoolean (iniText, section, itemName, info->partCanCenter[i]));

        /* Put autocenter attribute */
        sprintf (itemName, part->tagString, index, AUTOCENTER_TAG);
        errChk (Ini_PutBoolean (iniText, section, itemName, info->autoCenterPart[i]));
    }

Error:    
    return error;
}


static int WriteSettings (int panel, int control, IniText iniText,
                          char *section)
{
    int error = 0;
    int index, theCtrl;
    
    if (control == -1)
    {
        index = 1;
        GetPanelAttribute (panel, ATTR_PANEL_FIRST_CTRL, &theCtrl);
        while (theCtrl > 0)
        {
            errChk (WriteControlSettings (panel, theCtrl, iniText, 
                                          section, index, TRUE));
            index++;
            
            GetCtrlAttribute (panel, theCtrl, ATTR_NEXT_CTRL, &theCtrl);
        }
    }
    else
    {
        errChk (WriteControlSettings (panel, control, iniText, section, 1, FALSE));
    }

Error:    
    return error;
}


int CVIFUNC SaveMovableCtrlSettings (int panel, int control, 
                                     char *settingsName)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 128];     /* control prefix should never be anywhere near 128 characters */
    int prevCursor;
    
    GetMouseCursor (&prevCursor);
    SetMouseCursor (VAL_HOUR_GLASS_CURSOR);

    errChk (ReadConfigIni(&iniText));

    sprintf (section, "%s%s", MOVABLE_CTRL_SETTINGS_PREFIX, settingsName);
    Ini_RemoveSection (iniText, section);

    errChk (WriteSettings (panel, control, iniText, section));

    errChk (WriteConfigIni(iniText));

    SetMouseCursor (prevCursor);
    
Error:
    Ini_Dispose (iniText);
    return error;
}


static int ReadPartSettings (int panel, int control, int partType,
                             MovableCtrlInfo *ctrlInfo,
                             IniText iniText, char *section, int index)
{
    int     error = 0;
    char    geometryString[50];     /* Large enough to hold four ints */
    char    itemName[128];
    int     top, left, height, width;
    int     isSettable; 
    PartInfo *part;
    
    if (!CtrlHasPart (panel, control, partType))
        goto Error;

    part = GetPartInfo (partType);
        
    /* Get current position and size */
    sprintf (itemName, part->tagString, index, POSITION_TAG);
    errChk (Ini_GetStringIntoBuffer (iniText, section, itemName, geometryString,
                                     sizeof (geometryString)));
    sscanf (geometryString, "%d %d %d %d", &top, &left, &height, &width);
    
    SetCtrlAttribute (panel, control, part->topAttr, top);
    SetCtrlAttribute (panel, control, part->leftAttr, left);
    SetCtrlAttribute (panel, control, part->widthAttr, width);
    
    /* Height Attribute is not always valid for all control types */
    if (AttributeIsValidForControl (panel, control, part->heightAttr, 
                                    &isSettable))
    {
        if (isSettable)
            SetCtrlAttribute (panel, control, part->heightAttr, height);
    }
    
    /* Get movable attribute */
    sprintf (itemName, part->tagString, index, MOVABLE_TAG);
    errChk (Ini_GetBoolean (iniText, section, itemName, &ctrlInfo->partCanMove[part->type]));
    
    /* Get sizable attribute */         
    sprintf (itemName, part->tagString, index, SIZABLE_TAG);
    errChk (Ini_GetBoolean (iniText, section, itemName, &ctrlInfo->partCanSize[part->type]));

    /* Get centerable attribute */
    sprintf (itemName, part->tagString, index, CENTERABLE_TAG);
    errChk (Ini_GetBoolean (iniText, section, itemName, &ctrlInfo->partCanCenter[part->type]));

    /* Get autocenter attribute */
    sprintf (itemName, part->tagString, index, AUTOCENTER_TAG);
    errChk (Ini_GetBoolean (iniText, section, itemName, &ctrlInfo->autoCenterPart[part->type]));
    
Error:
    return error;
}


static int ReadControlSettings (int panel, int control, IniText iniText,
                                char *section, int index, int ignoreNormalControls,
                                int useCtrlIDFromFile)
{
    int     error = 0;
    char    itemName[128];
    int     i;
    int     ctrlId;
    MovableCtrlInfo *info;
    
    if (useCtrlIDFromFile)
    {
        sprintf (itemName, CTRL_ID_TAG, index);
        errChk (Ini_GetInt (iniText, section, itemName, &ctrlId));
    }
    else
        ctrlId = control;
    
    info = GetCtrlInfo (panel, ctrlId);
    if (!info)
    {
        if (!ignoreNormalControls)
            error = UIEInvalidControlType;

        goto Error;
    }
    
    sprintf (itemName, ENABLED_TAG, index);
    errChk (Ini_GetBoolean (iniText, section, itemName, &info->enabled));

    /***************************************************************/
    /* Must read the control part first so that the label/dig disp */
    /* don't get moved when the control position is set.           */
    /***************************************************************/
    ReadPartSettings (panel, ctrlId, kCtrlPart, info, 
                      iniText, section, index);
    
    for (i = 0 ; i < kNumParts ; i++)
    {
        if (i != kCtrlPart)
            ReadPartSettings (panel, ctrlId, i, info, 
                              iniText, section, index);
    }

Error:    
    return error;
}


static int ReadSettings (int panel, int control, IniText iniText,
                         char *section)
{
    int     error = 0;
    int     index;
    char    itemName[128];
    
    if (control == -1)
    {
        index = 1;
        sprintf (itemName, CTRL_ID_TAG, index);
        while (Ini_ItemExists (iniText, section, itemName))
        {
            errChk (ReadControlSettings (panel, 0, iniText, 
                                         section, index, TRUE, TRUE));
            index++;                                         
            sprintf (itemName, CTRL_ID_TAG, index);
        }
    }
    else
    {
        errChk (ReadControlSettings (panel, control, iniText, section, 1, FALSE, FALSE));
    }

Error:
    return error;
}


int CVIFUNC LoadMovableCtrlSettings (int panel, int control, 
                                     char *settingsName)
{
    int     error = 0;
    IniText iniText = 0;
    char    section[MAX_SETTING_NAME_LENGTH + 128];     /* control prefix should never be anywhere near 128 characters */
    int prevCursor;
    
    GetMouseCursor (&prevCursor);
    SetMouseCursor (VAL_HOUR_GLASS_CURSOR);
    
    errChk (ReadConfigIni(&iniText));

    sprintf (section, "%s%s", MOVABLE_CTRL_SETTINGS_PREFIX, settingsName);
    if (!Ini_SectionExists (iniText, section))
        sprintf (section, "%s%s", MOVABLE_CTRL_SETTINGS_PREFIX, DEFAULT_CONFIG_NAME);
    
    errChk (ReadSettings (panel, control, iniText, section));
    
    SetMouseCursor (prevCursor);
    
Error:
    return error;
}
