#include <formatio.h>
#include <utility.h>
#include <ansi_c.h>
#include <userint.h>
#include "toolbox.h"
#include "pwctrl.h"


#define DEFAULT_MASK_CHARACTER  '*'
#define PASSWORD_CTRL_TYPE      "Password Control"

typedef struct
    {
    int     panel;
    int     ctrl;
    int     maxPasswordLength;
    char    password[MAX_POSSIBLE_PASSWORD_LENGTH+1];
    char    maskCharacter;
    } *PasswordInfo;

static int  UpdateObscuringCharacters(int panel, int passwordCtrl);
static int  CreatePasswordInfo(int panel, int ctrl, PasswordInfo *newPasswordInfo);
static void DiscardPasswordInfo(PasswordInfo passwordInfo);

int  CVICALLBACK PasswordCtrl_Callback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2);


/*****************************************************************************************/

static int UpdateObscuringCharacters(int panel, int passwordCtrl)
{
    int             error = 0;
    char            maskBuffer[MAX_POSSIBLE_PASSWORD_LENGTH+1];
    int             strLength;
    PasswordInfo    pwInfo = 0;

    errChk( GetChainedCallbackData(panel, passwordCtrl, PASSWORD_CTRL_TYPE, (void **)&pwInfo));

    strLength = strlen(pwInfo->password);
    memset (maskBuffer, pwInfo->maskCharacter, strLength);
    maskBuffer[strLength] = 0;
    errChk(SetCtrlVal(panel, passwordCtrl, maskBuffer));

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC PasswordCtrl_Create(int panel, char *label, int top, int left)
{
    int stringCtrl;
    int error  = 0;

    errChk(stringCtrl = NewCtrl (panel, CTRL_STRING, label, top, left));

    return PasswordCtrl_ConvertFromString(panel, stringCtrl);

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC PasswordCtrl_ConvertFromString(int panel, int stringCtrl)
{
    int             error = 0;
    char            password[MAX_POSSIBLE_PASSWORD_LENGTH+1];
    PasswordInfo    pwInfo = 0;

    errChk( CreatePasswordInfo(panel, stringCtrl, &pwInfo));

    PasswordCtrl_SetAttribute(panel, stringCtrl, ATTR_PASSWORD_MASK_CHARACTER, DEFAULT_MASK_CHARACTER);
    PasswordCtrl_SetAttribute(panel, stringCtrl, ATTR_PASSWORD_MAX_LENGTH, MAX_POSSIBLE_PASSWORD_LENGTH);

    GetCtrlVal(panel, stringCtrl, password);
    errChk( PasswordCtrl_SetAttribute(panel, stringCtrl, ATTR_PASSWORD_VAL, password) );

    return stringCtrl;

Error:
    DiscardPasswordInfo(pwInfo);
    return error;
}

/*****************************************************************************************/

static int CreatePasswordInfo(int panel, int ctrl, PasswordInfo *newPasswordInfo)
{
    int             error = 0;
    PasswordInfo    pwInfo = 0;

    nullChk(pwInfo = (PasswordInfo)calloc(sizeof(*pwInfo), 1));

    pwInfo->panel = panel;
    pwInfo->ctrl = ctrl;

    errChk( ChainCtrlCallback(panel, ctrl, PasswordCtrl_Callback, (void *)pwInfo, PASSWORD_CTRL_TYPE));

    if (newPasswordInfo)
        *newPasswordInfo = pwInfo;

Error:
    if (error < 0)
        {
        if (newPasswordInfo)
            *newPasswordInfo = 0;
        return error;
        }
    else
        return error;
}

/*****************************************************************************************/

static void DiscardPasswordInfo(PasswordInfo passwordInfo)
{
    if (passwordInfo)
        {
        UnchainCtrlCallback(passwordInfo->panel, passwordInfo->ctrl, PASSWORD_CTRL_TYPE);
        free(passwordInfo);
        }
}

/*****************************************************************************************/

int CVIFUNC_C PasswordCtrl_SetAttribute(int panel, int passwordCtrl, int attribute, ...)
{
    int             error = 0;
    va_list         parmInfo;
    PasswordInfo    pwInfo = 0;

    errChk( GetChainedCallbackData(panel, passwordCtrl, PASSWORD_CTRL_TYPE, (void **)&pwInfo));

    va_start(parmInfo, attribute);

    switch (attribute)
        {
        case ATTR_PASSWORD_MAX_LENGTH:
            pwInfo->maxPasswordLength = va_arg(parmInfo, int);

            if (pwInfo->maxPasswordLength < 0)
                pwInfo->maxPasswordLength = 0;
            else
            if (pwInfo->maxPasswordLength > MAX_POSSIBLE_PASSWORD_LENGTH)
                pwInfo->maxPasswordLength = MAX_POSSIBLE_PASSWORD_LENGTH;

            PasswordCtrl_SetAttribute(panel, passwordCtrl, ATTR_PASSWORD_VAL, "");
            SetCtrlAttribute(panel, passwordCtrl, ATTR_MAX_ENTRY_LENGTH, pwInfo->maxPasswordLength);
            pwInfo->password[pwInfo->maxPasswordLength] = 0;    /* truncate any existing password */
            break;
        case ATTR_PASSWORD_VAL:
            {
            char    *newPasswordVal = (char *)va_arg(parmInfo, char *);

                if (!newPasswordVal)
                    newPasswordVal = "";

                strncpy (pwInfo->password, newPasswordVal, pwInfo->maxPasswordLength);
                UpdateObscuringCharacters(panel, passwordCtrl);
            }
            break;
        case ATTR_PASSWORD_MASK_CHARACTER:
            pwInfo->maskCharacter = va_arg(parmInfo, char);
            if (pwInfo->maskCharacter == 0)
                pwInfo->maskCharacter = DEFAULT_MASK_CHARACTER;

            UpdateObscuringCharacters(panel, passwordCtrl);
            break;
        default:
            errChk(UIEInvalidAttribute);
            break;
        }

    va_end(parmInfo);

Error:
    return error;
}

/*****************************************************************************************/

int CVIFUNC PasswordCtrl_GetAttribute(int panel, int passwordCtrl, int attribute, void *value)
{
    int             error = 0;
    PasswordInfo    pwInfo = 0;

    if (!value)
        errChk(UIENullPointerPassed);

    errChk( GetChainedCallbackData(panel, passwordCtrl, PASSWORD_CTRL_TYPE, (void **)&pwInfo));

    switch (attribute)
        {
        case ATTR_PASSWORD_MAX_LENGTH:
            *(int *)value = pwInfo->maxPasswordLength;
             break;
        case ATTR_PASSWORD_VAL:
            if (value)
                strcpy ((char *)value, pwInfo->password);
            break;
        case ATTR_PASSWORD_MASK_CHARACTER:
            *(char *)value = pwInfo->maskCharacter;
            break;
        default:
            errChk(UIEInvalidAttribute);
            break;
        }

Error:
    return error;
}

/*****************************************************************************************/

int  CVICALLBACK PasswordCtrl_Callback(int panel, int control, int event, void *callbackData, int eventData1, int eventData2)
{
    PasswordInfo    pwInfo = (PasswordInfo)callbackData;
    int             selectionStart, selectionLength, textLength;
    int             entireTextSelected;
    int             swallowEvent = 0;

    GetCtrlAttribute(panel, control, ATTR_TEXT_SELECTION_START, &selectionStart);
    GetCtrlAttribute(panel, control, ATTR_TEXT_SELECTION_LENGTH, &selectionLength);
    GetCtrlAttribute(panel, control, ATTR_STRING_TEXT_LENGTH, &textLength);

    entireTextSelected = textLength > 0 && textLength == selectionLength;

        /* is the entire text is not selected, the cursor must be at the end of the password */
    if (selectionStart != textLength && !entireTextSelected)
        {
        selectionStart = textLength;
        selectionLength = 0;
        SetCtrlAttribute(panel, control, ATTR_TEXT_SELECTION_START, selectionStart);
        SetCtrlAttribute(panel, control, ATTR_TEXT_SELECTION_LENGTH, selectionLength);
        }

    switch (event)
        {
        case EVENT_DISCARD:
            DiscardPasswordInfo(pwInfo);
            break;
        case EVENT_LEFT_CLICK:
        case EVENT_LEFT_DOUBLE_CLICK:
            SetActiveCtrl(panel, control);
            swallowEvent = 1;   /* Swallow clicks that move the cursor from the end */
            break;
        case EVENT_KEYPRESS:
            {
            int virtualKey;
            int asciiCode;

            virtualKey = eventData1 & VAL_VKEY_MASK;
            asciiCode = eventData1 & VAL_ASCII_KEY_MASK;

            if (eventData1 & VAL_MENUKEY_MODIFIER)  /* disallow cut & paste, etc.. */
                {
                swallowEvent = 1;
                break;
                }

            if  (virtualKey != 0)
                {
                switch (virtualKey)
                    {
                    case VAL_BACKSPACE_VKEY:  /* Check for Backspace Key */
                        if (entireTextSelected)
                            pwInfo->password[0] = 0;        /* delete whole password */
                        else
                        if (selectionStart > 0)
                          pwInfo->password[selectionStart-1] = '\0';    /* Remove one character from password */
                        break;
                    case VAL_FWD_DELETE_VKEY:
                    case VAL_INSERT_VKEY:
                    case VAL_UP_ARROW_VKEY:
                    case VAL_DOWN_ARROW_VKEY:
                    case VAL_LEFT_ARROW_VKEY:
                    case VAL_HOME_VKEY:
                        swallowEvent = 1;                   /* swallow keys that move the cursor from the end */
                        break;
                    case VAL_RIGHT_ARROW_VKEY:
                    case VAL_END_VKEY:
                        swallowEvent = !entireTextSelected;     /* no point in allowing unless all text is selected */
                        break;
                    }
                }
            else
            if (asciiCode)
                {
                if (selectionStart < pwInfo->maxPasswordLength)  /* Check password index */
                    {
                    if (entireTextSelected)
                        pwInfo->password[0] = 0;    /* type over entire password since it was selected */

                    Fmt(pwInfo->password,"%s[a]<%c", eventData1);   /* Store the entered keystroke  */
                    *(int*)(eventData2) = pwInfo->maskCharacter;
                    break;
                    }
                else                                        /* length limit reached */
                    {
                    Beep();                                 /* Beep to notify */
                    swallowEvent = 1;                       /* Return a non-zero to "swallow" KEYPRESS */
                    }
                }
            }
            break;
        }

    return swallowEvent;
}
