/*************************************************/
/* scroll.c - Source file for scroll bar control */
/*************************************************/


#include <userint.h>
#include <ansi_c.h>
#include "scroll.h"


/*** The following declarations were moved here from scrlutil.h: ***/
/* which part is being tracked */
enum {kHitUpButton, kHitDownButton, kHitPageUp, kHitPageDown, kHitThumb, kNoHit};

/* delay in milliseconds before repeating page up, up arrow, etc. */
#define kInitialDelay    500

/* flags for scroll bar */
#define kVertical           0x0001 /* set if scroll bar is vertical */
#define kHasPropThumb       0x0002 /* has proportional thumb */

/* structure for scroll bar */
typedef struct
    {
    int                     flags;
    int                     panel;
    int                     graph;              /* ctrlId of graph control */
    int                     height;
    int                     width;
    int                     color;
    unsigned int            viewSize;
    unsigned int            docMin;
    unsigned int            docMax;
    unsigned int            value;
    int                     thumbPlotHandles[5];/* save these so we can optimize and redraw thumb movements quickly */
    int                     thumbTop;           /* used to determine if user clicked on thumb */
    int                     thumbBottom;        /* used to determine if user clicked on thumb */
    CtrlCallbackPtr         userCallback;       /* users callback function */
    int                     trackingPart;       /* part of thumb that user clicked */
    DeferredCallbackPtr     deferredCallback;   /* used to update scrollbar when mouse is down */
    void                    *callbackData;      /* users callback data */
    double                  pixelRange;         /* used to convert click to document units */
    double                  pixelTop;           /* used to convert click to document units */
    int                     thumbClickOffset;   /* how many pixels from the middle of the thumb the user clicked */
    clock_t                 trackStartTime;     /* used to setup delay before repeat */
} ScrollBarType;


int InitScrollBar(int panel, int graph, int vertical,CtrlCallbackPtr callbackFunction,
                void *callbackData, int height, int width, int color);
int DrawScrollBar(int panel, int graph);
ScrollBarType *GetScrollBarPtr(int panel, int graph);
void GetHiliteAndShade(int color, int *hilite, int *shade);
int DrawButtonArrow(int panel, int graph, ScrollBarType *scrollBarPtr, int which);
int DrawScrollBarThumb(int panel, int graph);
int CVICALLBACK GraphCallback(int panel, int control, int event, void *data, int eventData1, int eventData2);
int DoHitTest(int panel, int graph);
void UpdateScrollBarPosition(int panel, int graph, int firstHit);
void CVICALLBACK ScrollBarDeferredCall(void *scrollBarPtr);
int DrawDownButtonPressed(int panel, int graph);
int DrawUpButtonPressed(int panel, int graph);
int GetThumbClickOffsetFromMiddle(int panel, int graph);
/*** end of declarations were moved from scrlutil.h ***/




/******************************************************************************/
/* Function: ScrollBar_Create                                                 */
/* Purpose:  Create a new scroll bar ctrl at the specifed location with       */
/*           callback function and callback data passed in                    */
/******************************************************************************/

int CVIFUNC ScrollBar_Create(int panel, int top, int left, int vertical, CtrlCallbackPtr callbackFunction, void *callbackData)
    {
    int graph = 0;

    graph = NewCtrl(panel, CTRL_GRAPH, "scroll bar", top, left);
    if (!graph)
        return SB_ERROR;

    if (InitScrollBar(panel, graph, vertical, callbackFunction,
                callbackData, vertical ? 120 : 16, vertical ? 16 : 120, VAL_LT_GRAY) != SB_NO_ERROR)
        {
        DiscardCtrl(panel, graph);
        return SB_ERROR;
        }
    if (DrawScrollBar(panel, graph) != SB_NO_ERROR)
        {
        DiscardCtrl(panel, graph);
        return SB_ERROR;
        }

    return graph;
    }


/******************************************************************************/
/* Function: ScrollBar_ConvertFromGraph                                       */
/* Purpose:  Convert the graph control passed in to a scroll bar control.     */
/*           Keep the graph control's position, size and color                */
/******************************************************************************/

int CVIFUNC ScrollBar_ConvertFromGraph(int panel, int graph, CtrlCallbackPtr callbackFunction, void *callbackData)
    {
    int height, width, style, vertical, color;

    GetCtrlAttribute(panel, graph, ATTR_CTRL_STYLE, &style);

    if (style != CTRL_GRAPH)
        return SB_ERROR_NOT_GRAPH;

        /*  If zero is passed for the callback function, use the graph's current callback
            function as the callback function.
        */
    if (callbackFunction == 0)
        GetCtrlAttribute(panel, graph, ATTR_CALLBACK_FUNCTION_POINTER, &callbackFunction);

    GetCtrlAttribute(panel, graph, ATTR_WIDTH, &width);
    GetCtrlAttribute(panel, graph, ATTR_HEIGHT, &height);
    GetCtrlAttribute(panel, graph, ATTR_PLOT_BGCOLOR, &color);
    vertical = height < width ? 0 : 1;
    if (vertical)
        {
        if (width < 16) width = 16;
        if (height < 64) height = 64;
        }
    else
        {
        if (height < 16) height = 16;
        if (width < 64) width = 64;
        }

    if (InitScrollBar(panel, graph, vertical, callbackFunction,
            callbackData, height, width, color) != SB_NO_ERROR)
        {
        DiscardCtrl(panel, graph);
        return SB_ERROR;
        }

    if (DrawScrollBar(panel, graph) != SB_NO_ERROR)
        {
        DiscardCtrl(panel, graph);
        return SB_ERROR;
        }

    return graph;
    }


/******************************************************************************/
/* Function: DiscardScrollBarCtrl                                             */
/* Purpose:  Discard the scroll bar's graph control and free its data         */
/******************************************************************************/

static int DiscardScrollBarCtrl(int panel, int graph)
    {
    ScrollBarType   *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    free(scrollBarPtr);
    SetCtrlAttribute(panel, graph, ATTR_CALLBACK_DATA, (void *)0);

    return SB_NO_ERROR;
    }


/******************************************************************************/
/* Function: ScrollBar_SetAttribute                                           */
/* Purpose:  Set the specified attribute to the value passed into the function*/
/******************************************************************************/

int CVIFUNC ScrollBar_SetAttribute(int panel, int graph, int attribute, int value)
    {
    int refresh = 0, refreshThumb = 0, error = SB_NO_ERROR;
    ScrollBarType *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);

    switch (attribute)
        {
        case ATTR_SB_PROP_THUMB:
            {
            int isProp;

            if (value != 0)
                value = 1;
            isProp = scrollBarPtr->flags & kHasPropThumb ? 1: 0;
            if (isProp != value)
                {
                if (value)
                    scrollBarPtr->flags |= kHasPropThumb;
                else
                    scrollBarPtr->flags &= ~kHasPropThumb;
                refreshThumb = 1;
                }
             }
             break;
        case ATTR_SB_VERTICAL:
            {
            int newWidth, newHeight, vertical;

            if (value != 0)
                value = 1;
            vertical = scrollBarPtr->flags & kVertical ? 1: 0;
            if (vertical != value)
                {
                if (value)
                    scrollBarPtr->flags |= kVertical;
                else
                    scrollBarPtr->flags &= ~kVertical;
                newWidth = scrollBarPtr->height;
                newHeight = scrollBarPtr->width;
                scrollBarPtr->width = newWidth;
                SetCtrlAttribute(panel, graph, ATTR_WIDTH, newWidth);
                SetAxisRange (panel, graph, VAL_MANUAL, 0.0, (double)(newWidth-3), VAL_NO_CHANGE, 0, 0);
                scrollBarPtr->height = newHeight;
                SetCtrlAttribute(panel, graph, ATTR_HEIGHT, newHeight);
                SetAxisRange (panel, graph, VAL_NO_CHANGE, 0, 0, VAL_MANUAL, 1, (double)(newHeight-2));
                refresh = 1;
                }
            }
            break;
        case ATTR_SB_COLOR:
            if (value > 0xffffff)
                error = SB_ERROR_BAD_VALUE;
            else if (value < 0)
                error = SB_ERROR_BAD_VALUE;
            else
                {
                scrollBarPtr->color = value;
                refresh = 1;
                }
            break;
        case ATTR_SB_WIDTH:
            {
            int minimum = scrollBarPtr->flags & kVertical ? 16 : 64;

            if (value > 32767)
                error = SB_ERROR_BAD_VALUE;
            else if (value < minimum)
                error = SB_ERROR_BAD_VALUE;
            else
                {
                scrollBarPtr->width = value;
                SetCtrlAttribute(panel, graph, ATTR_WIDTH, value);
                SetAxisRange (panel, graph, VAL_MANUAL, 0.0, (double)(value-3), VAL_NO_CHANGE, 0, 0);
                refresh = 1;
                }
            }
            break;
        case ATTR_SB_HEIGHT:
            {
            int minimum = scrollBarPtr->flags & kVertical ? 64 : 16;

            if (value > 32767)
                error = SB_ERROR_BAD_VALUE;
            else if (value < minimum)
                error = SB_ERROR_BAD_VALUE;
            else
                {
                scrollBarPtr->height = value;
                SetCtrlAttribute(panel, graph, ATTR_HEIGHT, value);
                SetAxisRange (panel, graph, VAL_NO_CHANGE, 0, 0, VAL_MANUAL, 1, (double)(value-2));
                refresh = 1;
                }
            }
            break;
        case ATTR_SB_VIEW_SIZE:
            scrollBarPtr->viewSize = (unsigned int)value;
            refreshThumb = 1;
            break;
        case ATTR_SB_DOC_MIN:
            if ((unsigned)value >= scrollBarPtr->docMax)
                error = SB_ERROR_BAD_VALUE;
            else
                {
                scrollBarPtr->docMin = (unsigned int)value;
                refreshThumb = 1;
                }
            break;
        case ATTR_SB_DOC_MAX:
            if ((unsigned)value <= scrollBarPtr->docMin)
                error = SB_ERROR_BAD_VALUE;
            else
                {
                scrollBarPtr->docMax = (unsigned int)value;
                refreshThumb = 1;
                }
            break;
        case ATTR_SB_VALUE:
            if ((unsigned)value < scrollBarPtr->docMin)
                error = SB_ERROR_BAD_VALUE;
            else if ((unsigned)value > scrollBarPtr->docMax)
                error = SB_ERROR_BAD_VALUE;
            scrollBarPtr->value = (unsigned int)value;
            refreshThumb = 1;
            break;
        default:
            error = SB_ERROR_BAD_ATTRIBUTE;
            break;
        }
    if (refresh)
        DrawScrollBar(panel, graph);
    else if (refreshThumb)
        DrawScrollBarThumb(panel, graph);

    return error;
    }

/******************************************************************************/
/* Function: ScrollBar_GetAttribute                                           */
/* Purpose:  Get the value of the attribute passed into the function          */
/******************************************************************************/

int CVIFUNC ScrollBar_GetAttribute(int panel, int graph, int attribute, int *value)
    {
    int error = SB_NO_ERROR;
    ScrollBarType *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);

    switch (attribute)
        {
        case ATTR_SB_PROP_THUMB:
            *value = scrollBarPtr->flags & kHasPropThumb ? 1 : 0;
            break;
        case ATTR_SB_VERTICAL:
            *value = scrollBarPtr->flags & kVertical ? 1 : 0;
            break;
        case ATTR_SB_COLOR:
            *value = scrollBarPtr->color;
            break;
        case ATTR_SB_WIDTH:
            *value = scrollBarPtr->width;
            break;
        case ATTR_SB_HEIGHT:
            *value = scrollBarPtr->height;
            break;
        case ATTR_SB_VIEW_SIZE:
            *value = scrollBarPtr->viewSize;
            break;
        case ATTR_SB_DOC_MIN:
            *value = scrollBarPtr->docMin;
            break;
        case ATTR_SB_DOC_MAX:
            *value = scrollBarPtr->docMax;
            break;
        case ATTR_SB_VALUE:
            *value = scrollBarPtr->value;
            break;
        default:
            error = SB_ERROR_BAD_ATTRIBUTE;
            break;
        }

    return error;
    }






/*****************************************************************************/
/* The Following Functions provide support for the Scroll Bar Implementation */
/* They were move here from scrlutil.c.                                      */
/*****************************************************************************/

/* private functions */
static int HandleScrollBarEvent(int panel, int graph, int change, int event);

/* pointer to scroll bar data structure is stored in graph controls callback data */
ScrollBarType *GetScrollBarPtr(int panel, int graph)
    {
    void *ptr;

    GetCtrlAttribute(panel, graph, ATTR_CALLBACK_DATA, &ptr);

    return (ScrollBarType *)ptr;
    }

int InitScrollBar(int panel, int graph, int vertical,
                CtrlCallbackPtr callbackFunction, void *callbackData, int height, int width, int color)
    {
    ScrollBarType   *scrollBarPtr;

    /* init graph */
    SetCtrlAttribute(panel, graph, ATTR_BORDER_VISIBLE, 0);
    SetCtrlAttribute(panel, graph, ATTR_EDGE_STYLE, VAL_FLAT_EDGE);
    SetCtrlAttribute(panel, graph, ATTR_PLOT_BGCOLOR, VAL_BLACK);
    SetCtrlAttribute(panel, graph, ATTR_XGRID_VISIBLE, 0);
    SetCtrlAttribute(panel, graph, ATTR_XLABEL_VISIBLE, 0);
    SetCtrlAttribute(panel, graph, ATTR_YGRID_VISIBLE, 0);
    SetCtrlAttribute(panel, graph, ATTR_XLABEL_VISIBLE, 0);
    SetCtrlAttribute(panel, graph, ATTR_LABEL_VISIBLE, 1);
    SetCtrlAttribute(panel, graph, ATTR_LABEL_TEXT, "scroll bar");
    SetCtrlAttribute(panel, graph, ATTR_LABEL_TOP, -1000);
    SetCtrlAttribute(panel, graph, ATTR_CTRL_MODE, VAL_HOT);
    SetCtrlAttribute(panel, graph, ATTR_NUM_CURSORS, 1);
    SetCtrlAttribute(panel, graph, ATTR_SMOOTH_UPDATE, 1);
    SetCursorAttribute (panel, graph, 1, ATTR_CROSS_HAIR_STYLE, VAL_NO_CROSS);
    SetCursorAttribute (panel, graph, 1, ATTR_CURSOR_POINT_STYLE, VAL_NO_POINT);
    SetCursorAttribute (panel, graph, 1, ATTR_CURSOR_COLOR, VAL_TRANSPARENT);
    scrollBarPtr = malloc(sizeof(ScrollBarType));
    InstallCtrlCallback(panel, graph, GraphCallback, scrollBarPtr);

    scrollBarPtr->flags |= kHasPropThumb;
    if (vertical)
        scrollBarPtr->flags |= kVertical;
    scrollBarPtr->height = height;
    SetCtrlAttribute(panel, graph, ATTR_HEIGHT, height);
    SetAxisRange (panel, graph, VAL_NO_CHANGE, 0, 0, VAL_MANUAL, 1, (double)(height-2));
    scrollBarPtr->width = width;
    SetCtrlAttribute(panel, graph, ATTR_WIDTH, width);
    SetAxisRange (panel, graph, VAL_MANUAL, 0.0, (double)(width-3), VAL_NO_CHANGE, 0, 0);

    scrollBarPtr->userCallback = callbackFunction;
    scrollBarPtr->callbackData = callbackData;
    scrollBarPtr->graph = graph;
    scrollBarPtr->panel = panel;
    scrollBarPtr->color = color;
    scrollBarPtr->viewSize = 10;
    scrollBarPtr->docMin = 0;
    scrollBarPtr->docMax = 100;
    scrollBarPtr->value = 0;
    scrollBarPtr->trackingPart = kNoHit;
    scrollBarPtr->deferredCallback = &ScrollBarDeferredCall;

    return SB_NO_ERROR;
    }

enum {kUp, kDown, kLeft, kRight};

/* draw scroll bar button arrow in direction specifed by which */
int DrawButtonArrow(int panel, int graph, ScrollBarType *scrollBarPtr, int which)
    {
    int top=0, left=0, length, x1Array[7], x2Array[7], yArray[7], i;

    length = 2;
    x1Array[0] = left;
    x2Array[0] = left + length;
    yArray[0] = top;
    top++;
    x1Array[1] = left;
    x2Array[1] = left + length;
    yArray[1] = top;
    top++;
    x1Array[2] = left;
    x2Array[2] = left + length;
    yArray[2] = top;
    top++;
    length += 4;
    left -= 2;
    x1Array[3] = left;
    x2Array[3] = left + length;
    yArray[3] = top;
    top++;
    length -= 2;
    left += 1;
    x1Array[4] = left;
    x2Array[4] = left + length;
    yArray[4] = top;
    top++;
    length -= 2;
    left += 1;
    x1Array[5] = left;
    x2Array[5] = left + length;
    yArray[5] = top;
    top++;
    length -= 2;
    left += 1;
    x1Array[6] = left;
    x2Array[6] = left + length;
    yArray[6] = top;

    if (which == kUp)
        {
        top = scrollBarPtr->height -11;
        left = scrollBarPtr->width / 2 - 3;
        for (i=0; i<=6; i++)
            if (PlotLine(panel, graph, left+x1Array[i], top+yArray[i], left+x2Array[i], top+yArray[i], VAL_BLACK) < 0)
                return SB_ERROR;
        }
    else if (which == kRight)
        {
        top = scrollBarPtr->height / 2 + 1;
        left = scrollBarPtr->width - 13;
        for (i=6; i>=0; i--)
            if (PlotLine(panel, graph, left+yArray[i], top-x1Array[i], left+yArray[i], top-x2Array[i], VAL_BLACK) < 0)
                return SB_ERROR;
        }
    else if (which == kLeft)
        {
        top = scrollBarPtr->height / 2 + 1;
        left = 9;
        for (i=6; i>=0; i--)
            if (PlotLine(panel, graph, left-yArray[i], top-x1Array[i], left-yArray[i], top-x2Array[i], VAL_BLACK) < 0)
                return SB_ERROR;
        }
    else if (which == kDown)
        {
        top = 11;
        left = scrollBarPtr->width / 2 - 3;
        for (i=6; i>=0; i--)
            if (PlotLine(panel, graph, left+x1Array[i], top-yArray[i], left+x2Array[i], top-yArray[i], VAL_BLACK) < 0)
                return SB_ERROR;
        }

    return SB_NO_ERROR;
    }

#define kThumbMiddleMinimum ((vertical ? 16 : 15) + halfThumbHeight)
#define kThumbMiddleMaximum ((vertical ? (scrollBarPtr->height -17) : (scrollBarPtr->width - 18)) - halfThumbHeight)

int DrawScrollBar(int panel, int graph)
    {
    ScrollBarType *scrollBarPtr;
    int hilite, shade, i, vertical;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    vertical = scrollBarPtr->flags & kVertical ? 1 : 0;
    GetHiliteAndShade(scrollBarPtr->color, &hilite, &shade);

    SetCtrlAttribute(panel, graph, ATTR_REFRESH_GRAPH, 0);
    DeleteGraphPlot(panel, graph, -1, VAL_NO_DRAW);
    for (i=0; i<=4; i++)
        scrollBarPtr->thumbPlotHandles[i] = 0;

    /* blank out scroll bar */
    if (PlotRectangle(panel, graph, 0, 0, scrollBarPtr->width, scrollBarPtr->height, scrollBarPtr->color, scrollBarPtr->color) < 0)
        return SB_ERROR;

    /* draw bottom(left) button */
    if (vertical)
        {
        if (PlotLine(panel, graph, 0, 15, scrollBarPtr->width, 15, VAL_BLACK) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 14, scrollBarPtr->width, 14, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, 0, 14, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, scrollBarPtr->width, 1, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, 1, scrollBarPtr->width-3, 14, shade) < 0)
            return SB_ERROR;
        if (DrawButtonArrow(panel, graph, scrollBarPtr, kDown) == SB_ERROR)
            return SB_ERROR;
        }
    else
        {
        if (PlotLine(panel, graph, 14, 0, 14,  scrollBarPtr->height, VAL_BLACK) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 0, 0, scrollBarPtr->height, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-2, 13, scrollBarPtr->height-2, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 13, 0, 13, scrollBarPtr->height, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, 13, 1, shade) < 0)
            return SB_ERROR;
        if (DrawButtonArrow(panel, graph, scrollBarPtr, kLeft) == SB_ERROR)
            return SB_ERROR;
        }

    /* draw top(right) button */
    if (vertical)
        {
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-16, scrollBarPtr->width, scrollBarPtr->height-16, VAL_BLACK) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-2, scrollBarPtr->width, scrollBarPtr->height-2, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height, 0, scrollBarPtr->height-15, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-15, scrollBarPtr->width, scrollBarPtr->height-15, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, scrollBarPtr->height, scrollBarPtr->width-3, scrollBarPtr->height-15, shade) < 0)
            return SB_ERROR;
        if (DrawButtonArrow(panel, graph, scrollBarPtr, kUp) == SB_ERROR)
            return SB_ERROR;
        }
    else
        {
        if (PlotLine(panel, graph, scrollBarPtr->width-17, 0, scrollBarPtr->width-17,  scrollBarPtr->height, VAL_BLACK) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-16, 0, scrollBarPtr->width-16, scrollBarPtr->height, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-16, scrollBarPtr->height-2, scrollBarPtr->width, scrollBarPtr->height-2, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, 0, scrollBarPtr->width-3, scrollBarPtr->height, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-16, 1, scrollBarPtr->width, 1, shade) < 0)
            return SB_ERROR;
        if (DrawButtonArrow(panel, graph, scrollBarPtr, kRight) == SB_ERROR)
            return SB_ERROR;
        }

    /* draw thumb */
    if (DrawScrollBarThumb(panel, graph) == SB_ERROR)
        return SB_ERROR;

    SetCtrlAttribute(panel, graph, ATTR_REFRESH_GRAPH, 1);

    return SB_NO_ERROR;
    }

int DrawScrollBarThumb(int panel, int graph)
    {
    ScrollBarType *scrollBarPtr;
    int hilite, shade, i, vertical;
    double viewRange, docRange, halfThumbHeight, thumbMiddle, docRatio;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    vertical = scrollBarPtr->flags & kVertical ? 1 : 0;
    GetHiliteAndShade(scrollBarPtr->color, &hilite, &shade);

    SetCtrlAttribute(panel, graph, ATTR_REFRESH_GRAPH, 0);

    for (i=0; i<=4; i++)
        if (scrollBarPtr->thumbPlotHandles[i])
            {
            DeleteGraphPlot(panel, graph, scrollBarPtr->thumbPlotHandles[i], VAL_NO_DRAW);
            scrollBarPtr->thumbPlotHandles[i] = 0;
            }
    if (vertical)
        PlotRectangle(panel, graph, 0, 15, scrollBarPtr->width-2, scrollBarPtr->height - 17, scrollBarPtr->color, scrollBarPtr->color);
    else
        PlotRectangle(panel, graph, 15, 0, scrollBarPtr->width-17, scrollBarPtr->height - 2, scrollBarPtr->color, scrollBarPtr->color);
    viewRange = scrollBarPtr->viewSize;
    docRange = scrollBarPtr->docMax - scrollBarPtr->docMin;
    if (viewRange > docRange)
        viewRange = docRange;
    scrollBarPtr->thumbTop = -1;
    scrollBarPtr->thumbBottom = -1;
    if ((docRange != 0) && (viewRange != 0) /*&& (viewRange < docRange) */)
        {
        if (vertical)
            scrollBarPtr->pixelRange = scrollBarPtr->height-33;
        else
            scrollBarPtr->pixelRange = scrollBarPtr->width-33;
        if (scrollBarPtr->flags & kHasPropThumb)
            {
            halfThumbHeight = viewRange / docRange * scrollBarPtr->pixelRange / 2;
            halfThumbHeight = (double)(int)(halfThumbHeight + 0.5); /* round to nearest int */
            if (halfThumbHeight < 8) /* make thumb minimum size */
                halfThumbHeight = 8;
            }
        else
            halfThumbHeight = 8;
        scrollBarPtr->pixelRange -= (2*halfThumbHeight);
        if (vertical)
            scrollBarPtr->pixelTop = 17+halfThumbHeight;
        else
            scrollBarPtr->pixelTop = 15+halfThumbHeight;
        if (scrollBarPtr->value > scrollBarPtr->docMin)
            docRatio = ((scrollBarPtr->value - scrollBarPtr->docMin) / docRange);
        else
            docRatio = 0;
        if (vertical)
            thumbMiddle = scrollBarPtr->height - scrollBarPtr->pixelTop - (docRatio * scrollBarPtr->pixelRange);
        else
            thumbMiddle = scrollBarPtr->pixelTop + (docRatio * scrollBarPtr->pixelRange);
        thumbMiddle = (double)(int)(thumbMiddle + 0.5); /* round to nearest int */
        if (thumbMiddle < kThumbMiddleMinimum) /* make sure thumb is in range */
            thumbMiddle = kThumbMiddleMinimum;
        if (thumbMiddle > kThumbMiddleMaximum)
            thumbMiddle = kThumbMiddleMaximum;
        scrollBarPtr->thumbTop = (int)(thumbMiddle + halfThumbHeight);
        scrollBarPtr->thumbBottom = (int)(thumbMiddle - halfThumbHeight);
        if (vertical)
            scrollBarPtr->thumbPlotHandles[0] = PlotRectangle(panel, graph, 0, thumbMiddle - halfThumbHeight, scrollBarPtr->width, thumbMiddle + halfThumbHeight, scrollBarPtr->color, scrollBarPtr->color);
        else
            scrollBarPtr->thumbPlotHandles[0] = PlotRectangle(panel, graph, thumbMiddle - halfThumbHeight, 0, thumbMiddle + halfThumbHeight, scrollBarPtr->height, scrollBarPtr->color, scrollBarPtr->color);
        if (scrollBarPtr->thumbPlotHandles[0] < 0)
            {
            scrollBarPtr->thumbPlotHandles[0] = 0;
            return SB_ERROR;
            }
        if (vertical)
            scrollBarPtr->thumbPlotHandles[1] = PlotLine(panel, graph, 0, thumbMiddle + halfThumbHeight, scrollBarPtr->width-3, thumbMiddle + halfThumbHeight, hilite);
        else
            scrollBarPtr->thumbPlotHandles[1] = PlotLine(panel, graph, thumbMiddle - halfThumbHeight, 0, thumbMiddle - halfThumbHeight, scrollBarPtr->height-2, hilite);
        if (scrollBarPtr->thumbPlotHandles[1] < 0)
            {
            scrollBarPtr->thumbPlotHandles[1] = 0;
            return SB_ERROR;
            }
        if (vertical)
            scrollBarPtr->thumbPlotHandles[2] = PlotLine(panel, graph, 0, thumbMiddle - halfThumbHeight, 0, thumbMiddle + halfThumbHeight, hilite);
        else
            scrollBarPtr->thumbPlotHandles[2] = PlotLine(panel, graph, thumbMiddle - halfThumbHeight, scrollBarPtr->height-2, thumbMiddle + halfThumbHeight, scrollBarPtr->height-2, hilite);
        if (scrollBarPtr->thumbPlotHandles[2] < 0)
            {
            scrollBarPtr->thumbPlotHandles[2] = 0;
            return SB_ERROR;
            }
        if (vertical)
            scrollBarPtr->thumbPlotHandles[3] = PlotLine(panel, graph, 0, thumbMiddle - halfThumbHeight, scrollBarPtr->width-3, thumbMiddle - halfThumbHeight, shade);
        else
            scrollBarPtr->thumbPlotHandles[3] = PlotLine(panel, graph, thumbMiddle + halfThumbHeight, 0, thumbMiddle + halfThumbHeight, scrollBarPtr->height-2, shade);
        if (scrollBarPtr->thumbPlotHandles[3] < 0)
            {
            scrollBarPtr->thumbPlotHandles[3] = 0;
            return SB_ERROR;
            }
        if (vertical)
            scrollBarPtr->thumbPlotHandles[4] = PlotLine(panel, graph, scrollBarPtr->width-3, thumbMiddle - halfThumbHeight, scrollBarPtr->width-3, thumbMiddle + halfThumbHeight, shade);
        else
            scrollBarPtr->thumbPlotHandles[4] = PlotLine(panel, graph, thumbMiddle - halfThumbHeight, 1, thumbMiddle + halfThumbHeight, 1, shade);
        if (scrollBarPtr->thumbPlotHandles[4] < 0)
            {
            scrollBarPtr->thumbPlotHandles[4] = 0;
            return SB_ERROR;
            }
        }
    SetCtrlAttribute(panel, graph, ATTR_REFRESH_GRAPH, 1);
    return SB_NO_ERROR;
    }

#define RedComponent(c)     ((c>>16) & 0xff)
#define GreenComponent(c)   ((c>>8) & 0xff)
#define BlueComponent(c)    (c & 0xff)

void GetHiliteAndShade(int color, int *hilite, int *shade)
    {
    *hilite = MakeColor((0x300+RedComponent(color))>>2,
                         (0x300+GreenComponent(color))>>2,
                         (0x300+BlueComponent(color))>>2);
    *shade = MakeColor((RedComponent(color)>>2)*3,
                        (GreenComponent(color)>>2)*3,
                        (BlueComponent(color)>>2)*3);
    }

void CVICALLBACK ScrollBarDeferredCall(void *scrollBarPtr)
    {
    double delay;

    delay = (clock() - ((ScrollBarType *)scrollBarPtr)->trackStartTime) * 1000.0 / CLOCKS_PER_SEC;
    if (((ScrollBarType *)scrollBarPtr)->trackingPart != kNoHit)
        {
        if (delay > kInitialDelay)
            UpdateScrollBarPosition(((ScrollBarType *)scrollBarPtr)->panel, ((ScrollBarType *)scrollBarPtr)->graph, ((ScrollBarType *)scrollBarPtr)->trackingPart);
        PostDeferredCall(((ScrollBarType *)scrollBarPtr)->deferredCallback, scrollBarPtr);
        }
    }

int CVICALLBACK GraphCallback(int panel, int graph, int event, void *data, int eventData1, int eventData2)
    {
    ScrollBarType *scrollBarPtr;
    int value;

    scrollBarPtr = GetScrollBarPtr(panel, graph);

    if (event == EVENT_KEYPRESS)
        {
        if ((eventData1 == VAL_TAB_VKEY) || (eventData1 == (VAL_TAB_VKEY | VAL_SHIFT_MODIFIER)))
            return 0;

        value = scrollBarPtr->value;

        if (eventData1 == VAL_UP_ARROW_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->value-1, EVENT_SB_UP_BUTTON);
        else if (eventData1 == VAL_DOWN_ARROW_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->value+1, EVENT_SB_DOWN_BUTTON);
        else if (eventData1 == VAL_PAGE_UP_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->value - (int)scrollBarPtr->viewSize, EVENT_SB_PAGE_UP);
        else if (eventData1 == VAL_PAGE_DOWN_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->value + (int)scrollBarPtr->viewSize, EVENT_SB_PAGE_DOWN);
        else if (eventData1 == VAL_HOME_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->docMin, EVENT_SB_THUMB_CHANGE);
        else if (eventData1 == VAL_END_VKEY)
            HandleScrollBarEvent(panel, graph, scrollBarPtr->docMax, EVENT_SB_THUMB_CHANGE);

        if ((unsigned)value != scrollBarPtr->value)
            {
            value = scrollBarPtr->value;
            if (scrollBarPtr->userCallback)
                (*(scrollBarPtr->userCallback))(panel, graph, EVENT_SB_COMMIT, scrollBarPtr->callbackData, value, (int)&value);
            if ((unsigned)value != scrollBarPtr->value)
                {
                scrollBarPtr->value = value;
                DrawScrollBar(panel, graph);
                }
            }

        return 1;
        }
    else
    if (event == EVENT_VAL_CHANGED)
        {
        if (scrollBarPtr->trackingPart == kNoHit)
            {
            scrollBarPtr->trackStartTime = clock();
            scrollBarPtr->trackingPart = DoHitTest(panel, graph);
            GetThumbClickOffsetFromMiddle(panel, graph);
            if (scrollBarPtr->trackingPart == kHitUpButton)
                DrawUpButtonPressed(panel, graph);
            else if (scrollBarPtr->trackingPart == kHitDownButton)
                DrawDownButtonPressed(panel, graph);
            UpdateScrollBarPosition(panel, graph, scrollBarPtr->trackingPart);
            if (scrollBarPtr->trackingPart != kHitThumb)
                PostDeferredCall(scrollBarPtr->deferredCallback, scrollBarPtr);
            }
        else if (scrollBarPtr->trackingPart == kHitThumb)
            UpdateScrollBarPosition(panel, graph, scrollBarPtr->trackingPart);
        }
    else
    if (event == EVENT_COMMIT)
        {
        SetGraphCursor(panel, graph, 1, 1, 1);
        DrawScrollBar(panel, graph);
        scrollBarPtr->trackingPart = kNoHit;
        value = scrollBarPtr->value;
        if (scrollBarPtr->userCallback)
            (*(scrollBarPtr->userCallback))(panel, graph, EVENT_SB_COMMIT, scrollBarPtr->callbackData, value, (int)&value);
        if ((unsigned)value != scrollBarPtr->value)
            {
            scrollBarPtr->value = value;
            DrawScrollBar(panel, graph);
            }
        }
    else
    if (event == EVENT_DISCARD)
        {
        DiscardScrollBarCtrl(panel, graph);
        }
    return 0;
    }

void UpdateScrollBarPosition(int panel, int graph, int firstHit)
    {
    int hit;
    ScrollBarType *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    hit = DoHitTest(panel, graph);
    if ((hit == firstHit) || (firstHit == kHitThumb))
        {
        switch (firstHit)
            {
            case kHitThumb:
                {
                double x, y, ratio, newDValue, pixelValue;
                unsigned int newUValue;
                int vertical;

                GetGraphCursor(panel, graph, 1, &x, &y);
                vertical = scrollBarPtr->flags & kVertical ? 1 : 0;
                pixelValue = vertical ? y : x;
                pixelValue = (double)(int)(pixelValue + 0.5); /* round first */
                pixelValue -= scrollBarPtr->pixelTop;
                pixelValue += scrollBarPtr->thumbClickOffset;
                if (pixelValue < 0)
                    pixelValue = 0;
                if ( pixelValue > scrollBarPtr->pixelRange)
                    pixelValue = scrollBarPtr->pixelRange;
                if (vertical)
                    ratio = 1.0 - pixelValue/scrollBarPtr->pixelRange;
                else
                    ratio = pixelValue/scrollBarPtr->pixelRange;
                newDValue = scrollBarPtr->docMin + (ratio * (scrollBarPtr->docMax - scrollBarPtr->docMin));
                newUValue = (unsigned int)newDValue;
                if (newUValue != scrollBarPtr->value)
                    HandleScrollBarEvent(panel, graph, newUValue, EVENT_SB_THUMB_CHANGE);
                }
                break;
            case kHitUpButton:
                HandleScrollBarEvent(panel, graph, scrollBarPtr->value - 1, EVENT_SB_UP_BUTTON);
                break;
            case kHitDownButton:
                HandleScrollBarEvent(panel, graph, scrollBarPtr->value + 1, EVENT_SB_DOWN_BUTTON);
                break;
            case kHitPageUp:
                HandleScrollBarEvent(panel, graph, scrollBarPtr->value - (int)scrollBarPtr->viewSize, EVENT_SB_PAGE_UP);
                break;
            case kHitPageDown:
                HandleScrollBarEvent(panel, graph, scrollBarPtr->value + (int)scrollBarPtr->viewSize, EVENT_SB_PAGE_DOWN);
                break;
            case kNoHit:
                break;
            }
        }
    }

int DoHitTest(int panel, int graph)
    {
    double x, y, pixelValue;
    int vertical;
    ScrollBarType *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);

    vertical = scrollBarPtr->flags & kVertical ? 1 : 0;
    GetGraphCursor(panel, graph, 1, &x, &y);
    pixelValue = vertical ? y : x;
    pixelValue = (double)(int)(pixelValue + 0.5); /* round first */
    if ((scrollBarPtr->thumbTop ==  -1) || (scrollBarPtr->thumbBottom == -1))
        return kNoHit;
    else if ((vertical && (pixelValue > (scrollBarPtr->height - 17)))
                || (!vertical && (pixelValue < 15)))
        return kHitUpButton;
    else if ((vertical && (pixelValue < 16))
                || (!vertical && (pixelValue > (scrollBarPtr->width - 18))))
        return kHitDownButton;
    else if (pixelValue > scrollBarPtr->thumbTop)
        return vertical ? kHitPageUp : kHitPageDown;
    else if (pixelValue < scrollBarPtr->thumbBottom)
        return vertical ? kHitPageDown : kHitPageUp;
    else
        return kHitThumb;
    }

int DrawUpButtonPressed(int panel, int graph)
    {
    ScrollBarType *scrollBarPtr;
    int hilite, shade;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    GetHiliteAndShade(scrollBarPtr->color, &hilite, &shade);
    if (scrollBarPtr->flags & kVertical)
        {
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-2, scrollBarPtr->width, scrollBarPtr->height-2, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height, 0, scrollBarPtr->height-15, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-15, scrollBarPtr->width, scrollBarPtr->height-15, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, scrollBarPtr->height, scrollBarPtr->width-3, scrollBarPtr->height-15, hilite) < 0)
            return SB_ERROR;
        }
    else
        {
        if (PlotLine(panel, graph, 0, 0, 0, scrollBarPtr->height, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, scrollBarPtr->height-2, 13, scrollBarPtr->height-2, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 13, 0, 13, scrollBarPtr->height, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, 13, 1, hilite) < 0)
            return SB_ERROR;
        }

    return SB_NO_ERROR;
    }

int DrawDownButtonPressed(int panel, int graph)
    {
    ScrollBarType *scrollBarPtr;
    int hilite, shade;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    GetHiliteAndShade(scrollBarPtr->color, &hilite, &shade);
    if (scrollBarPtr->flags & kVertical)
        {
        if (PlotLine(panel, graph, 0, 14, scrollBarPtr->width, 14, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, 0, 14, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, 0, 1, scrollBarPtr->width, 1, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, 1, scrollBarPtr->width-3, 14, hilite) < 0)
            return SB_ERROR;
        }
    else
        {
        if (PlotLine(panel, graph, scrollBarPtr->width-16, 0, scrollBarPtr->width-16, scrollBarPtr->height, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-16, scrollBarPtr->height-2, scrollBarPtr->width, scrollBarPtr->height-2, shade) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-3, 0, scrollBarPtr->width-3, scrollBarPtr->height, hilite) < 0)
            return SB_ERROR;
        if (PlotLine(panel, graph, scrollBarPtr->width-16, 1, scrollBarPtr->width, 1, hilite) < 0)
            return SB_ERROR;
        }
    return SB_NO_ERROR;
    }

/* need to get and use this value for scrolling with the thumb */
int GetThumbClickOffsetFromMiddle(int panel, int graph)
    {
    double x, y, pixelValue;
    ScrollBarType *scrollBarPtr;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    GetGraphCursor(panel, graph, 1, &x, &y);
    pixelValue = scrollBarPtr->flags & kVertical ? y : x;
    pixelValue = (double)(int)(pixelValue + 0.5); /* round first */

    scrollBarPtr->thumbClickOffset = (int)(((scrollBarPtr->thumbTop - scrollBarPtr->thumbBottom) / 2 + scrollBarPtr->thumbBottom) - pixelValue);

    return SB_NO_ERROR;
    }


static int HandleScrollBarEvent(int panel, int graph, int newValue, int event)
    {
    ScrollBarType *scrollBarPtr;
    double newDValue;
    int swallowed = 0;
    int error = SB_NO_ERROR;

    scrollBarPtr = GetScrollBarPtr(panel, graph);
    newDValue = (double)newValue;
    if (newDValue < scrollBarPtr->docMin)
        newValue = scrollBarPtr->docMin;
    else if (newDValue > scrollBarPtr->docMax)
        newValue = scrollBarPtr->docMax;
    if (scrollBarPtr->userCallback)
        {
        swallowed = (*(scrollBarPtr->userCallback))(panel, graph, event, scrollBarPtr->callbackData, newValue, (int)&newValue);
        }
    if (!swallowed)
        {
        newDValue = (double)newValue;
        if (newDValue < scrollBarPtr->docMin)
            newDValue = scrollBarPtr->docMin;
        else if (newDValue > scrollBarPtr->docMax)
            newDValue = scrollBarPtr->docMax;
        scrollBarPtr->value = (unsigned int)newDValue;
        error = DrawScrollBarThumb(panel, graph);
        }

    return error;
    }

