
#include <utility.h>
#include <ansi_c.h>
#include <userint.h>
#include <toolbox.h>
#include "toolbar.h"


/********************************************************/
/*    Special prototypes from winuser.h, winbase.h      */
/********************************************************/

#if _NI_mswin32_
#define COLOR_BTNTEXT           18
#define COLOR_INFOTEXT          23
#define COLOR_INFOBK            24

__declspec(dllimport) int __stdcall GetSysColor(int nIndex);
__declspec(dllimport) int __stdcall GetVersion(void);
#endif


/********************/
/*      Macros      */
/********************/

#define kToolbarWidth 4096

/**************************************************************************/
/* ASSUMPTION:  All toolbar buttons are square and will fit properly in a */
/*              picture button of the size below.  This assumption is     */
/*              necessary to support the drawing speed optimizations      */
/**************************************************************************/
#define kToolbarButtonSize  24

/* Width of a separator */
#define kSeparatorWidth             7

/* Separation between buttons on the toolbar */
#define kButtonSeparation           -1

/* Frequency of help polling timer when help is not active */
#define kToolbarHelpInactivePollPeriod  2.0

/* Frequency of help polling timer when help is active */
#define kToolbarHelpActivePollPeriod    0.05

/* Period of help removal timer */
#define kToolbarHelpAutoCloseTime       4.0

/* Number of pixels from bottom of control to top of help panel */
#define kToolbarHelpVerticalOffset  8

/* Toolbar background color */
#define kToolbarPanelColor              VAL_PANEL_GRAY

/* Help panel coloring */
#define kToolbarHelpPanelColor          16777088
#define kToolbarHelpTextColor           VAL_BLACK
#define kToolbarHelpTextBackgroundColor VAL_TRANSPARENT

#define kToolButtonTop                  0
#define kToolButtonLeft                 0

/* Default color for command button text */
#define kSystemButtonTextColor          VAL_BLACK

/* Windows 40 UI constants */  
#define kWin40ToolButtonTop             3
#define kWin40ToolButtonLeft            6
#define kWin40ToolButtonSeparation      0
#define kWin40ToolButtonSize            21 
#define kWin40ToolbarBevelHeight        2


/******************/
/*     Types      */
/******************/

typedef enum
{
    kShellWin95,
    kShellWin31,
    kNumShellStyles,
    kShellUninitialized
} WindowsShellStyles;


typedef struct ToolbarItemStruct 
{
    int type;
    int controlId;
    int active;                 /* REAL activation state */
    int currActive;             /* activation state based on drawn or not */
    char *description;
    int callbackType;
    int menuItem;
    void *callbackFunc;
    void *callbackData;
    ToolbarType toolbar;
} ToolbarItem, *ToolbarItemPtr;

typedef struct ToolbarStruct 
{
    ListType items;             /* list of toolbar items */
    int menuBar;                /* 0 if there is no menu bar */
    int parentPanel;            /* panel id of toolbar's parent panel */
    int panelId;                /* panel id of toolbar panel */
    int frameCtrlId;            /* control used as background to draw win 4.0 style bevel */
    int helpPanel;              /* panel id of toolbar's help panel */ 
    int helpText;               /* ctrl id of help panel text msg */
    int helpTimer;              /* ctrl id of the help panel expiration timer */
    int helpEnabled;            /* Is the toolbar help enabled? */
    int helpActive;             /* Is the help active (being displayed) */
    int pollTimer;              /* ctrl id of the help panel polling timer */
    int sizeValid;              /* Is the current toolbar sized to the items? */
    int itemWidth;              /* width of items on the toolbar */
    int conformBitmapColors;    /* TRUE if the toolbar button bitmaps should */
                                /* be conformed to the toolbar colors */
} ToolbarRec, **ToolbarPtr;


/****************************************/
/*     Private function prototypes      */
/****************************************/

static int  ToolbarIsVisible (ToolbarType toolbar);
static int  NextItemIsButton (ListType itemList, int i);
static int  GetItemIndexFromPosition (ToolbarType toolbar, int xPosition);
static int  CreateToolbarHelpPanel (ToolbarType toolbar);
static int DiscardToolbarResources (ToolbarType toolbar);
static ToolbarItemPtr CopyToolbarItem (ToolbarItemPtr toolbarItem);
static void DisposeToolbarItem (ToolbarItemPtr toolbarItem);
static ListType CopyToolbarItemList (ListType itemList);
static void DisposeToolbarItemList (ListType itemList);
static int  SizeToolbarToItems (ToolbarType toolbar);
static int  SetPollTimer (ToolbarType toolbar, int enabled, int active,
                          int resetTimer);
static void SetCurrentHelpItem (ToolbarItemPtr itemPtr, ToolbarType toolbar);
static ToolbarItemPtr CurrentHelpItem (void);
static int  CVICALLBACK PollTimerCallback (int panel, int control, int event,
                               void *callbackData, int eventData1,
                               int eventData2);
static int  CVICALLBACK HelpTimerCallback (int panel, int control, int event,
                                    void *callbackData, int eventData1,
                                    int eventData2);
static int  CVICALLBACK HelpPanelCallback (int panel, int event, void *callbackData,
                               int eventData1, int eventData2);
static int  HideToolbarHelp (ToolbarType toolbar, int keepTracking,
                             int rememberCurrentItem);
static int  DisplayToolbarHelp (ToolbarItemPtr toolbarItem, int yPosition);
static int  CVICALLBACK ToolbarPanelCallback (int panel, int event, void *callbackData,
                                  int eventData1, int eventData2);
static int  CVICALLBACK GenericToolbarItemCallbackFunc (int panel, int control, int event,
                                            void *callbackData, int eventData1,
                                            int eventData2);
static int  ToolbarItemPtrCmpFn (void *itemA, void *itemB);
static void CoerceExclusiveToggleStates (ToolbarItemPtr toolbarItem);
static int  CreateNewToolbarItemCtrl (ToolbarItemPtr toolbarItem, char imageFile[]);

static int  GetCurrentShell (void);
static int  GetToolbarButtonSize (void);
static int  GetToolbarButtonTop (void);
static int  GetToolbarButtonLeft (void);
static int  GetToolbarButtonSeparation (void);
static int  GetToolbarPanelColor (void);
static int  GetWin313dObjectColor (void);
static int  GetSystemTextColor (void);
static int  GetToolbarHelpBGColor (void);
static int  GetToolbarHelpTextColor (void);
static int  WinColorToCVIColor(int c);


/****************************************/
/*              Functions               */
/****************************************/


static int ToolbarIsVisible (ToolbarType toolbar)
{
    int error = -UIENoError;
    int visible = 0;

    if (!toolbar)
        goto Error;

    errChk (GetPanelAttribute (toolbar->panelId, ATTR_VISIBLE, &visible));

Error :
    return visible;
}


static int CreateToolbarHelpPanel (ToolbarType toolbar)
{
    int error = -UIENoError;
    int tmpPanel = 0;

    if (!toolbar)
        return UIEHandleInvalid;

    errChk (toolbar->helpPanel = NewPanel (toolbar->parentPanel, "Toolbar Help Panel", 0, 0, 0, 0));

    errChk (SetPanelAttribute (toolbar->helpPanel, ATTR_FRAME_STYLE, VAL_OUTLINED_FRAME));
    errChk (SetPanelAttribute (toolbar->helpPanel, ATTR_ACTIVATE_WHEN_CLICKED_ON, 0));
    errChk (SetPanelAttribute (toolbar->helpPanel, ATTR_BACKCOLOR, GetToolbarHelpBGColor()));
    errChk (SetPanelAttribute (toolbar->helpPanel, ATTR_TITLEBAR_VISIBLE, 0));
    errChk (SetPanelAttribute (toolbar->helpPanel, ATTR_SIZABLE, 0));

    errChk (toolbar->helpText = NewCtrl (toolbar->helpPanel, CTRL_TEXT_MSG, "", 0,0));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpText,
                              ATTR_TEXT_BOLD, 0));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpText,
                              ATTR_TEXT_POINT_SIZE, 12));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpText,
                              ATTR_TEXT_COLOR, GetToolbarHelpTextColor()));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpText,
                              ATTR_TEXT_BGCOLOR, kToolbarHelpTextBackgroundColor));

    errChk (toolbar->helpTimer = NewCtrl (toolbar->helpPanel, CTRL_TIMER,
                                          "Help Timer", 0, 0));
    errChk (InstallCtrlCallback (toolbar->helpPanel, toolbar->helpTimer,
                                 HelpTimerCallback, (void *)toolbar));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpTimer,
                              ATTR_INTERVAL, kToolbarHelpAutoCloseTime));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpTimer,
                              ATTR_ENABLED, FALSE));

    errChk (toolbar->pollTimer = NewCtrl (toolbar->helpPanel, CTRL_TIMER,
                                          "Help Poll Timer", -150, -150));
    errChk (InstallCtrlCallback (toolbar->helpPanel, toolbar->pollTimer,
                                 PollTimerCallback, (void *)toolbar));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpTimer,
                              ATTR_INTERVAL, kToolbarHelpInactivePollPeriod));
    toolbar->helpEnabled = FALSE;
    toolbar->helpActive = FALSE;
    errChk (SetPollTimer (toolbar, FALSE, FALSE, TRUE));

Error :
    if (error < 0) {
        if (tmpPanel)
            DiscardPanel (tmpPanel);
        if (toolbar->helpPanel) {
            DiscardPanel (toolbar->helpPanel);
            toolbar->helpPanel = 0;
            toolbar->helpText = 0;
            toolbar->helpTimer = 0;
            toolbar->pollTimer = 0;
        }
    }

    return error;
}


int CVIFUNC Toolbar_New (int parentPanel, int menuBar, char title[],
                         int top, int left, int conformBitmapColors,
                         int conformToSystem, ToolbarType *toolbar)
{
    int error = -UIENoError;
    ToolbarType newToolbar = NULL;
    int menuVisible, menuHeight;
    int panelColor, titleTextColor;
    int bole;
    
    if (!parentPanel) {
        MessagePopup ("Toolbar API Warning",
                      "Toolbars as top-level panels not supported yet.\n"
                      "This would be a fairly easy enhancement.");
        return 0;
    }

    nullChk (newToolbar = malloc (sizeof (ToolbarRec)));
    newToolbar->parentPanel = parentPanel;
    newToolbar->menuBar = menuBar;
    errChk (GetPanelAttribute (newToolbar->parentPanel, ATTR_MENU_BAR_VISIBLE,
                               &menuVisible));
    if (menuVisible) {
        errChk (GetPanelAttribute (newToolbar->parentPanel, ATTR_MENU_HEIGHT,
                                   &menuHeight));
        top +=  menuHeight;
    }
    newToolbar->items = NULL;
    newToolbar->sizeValid = 0;
    newToolbar->conformBitmapColors = conformBitmapColors;
    newToolbar->helpEnabled = 0;
    newToolbar->helpActive = 0;
    errChk (CreateToolbarHelpPanel (newToolbar));
    errChk (newToolbar->panelId = NewPanel (parentPanel, title, top, left,
                                            kToolbarButtonSize, 0));
    nullChk (newToolbar->items = ListCreate (sizeof (ToolbarItemPtr)));

    errChk (SetPanelAttribute (newToolbar->panelId, ATTR_ACTIVATE_WHEN_CLICKED_ON, 0));
    if (!title || !(*title)) {
        errChk (SetPanelAttribute (newToolbar->panelId, ATTR_TITLEBAR_VISIBLE, 0));
    }
    errChk (SetPanelAttribute (newToolbar->panelId, ATTR_SIZABLE, 0));
    
    if (conformToSystem)
    {
        errChk (SetPanelAttribute (newToolbar->panelId, 
                                   ATTR_CONFORM_TO_SYSTEM, 1));
    }
    else
    {
        /* Get the color attributes of the parent panel */
        errChk (GetPanelAttribute (parentPanel, ATTR_BACKCOLOR, &panelColor));
        bole = SetBOLE (0);
        if (GetPanelAttribute (parentPanel, ATTR_TITLE_COLOR, &titleTextColor) < 0)
            titleTextColor = GetSystemTextColor();
        SetBOLE (bole);     
        errChk (GetPanelAttribute (parentPanel, ATTR_CONFORM_TO_SYSTEM,
                                   &conformToSystem));
        
        /* Copy the color attributes to the toolbar panel */                                   
        errChk (SetPanelAttribute (newToolbar->panelId, ATTR_BACKCOLOR, 
                                   panelColor));
        errChk (SetPanelAttribute (newToolbar->panelId, ATTR_TITLE_COLOR, 
                                   titleTextColor));
    }
    
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
    {
        errChk (newToolbar->frameCtrlId = NewCtrl (newToolbar->panelId, 
                                                   CTRL_SQUARE_BUTTON, "", 
                                                   -kWin40ToolbarBevelHeight, 
                                                   -kWin40ToolbarBevelHeight));
        errChk (SetCtrlAttribute (newToolbar->panelId, newToolbar->frameCtrlId, 
                                  ATTR_CTRL_MODE, VAL_INDICATOR)); 
        errChk (GetPanelAttribute (newToolbar->panelId, ATTR_BACKCOLOR, 
                                   &panelColor));
        errChk (SetCtrlAttribute (newToolbar->panelId, newToolbar->frameCtrlId, 
                                  ATTR_OFF_COLOR, panelColor)); 
        errChk (SetPanelAttribute (newToolbar->panelId, 
                                   ATTR_FRAME_STYLE, VAL_HIDDEN_FRAME));
        errChk (SetPanelAttribute (newToolbar->panelId, ATTR_FRAME_THICKNESS, 
                                   0));
    }       
    else
#endif
    {
        errChk (SetPanelAttribute (newToolbar->panelId, ATTR_FRAME_STYLE, VAL_OUTLINED_FRAME));
    }

    errChk (InstallPanelCallback (newToolbar->panelId, ToolbarPanelCallback,
                                  (void *)newToolbar));

    *toolbar = newToolbar;

Error :
    if (error < 0) {
        if (newToolbar)
            Toolbar_Discard (newToolbar);
    }

    return error;
}


int CVIFUNC Toolbar_Discard (ToolbarType toolbar)
{
    int error = -UIENoError;
    int menuBar;
    
    menuBar = GetPanelMenuBar (toolbar->parentPanel);
    if (menuBar > 0) {
        errChk (SetMenuBarAttribute (menuBar, 0, ATTR_DRAW_LIGHT_BEVEL, FALSE));
    }
    errChk (DiscardPanel (toolbar->helpPanel));
    errChk (DiscardPanel (toolbar->panelId));

Error :
    return error;
}


static int DiscardToolbarResources (ToolbarType toolbar)
{
    int error = -UIENoError;

    if (!toolbar)
        return UIEHandleInvalid;

    DisposeToolbarItemList (toolbar->items);
    free (toolbar);

    return error;
}


static ToolbarItemPtr CopyToolbarItem (ToolbarItemPtr toolbarItem)
{
    int error = -UIENoError;
    ToolbarItemPtr newItem = NULL;

    if (toolbarItem == NULL)
        return NULL;

    nullChk (newItem = malloc (sizeof(ToolbarItem)));
    *newItem = *toolbarItem;
    newItem->description = NULL;    /* this is to prevent disposal of */
    if (toolbarItem->description) {
        nullChk (newItem->description = StrDup (toolbarItem->description));
    }

Error :
    if (error < 0) {
        if (newItem) {
            DisposeToolbarItem (newItem);
            newItem = NULL;
        }
    }

    return newItem;
}


static void DisposeToolbarItem (ToolbarItemPtr toolbarItem)
{
    if (!toolbarItem)
        return;

    if (toolbarItem->description)
        free (toolbarItem->description);

    free (toolbarItem);
}


static ListType CopyToolbarItemList (ListType itemList)
{
    int error = -UIENoError;
    ListType newList = NULL;
    int i, numItems;
    ToolbarItemPtr itemPtr, newItemPtr;

    nullChk (newList = ListCreate (sizeof(ToolbarItemPtr)));
    numItems = ListNumItems (itemList);

    /**************************************************************************/
    /* Set all the char pointers to NULL in new list since we don't want them */
    /* to get discarded twice if we get an error.                             */
    /**************************************************************************/
    for (i = 1 ; i <= numItems ; i++) 
    {
        ListGetItem (itemList, &itemPtr, i);
        nullChk (newItemPtr = CopyToolbarItem (itemPtr));
        if (!ListInsertItem (newList, &newItemPtr, END_OF_LIST))
            errChk(UIEOutOfMemory);
    }

Error :
    if (error < 0) {
        if (newList) {
            DisposeToolbarItemList (newList);
            newList = NULL;
        }
    }

    return newList;
}


static void DisposeToolbarItemList (ListType itemList)
{
    ToolbarItemPtr itemPtr;
    int i, numItems;

    numItems = ListNumItems (itemList);
    for (i = 1 ; i <= numItems ; i++) {
        ListGetItem (itemList, &itemPtr, i);
        DisposeToolbarItem (itemPtr);
    }

    ListDispose (itemList);
}


int CVIFUNC Toolbar_Duplicate (int destPanel, ToolbarType sourceToolbar, 
                               char title[], int top, int left, 
                               ToolbarType *destToolbar)
{
    int error = -UIENoError;
    ToolbarType newToolbar = NULL;
    ToolbarItemPtr currentItem;
    int menuVisible, menuHeight;
    int numItems, i;

    if (!sourceToolbar)
        return UIEHandleInvalid;

    if (!destPanel) {
        MessagePopup ("Toolbar API Warning",
                      "Toolbars as top-level panels not supported yet.\n"
                      "This would be a fairly easy enhancement.");
        goto Error;
    }

    nullChk (newToolbar = malloc (sizeof (ToolbarRec)));
    newToolbar->parentPanel = destPanel;
    newToolbar->menuBar = sourceToolbar->menuBar;
    errChk (GetPanelAttribute (newToolbar->parentPanel, ATTR_MENU_BAR_VISIBLE,
                               &menuVisible));
    if (menuVisible) {
        errChk (GetPanelAttribute (newToolbar->parentPanel, ATTR_MENU_HEIGHT,
                                   &menuHeight));
        top +=  menuHeight;
    }

    newToolbar->sizeValid = sourceToolbar->sizeValid;
    newToolbar->itemWidth = sourceToolbar->itemWidth;
    errChk (CreateToolbarHelpPanel (newToolbar));
    errChk (newToolbar->panelId = DuplicatePanel (destPanel, sourceToolbar->panelId,
                                                  title, top, left));
    newToolbar->frameCtrlId = sourceToolbar->frameCtrlId;                                                 
    nullChk (newToolbar->items = CopyToolbarItemList (sourceToolbar->items));

    /*******************************************************/
    /* Make all the toolbar items point to the new toolbar */
    /* Also, set the callback data correctly for each item */
    /*******************************************************/
    numItems = ListNumItems (newToolbar->items);
    for (i = 1 ; i <= numItems ; i++) {
        ListGetItem (newToolbar->items, &currentItem, i);
        if (!currentItem)
            continue;

        currentItem->toolbar = newToolbar;
        if (currentItem->type != kSeparator) 
        {
            errChk (SetCtrlAttribute (newToolbar->panelId, currentItem->controlId,
                                      ATTR_CALLBACK_DATA, (void *)currentItem));
        }
    }

    errChk (InstallPanelCallback (newToolbar->panelId, ToolbarPanelCallback,
                                  (void *)newToolbar));

    *destToolbar = newToolbar;

Error :
    if (error < 0) {
        if (newToolbar)
            Toolbar_Discard (newToolbar);
    }

    return error;
}


static int SizeToolbarToItems (ToolbarType toolbar)
{
    int error = -UIENoError;
    int top = 0, left = 0, maxHeight = 0;
    int brTop, brLeft, brHeight, brWidth;
    int numItems, i, buttonSeparation, panelHeight;
    ToolbarItemPtr currentItem;

    if (!toolbar)
        return UIEHandleInvalid;

    top = GetToolbarButtonTop();
    left = GetToolbarButtonLeft();
    buttonSeparation = GetToolbarButtonSeparation();

    numItems = (toolbar->items) ? (ListNumItems (toolbar->items)) : 0;
    for (i = 1 ; i <= numItems ; i++) 
    {
        ListGetItem (toolbar->items, &currentItem, i);
        if (!currentItem)
            continue;

        switch (currentItem->type) 
        {
            case kCommandButton:
            case kToggleButton:
            case kExclusiveToggleButton:
                errChk (SetCtrlAttribute (toolbar->panelId, currentItem->controlId,
                                          ATTR_VISIBLE, currentItem->active));
                if (currentItem->active) 
                {
                    errChk (SetCtrlAttribute (toolbar->panelId, currentItem->controlId,
                                                ATTR_TOP, top));
                    errChk (SetCtrlAttribute (toolbar->panelId, currentItem->controlId,
                                                ATTR_LEFT, left));
                    errChk (GetCtrlBoundingRect (toolbar->panelId, currentItem->controlId, &brTop,
                                                 &brLeft, &brHeight, &brWidth));
                    toolbar->itemWidth = brWidth;       /* record item width (ASSUME : all buttons are the same size) */

                    if (brHeight > maxHeight)
                        maxHeight = brHeight;

                    left += brWidth;
                    if (NextItemIsButton (toolbar->items, i))
                        left += buttonSeparation;
                }
                break;
            case kSeparator:
                if (currentItem->active)
                    left += kSeparatorWidth;
                break;
        }

        currentItem->currActive = currentItem->active;
    }

    panelHeight = maxHeight + 2*top;
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
        {
        panelHeight += kWin40ToolbarBevelHeight;
        errChk (SetCtrlAttribute (toolbar->panelId, toolbar->frameCtrlId, ATTR_HEIGHT, panelHeight + kWin40ToolbarBevelHeight)); 
        errChk (SetCtrlAttribute (toolbar->panelId, toolbar->frameCtrlId, ATTR_WIDTH, kToolbarWidth)); 
        }
#endif
    errChk (SetPanelAttribute (toolbar->panelId, ATTR_HEIGHT, panelHeight));
    errChk (SetPanelAttribute (toolbar->panelId, ATTR_WIDTH, kToolbarWidth));

    toolbar->sizeValid = 1;

Error :
    return error;
}


/*************************************************************/
/* enabled means the timer is on                             */
/* active means the help is on so the poll rate is different */
/*************************************************************/

static int SetPollTimer (ToolbarType toolbar, int enabled, int active,
                         int resetTimer)
{
    int error = -UIENoError;

    if (enabled != toolbar->helpEnabled) {
        errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->pollTimer,
                                    ATTR_ENABLED, enabled));
        toolbar->helpEnabled = enabled;
    }
    if (enabled && (toolbar->helpActive != active)) {
        errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->pollTimer,
                                  ATTR_INTERVAL,
                                  active ? kToolbarHelpActivePollPeriod
                                         : kToolbarHelpInactivePollPeriod));
        toolbar->helpActive = active;
    }

    if (resetTimer) {
        errChk (ResetTimer (toolbar->helpPanel, toolbar->pollTimer));
    }

Error :
    return error;
}


int CVIFUNC Toolbar_Display (ToolbarType toolbar)
{
    int error = -UIENoError;
    int menuBar;
    
    if (!toolbar)
        return UIEHandleInvalid;

    errChk (SetPanelAttribute (toolbar->panelId, ATTR_VISIBLE, FALSE));
    if (!toolbar->sizeValid) 
    {
        errChk (SizeToolbarToItems (toolbar));
    }
    menuBar = GetPanelMenuBar (toolbar->parentPanel);
    if (menuBar > 0) {
        errChk (SetMenuBarAttribute (menuBar, 0, ATTR_DRAW_LIGHT_BEVEL, TRUE));
    }
    
    errChk (SetPanelAttribute (toolbar->panelId, ATTR_VISIBLE, TRUE));
    errChk (SetPanelAttribute (toolbar->panelId, ATTR_ZPLANE_POSITION, 0));
    errChk (SetPollTimer (toolbar, TRUE, FALSE, TRUE));

Error :
    return error;
}


int ToolbarPanel (ToolbarType toolbar)
{
    return (toolbar) ? toolbar->panelId : UIEHandleInvalid;
}


int CVIFUNC Toolbar_Hide (ToolbarType toolbar)
{
    int error = -UIENoError;
    int menuBar;
    
    if (!toolbar)
        return UIEHandleInvalid;

    HideToolbarHelp (toolbar, FALSE, FALSE);
    errChk (SetPollTimer (toolbar, FALSE, FALSE, FALSE));
    errChk (SetPanelAttribute (toolbar->panelId, ATTR_VISIBLE, 0));
    menuBar = GetPanelMenuBar (toolbar->parentPanel);
    if (menuBar > 0) {
        errChk (SetMenuBarAttribute (menuBar, 0, ATTR_DRAW_LIGHT_BEVEL, FALSE));
    }

Error :
    return error;
}


#if 0       /* For future expansion */

int CVIFUNC_C Toolbar_SetAttribute (ToolbarType toolbar, int attribute, ...)
{
    int error = -UIENoError;

Error :
    return error;
}


int CVIFUNC_C Toolbar_GetAttribute (ToolbarType toolbar, int attribute, ...)
{
    int error = -UIENoError;

Error :
    return error;
}

#endif  /* 0 */


static ToolbarItemPtr gCurrentHelpItem = NULL;


static void SetCurrentHelpItem (ToolbarItemPtr itemPtr, ToolbarType toolbar)
{
    if ((!itemPtr) && (gCurrentHelpItem) &&
        (gCurrentHelpItem->toolbar != toolbar))
        return;

    gCurrentHelpItem = itemPtr;
}


static ToolbarItemPtr CurrentHelpItem (void)
{
    return gCurrentHelpItem;
}


/**********************************************************************/
/* Hidden UI Library function -- returns true if there is currently a */
/* menu open.                                                         */
/**********************************************************************/

extern int CVIFUNC_C MenuIsOpen (void);


static int CVICALLBACK PollTimerCallback (int panel, int control, int event,
                              void *callbackData, int eventData1,
                              int eventData2)
{
    int error = -UIENoError;
    int x, y, panelHeight, parentWidth;
    int i, visible;
    ToolbarItemPtr itemPtr = NULL;
    ToolbarItemPtr currentHelpItem = NULL;
    ToolbarType toolbar;

    toolbar = (ToolbarType)callbackData;

    /* Ignore anything besides a tick (eg EVENT_DISCARD) */
    if (event != EVENT_TIMER_TICK)
        return (FALSE);

    /* If toolbar is not visible, do not respond to this event */
    GetPanelAttribute (toolbar->panelId, ATTR_VISIBLE, &visible);
    if (!visible)
        return (FALSE);

    if (MenuIsOpen ()) {
        HideToolbarHelp (toolbar, FALSE, FALSE);
        SetCurrentHelpItem (NULL, toolbar);

        return (FALSE);
    }

    GetRelativeMouseState (toolbar->panelId, 0, &x, &y, NULL, NULL, NULL);
    GetPanelAttribute (toolbar->panelId, ATTR_HEIGHT, &panelHeight);
    if ((x < 0) || (y < 0) || (y > panelHeight)) {
        /*****************************************************/
        /* Mouse is to the left, above, or below the toolbar */
        /*****************************************************/
        HideToolbarHelp (toolbar, FALSE, FALSE);
        SetCurrentHelpItem (NULL, toolbar);

        return (FALSE);
    }

    /***********************************************************************/
    /* We must use the toolbar panel's parent to determine if the cursor   */
    /* is over the toolbar, since the toolbar itself is ~4K pixels wide    */
    /*                                                                     */
    /* NOTE : If the parent panel is also clipped inside its parent panel, */
    /* this will not be any more useful than checking the width of the     */
    /* toolbar.                                                            */
    /***********************************************************************/
    GetPanelAttribute (toolbar->parentPanel, ATTR_WIDTH, &parentWidth);
    if (x > parentWidth) {
        HideToolbarHelp (toolbar, FALSE, FALSE);
        SetCurrentHelpItem (NULL, toolbar);

        return (FALSE);
    }

    /*******************************************************/
    /* At this point we know the mouse is over the toolbar */
    /*******************************************************/
    i = GetItemIndexFromPosition (toolbar, x);
    if (i == -1)
        HideToolbarHelp (toolbar, TRUE, FALSE);
    else {
        ListGetItem (toolbar->items, &itemPtr, i);
        currentHelpItem = CurrentHelpItem();
        if (itemPtr != currentHelpItem) 
        {
            currentHelpItem = itemPtr;
            if (currentHelpItem->type == kSeparator)
                HideToolbarHelp (toolbar, TRUE, FALSE);
            else 
            {
                DisplayToolbarHelp (itemPtr, y);
                SetPollTimer (toolbar, TRUE, TRUE, FALSE);
            }
        }
        SetCurrentHelpItem (currentHelpItem, toolbar);

    }

    return (FALSE);
}


static int CVICALLBACK HelpTimerCallback (int panel, int control, int event,
                                   void *callbackData, int eventData1,
                                   int eventData2)
{
    ToolbarType toolbar = (ToolbarType)callbackData;

    if (GetCurrentShell() != kShellWin95)
        if (event == EVENT_TIMER_TICK)
            HideToolbarHelp (toolbar, FALSE, TRUE);

    return (0);
}


static int HideToolbarHelp (ToolbarType toolbar, int keepTracking,
                            int rememberCurrentItem)
{
    int error = -UIENoError;

    if (!toolbar)
        goto Error;

    errChk (HidePanel (toolbar->helpPanel));

    errChk (SetPollTimer (toolbar, TRUE, keepTracking, !keepTracking));
    if (!keepTracking) {
        errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpTimer,
                                  ATTR_ENABLED, FALSE));
    }
    if (!rememberCurrentItem)
        SetCurrentHelpItem (NULL, toolbar);

Error :
    return error;
}


static int DisplayToolbarHelp (ToolbarItemPtr toolbarItem, int yPosition)
{
    int error = -UIENoError;
    int panelTop, panelLeft, panelHeight, panelWidth;
    int ctrlTop, ctrlLeft, ctrlHeight, ctrlWidth, ctrlCenter;
    int rightBorder, childLeft, childTop;
    int panel, helpPanel;
    ToolbarType toolbar;

    /* No help for this control */
    if (!toolbarItem->description || !(toolbarItem->description[0]))
        goto Error;

    toolbar = toolbarItem->toolbar;
    panel = toolbar->panelId;
    if (!toolbar->helpPanel) {      /* help panel not created yet */
        errChk (CreateToolbarHelpPanel (toolbar));
    }
    helpPanel = toolbar->helpPanel;

    errChk (GetPanelAttribute (panel, ATTR_TOP, &childTop));
    errChk (GetPanelAttribute (panel, ATTR_LEFT, &childLeft));
    errChk (GetCtrlBoundingRect (panel, toolbarItem->controlId, &ctrlTop, &ctrlLeft,
                                 &ctrlHeight, &ctrlWidth));
    ctrlCenter = ctrlLeft + (ctrlWidth / 2);
    panelTop = childTop + ctrlTop + ctrlHeight + kToolbarHelpVerticalOffset;

    errChk (SetCtrlVal (helpPanel, toolbar->helpText, toolbarItem->description));
    errChk (GetCtrlBoundingRect (helpPanel, toolbar->helpText, &ctrlTop, &ctrlLeft,
                                 &ctrlHeight, &ctrlWidth));
    panelHeight = ctrlHeight;
    panelWidth = ctrlWidth;

    errChk (GetPanelAttribute (toolbar->parentPanel, ATTR_WIDTH, &rightBorder));
    panelLeft = childLeft + ctrlCenter - (panelWidth / 2);
    if (panelLeft <= 0)
        panelLeft = 1;
    if (panelLeft + panelWidth >= rightBorder)
        panelLeft = rightBorder - panelWidth - 1;

    errChk (SetPanelAttribute (helpPanel, ATTR_TOP, panelTop));
    errChk (SetPanelAttribute (helpPanel, ATTR_LEFT, panelLeft));
    errChk (SetPanelAttribute (helpPanel, ATTR_HEIGHT, panelHeight));
    errChk (SetPanelAttribute (helpPanel, ATTR_WIDTH, panelWidth));

    errChk (SetPanelAttribute (helpPanel, ATTR_ZPLANE_POSITION, 0));
    errChk (SetPanelAttribute (helpPanel, ATTR_VISIBLE, TRUE));

    errChk (ResetTimer (toolbar->helpPanel, toolbar->helpTimer));
    errChk (SetCtrlAttribute (toolbar->helpPanel, toolbar->helpTimer,
                              ATTR_ENABLED, TRUE));

Error :
    return error;
}


/***************************************************************/
/* Returns TRUE iff the next item in the item list after item i */
/* is a button.                                                 */
/***************************************************************/

static int NextItemIsButton (ListType itemList, int i)
{
    ToolbarItemPtr nextItem;
    int isButton = FALSE;
    int done = FALSE;
    int numItems;

    if (!itemList)
        return FALSE;

    numItems = ListNumItems (itemList);

    while (!done) 
    {
        if (i >= numItems)
            return FALSE;

        ListGetItem (itemList, &nextItem, i+1);
        if (!nextItem->active) 
        {
            i++;
            continue;
        }

        switch (nextItem->type) 
        {
            case kCommandButton:
            case kToggleButton:
            case kExclusiveToggleButton:
                isButton = TRUE;
                done = TRUE;
                break;
            default:
                done = TRUE;
                break;
        }
    }

    return isButton;
}


/************************************************************************/
/* Return the index associated with the given x position in the toolbar */
/* panel.  This is used for the toolbar help so, given the mouse        */
/* coordinates relative to the panel, we can compute which item the     */
/* cursor is currently over. (return -1 if no item matches)             */
/************************************************************************/

static int GetItemIndexFromPosition (ToolbarType toolbar, int xPosition)
{
    int numItems, i;
    ToolbarItemPtr itemPtr;
    int left = 0;
    int buttonSeparation;

    if (!toolbar || (xPosition < 0))
        return -1;

    left = GetToolbarButtonLeft ();
    buttonSeparation = GetToolbarButtonSeparation();

    numItems = (toolbar->items) ? (ListNumItems (toolbar->items)) : 0;
    for (i = 1 ; i <= numItems ; i++) 
    {
        ListGetItem (toolbar->items, &itemPtr, i);

        if (itemPtr->currActive) 
        {
            switch (itemPtr->type) 
            {
                case kCommandButton:
                case kToggleButton:
                case kExclusiveToggleButton:
                    left += toolbar->itemWidth;
                    if (NextItemIsButton (toolbar->items, i))
                        left += buttonSeparation;
                    break;
                case kSeparator:
                    left += kSeparatorWidth;
                    break;
            }
        }

        if (xPosition < left)
            return i;
    }

    return -1;
}


static int CVICALLBACK ToolbarPanelCallback (int panel, int event, void *callbackData,
                                 int eventData1, int eventData2)
{
    ToolbarItemPtr itemPtr = NULL;
    ToolbarType toolbar = NULL;

    toolbar = (ToolbarType)callbackData;

    switch (event) 
    {
        case EVENT_RIGHT_CLICK:
        case EVENT_RIGHT_DOUBLE_CLICK:
            SetCurrentHelpItem (NULL, toolbar);
            SetPollTimer (toolbar, TRUE, TRUE, FALSE);
            break;
        case EVENT_DISCARD:
            DiscardToolbarResources (toolbar);
            break;
    }

    return (FALSE);
}


static int CVICALLBACK GenericToolbarItemCallbackFunc (int panel, int control, int event,
                                           void *callbackData, int eventData1,
                                           int eventData2)
{
    ToolbarItemPtr toolbarItem = NULL;
    int swallow = FALSE;
    int value;
    
    toolbarItem = (ToolbarItemPtr)callbackData;

    switch (event) 
    {
        case EVENT_LEFT_CLICK:
        case EVENT_LEFT_DOUBLE_CLICK:
            /*****************************************************************/
            /* Exclusive toggle button: If the button is already depressed,  */
            /* swallow the click so nothing happens (the button stays down). */
            /*****************************************************************/
            if (toolbarItem->type == kExclusiveToggleButton)
            {
                GetCtrlVal (panel, control, &value);
                if (value)
                    return (1);
            }
            break;
        case EVENT_COMMIT:
            HideToolbarHelp (toolbarItem->toolbar, FALSE, TRUE);
            
            /**********************************************************/
            /* Exclusive toggles can only get a commit when they have */
            /* been turned on.                                        */
            /**********************************************************/
            if (toolbarItem->type == kExclusiveToggleButton)
                CoerceExclusiveToggleStates (toolbarItem);
                
            if (toolbarItem->callbackFunc) 
            {
                switch (toolbarItem->callbackType) 
                {
                    case kMenuCallback:
                        GetMenuBarAttribute (toolbarItem->toolbar->menuBar,
                                             toolbarItem->menuItem, 
                                             ATTR_CALLBACK_DATA,
                                             &toolbarItem->callbackData);
                        ((MenuCallbackPtr)toolbarItem->callbackFunc)
                                    (toolbarItem->toolbar->menuBar,
                                     toolbarItem->menuItem,
                                     toolbarItem->callbackData,
                                     toolbarItem->toolbar->parentPanel);
                        break;
                    case kControlCallback:
                        swallow = ((CtrlCallbackPtr)toolbarItem->callbackFunc)
                                        (panel, control, event,
                                         toolbarItem->callbackData,
                                         eventData1, eventData2);
                        break;
                }
            }
            break;
    }

    return (0);
}


static int ToolbarItemPtrCmpFn (void *itemA, void *itemB)
{
    ToolbarItemPtr a = *((ToolbarItemPtr *)itemA);
    ToolbarItemPtr b = *((ToolbarItemPtr *)itemB);
    
    return (a - b);
}


/*******************************************************************/
/* If this function is called on a toolbar item which is in the on */
/* state, it will turn off all the other exclusive toggles in the  */
/* set.  If the the item is in the off state, then the function    */
/* will not change the set button states (if a button is already   */
/* on), or it will make the item on (if no buttons are on).        */
/*******************************************************************/

static void CoerceExclusiveToggleStates (ToolbarItemPtr toolbarItem)
{
    int position, firstExclusive, lastExclusive;
    int i, numItems;
    ToolbarType toolbar = toolbarItem->toolbar;
    ToolbarItemPtr itemPtr;
    int turnCurrentOn, turnOthersOff;
    int currentVal;
    int menuBar;
    
    /* Need to find the position of the item on the toolbar */
    position = ListFindItem (toolbar->items, &toolbarItem, FRONT_OF_LIST,
                             ToolbarItemPtrCmpFn);

    numItems = ListNumItems (toolbar->items);

    /*******************************************************************/
    /* Find position of first exclusive toggle in this set (separators */
    /* terminate a set, as do command buttons).                        */
    /*******************************************************************/
    firstExclusive = position;
    for (i = position - 1 ; i >= 1 ; i--)
    {
        ListGetItem (toolbar->items, &itemPtr, i);
        if (itemPtr->type == kExclusiveToggleButton)
            firstExclusive = i;
        else
            break;
    }
    
    /*******************************************************************/
    /* Find position of last exclusive toggle in this set (separators  */
    /* terminate a set, as do command buttons).                        */
    /*******************************************************************/
    lastExclusive = position;
    for (i = position + 1 ; i <= numItems ; i++)
    {
        ListGetItem (toolbar->items, &itemPtr, i);
        if (itemPtr->type == kExclusiveToggleButton)
            lastExclusive = i;
        else
            break;
    }

    /* If item is on, turn all others off */
    GetCtrlVal (toolbarItem->toolbar->panelId, toolbarItem->controlId, 
                &currentVal);
    turnOthersOff = currentVal ? TRUE : FALSE;
    turnCurrentOn = TRUE;   /* TRUE by default, becomes FALSE if we find an already active item */
    
    for (i = firstExclusive ; i <= lastExclusive ; i++)
    {
        if (i != position)
        {
            ListGetItem (toolbar->items, &itemPtr, i);
            if (turnOthersOff)
            {
                SetCtrlVal (toolbar->panelId, itemPtr->controlId, 0);
                if (itemPtr->callbackType == kMenuCallback)
                {
                    if (menuBar = GetPanelMenuBar (toolbar->parentPanel))
                        SetMenuBarAttribute (menuBar, itemPtr->menuItem, 
                                             ATTR_CHECKED, FALSE);
                }
            }
            else
            {
                /******************************************************/
                /* If we find an item which is already on, don't turn */
                /* on the specified item.                             */
                /******************************************************/
                GetCtrlVal (toolbar->panelId, itemPtr->controlId,
                            &currentVal);
                if (currentVal)         /* found an already active item */
                {
                    turnCurrentOn = FALSE;
                    break;
                }
            }
        }
    }
    
    if (turnCurrentOn)
    {
        SetCtrlVal (toolbar->panelId, toolbarItem->controlId, 1);
        if (toolbarItem->callbackType == kMenuCallback)
        {
            if (menuBar = GetPanelMenuBar (toolbar->parentPanel))
                SetMenuBarAttribute (menuBar, toolbarItem->menuItem, ATTR_CHECKED,
                                     TRUE);
        }
    }
}


/**********************************************************************/
/* Assumes that the toolbar and the type have already been validated. */
/**********************************************************************/

static int CreateNewToolbarItemCtrl (ToolbarItemPtr toolbarItem, char imageFile[])
{
    int error = -UIENoError;
    int ctrlType;
    int toolBarButtonSize;
    int panelColor;
    
    if (!toolbarItem)
        return UIEHandleInvalid;

    if (toolbarItem->type == kSeparator)    /* Don't need a control */
        goto Error;

    toolBarButtonSize = GetToolbarButtonSize();

    switch (toolbarItem->type) 
    {
        case kCommandButton:
            ctrlType = CTRL_PICTURE_COMMAND_BUTTON;
            break;
        case kToggleButton:
        case kExclusiveToggleButton:
            ctrlType = CTRL_PICTURE_TOGGLE_BUTTON;
            break;
    }

    errChk (toolbarItem->controlId = NewCtrl (toolbarItem->toolbar->panelId,
                                        ctrlType, "", 0, 0));
    errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                              toolbarItem->controlId, ATTR_HEIGHT,
                              toolBarButtonSize));
    errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                              toolbarItem->controlId, ATTR_WIDTH,
                              toolBarButtonSize));

    errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                              toolbarItem->controlId, ATTR_FAST_DRAW_BUTTON,
                              TRUE));

    errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                              toolbarItem->controlId, ATTR_IMAGE_FILE,
                              imageFile));

    errChk (InstallCtrlCallback (toolbarItem->toolbar->panelId,
                                 toolbarItem->controlId,
                                 GenericToolbarItemCallbackFunc,
                                 (void *)toolbarItem));

    /* This will make the toolbar buttons have the same color as the toolbar */
    errChk (GetPanelAttribute (toolbarItem->toolbar->panelId, 
                               ATTR_BACKCOLOR, &panelColor));
    if (toolbarItem->type == kCommandButton)
    {
        errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                                  toolbarItem->controlId,
                                  ATTR_CMD_BUTTON_COLOR, panelColor));
    }
    else
    {
        errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                                  toolbarItem->controlId,
                                  ATTR_ON_COLOR, panelColor));
        errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                                  toolbarItem->controlId,
                                  ATTR_OFF_COLOR, panelColor));
    }
                               
    if (toolbarItem->toolbar->conformBitmapColors)
    {
        errChk (ConformCtrlBitmapToPanel (toolbarItem->toolbar->panelId,
                                          toolbarItem->controlId, 0, 
                                          VAL_PANEL_GRAY, -1, VAL_BLACK, -1));
    }
    
Error :
    return error;
}


int CVIFUNC Toolbar_InsertItem (ToolbarType toolbar, int position, int type, 
                                int active, char description[], 
                                int callbackType, int menuItem,
                                CtrlCallbackPtr callbackFunc, 
                                void *callbackData,
                                char imageFile[])
{
    int error = -UIENoError;
    ToolbarItemPtr newItem = NULL;
    MenuCallbackPtr menuCallback = NULL;
    
    if (!toolbar)
        return UIEHandleInvalid;

    nullChk (newItem = malloc (sizeof (ToolbarItem)));
    newItem->type = type;
    newItem->active = active;
    newItem->currActive = FALSE;
    newItem->toolbar = toolbar;
    if (description) 
    {
        nullChk (newItem->description = StrDup (description));
    }
    else
        newItem->description = NULL;

    switch (callbackType) {
        case kMenuCallback:
            newItem->callbackType = callbackType;
            newItem->menuItem = menuItem;
            errChk (GetMenuBarAttribute (toolbar->menuBar, menuItem,
                                         ATTR_CALLBACK_FUNCTION_POINTER, &menuCallback));
            newItem->callbackFunc = (void *)menuCallback;
            errChk (GetMenuBarAttribute (toolbar->menuBar, menuItem,
                                         ATTR_CALLBACK_DATA, &newItem->callbackData));
            break;
        case kControlCallback:
            newItem->callbackType = callbackType;
            newItem->callbackFunc = (void *)callbackFunc;
            newItem->callbackData = callbackData;
            newItem->menuItem = menuItem;
            break;
        default:
            newItem->callbackType = kNoCallback;
            newItem->callbackFunc = NULL;
            newItem->callbackData = NULL;
            newItem->menuItem = 0;
            break;
    }

    errChk (CreateNewToolbarItemCtrl (newItem, imageFile));

    if (!ListInsertItem (toolbar->items, &newItem, position))
        errChk(UIEIndexOutOfRange);

    if (newItem->type == kExclusiveToggleButton)
        CoerceExclusiveToggleStates (newItem);
    
    if (active)
    {
        toolbar->sizeValid = 0;
        if (ToolbarIsVisible (toolbar))
        {
            errChk (Toolbar_Display (toolbar));
        }
    }

Error :
    if (error < 0) {
        if (newItem) {
            if (newItem->controlId)
                DiscardCtrl (toolbar->panelId, newItem->controlId);
            DisposeToolbarItem (newItem);
        }
    }

    return error;
}


int CVIFUNC Toolbar_RemoveItem (ToolbarType toolbar, int position)
{
    int error = -UIENoError;
    ToolbarItemPtr itemPtr;

    if (!toolbar)
        return UIEHandleInvalid;

    if ((position < 0) || (position > ListNumItems (toolbar->items)))
        return UIEIndexOutOfRange;

    ListRemoveItem (toolbar->items, &itemPtr, position);
    if (itemPtr->type != kSeparator)
    {
        errChk (DiscardCtrl (toolbar->panelId, itemPtr->controlId));
    }
    DisposeToolbarItem(itemPtr);
    toolbar->sizeValid = 0;
    if (ToolbarIsVisible (toolbar))
    {
        errChk (Toolbar_Display (toolbar));
    }

Error :
    return error;
}


int CVIFUNC Toolbar_SetItemVal (ToolbarType toolbar, int position, int value)
{
    int error = -UIENoError;
    ToolbarItemPtr toolbarItem;

    if (!toolbar)
        return UIEHandleInvalid;

    if ((position < 0) || (position > ListNumItems (toolbar->items)))
        return UIEIndexOutOfRange;

    ListGetItem (toolbar->items, &toolbarItem, position);

    if (toolbarItem->type != kSeparator) 
    {
        if (toolbarItem->type == kExclusiveToggleButton)
        {
            /* Cannot set an exclusive toggle to "off" */
            if (value)
            {
                errChk (SetCtrlVal (toolbarItem->toolbar->panelId,
                                    toolbarItem->controlId, value));
                CoerceExclusiveToggleStates (toolbarItem);
            }
        }
        else
        {
            errChk (SetCtrlVal (toolbarItem->toolbar->panelId,
                                toolbarItem->controlId, value));
        }
    }

Error :
    return error;
}


int CVIFUNC Toolbar_DimItem (ToolbarType toolbar, int position, int dimmed)
{
    int error = -UIENoError;
    ToolbarItemPtr toolbarItem;

    if (!toolbar)
        return UIEHandleInvalid;

    if ((position < 0) || (position > ListNumItems (toolbar->items)))
        return UIEIndexOutOfRange;

    ListGetItem (toolbar->items, &toolbarItem, position);

    if (toolbarItem->type != kSeparator) {
        errChk (SetCtrlAttribute (toolbarItem->toolbar->panelId,
                                  toolbarItem->controlId,
                                  ATTR_DIMMED, dimmed));
    }

Error :
    return error;
}


int CVIFUNC Toolbar_ActivateItem (ToolbarType toolbar, int position, 
                                  int active, int refresh)
{
    int error = -UIENoError;
    ToolbarItemPtr toolbarItem;

    if (!toolbar)
        return UIEHandleInvalid;

    if ((position < 0) || (position > ListNumItems (toolbar->items)))
        return UIEIndexOutOfRange;

    ListGetItem (toolbar->items, &toolbarItem, position);

    active = (active) ? 1 : 0;
    if (toolbarItem->active != active) {
        toolbarItem->active = active;
        toolbarItem->toolbar->sizeValid = 0;
    }

    if (refresh && ToolbarIsVisible (toolbarItem->toolbar)) {
        errChk (Toolbar_Display (toolbarItem->toolbar));
    }

Error :
    return error;
}


int CVIFUNC Toolbar_GetNumItems (ToolbarType toolbar, int *count)
{
    int error = -UIENoError;

    if (!toolbar)
        return UIEHandleInvalid;

    if (!count)
        goto Error;

    *count = (toolbar->items) ? ListNumItems(toolbar->items) : 0;

Error :
    return error;
}


int CVIFUNC Toolbar_GetIndexFromDescription (ToolbarType toolbar, 
                                             char label[], int *index)
{
    int error = -UIENoError;
    int i, numItems;
    ToolbarItemPtr toolbarItem;

    if (!index)
        goto Error;

    *index = -1;

    numItems = ListNumItems (toolbar->items);
    for (i = 1 ; i <= numItems ; i++) {
        ListGetItem (toolbar->items, &toolbarItem, i);
        if (!strcmp (label, toolbarItem->description)) {
            *index = i;
            break;
        }
    }

Error :
    return error;
}


int CVIFUNC Toolbar_GetIndexFromMenu (ToolbarType toolbar, int menuItem, 
                                      int *index)
{
    int error = -UIENoError;
    int i, numItems;
    ToolbarItemPtr toolbarItem;

    if (!toolbar)
        return UIEHandleInvalid;

    if (!index)
        goto Error;

    *index = -1;

    numItems = ListNumItems (toolbar->items);
    for (i = 1 ; i <= numItems ; i++) {
        ListGetItem (toolbar->items, &toolbarItem, i);
        if (toolbarItem->menuItem == menuItem) {
            *index = i;
            break;
        }
    }

Error :
    return error;
}


static int GetToolbarButtonSeparation (void)
{
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
        return kWin40ToolButtonSeparation;
    else
#endif
        return kButtonSeparation;
}


#define LOBYTE(w)   ((unsigned char)(w))
#define LOWORD(l)   ((unsigned short)(unsigned long)(l))

static int GetCurrentShell (void)
{
    static int currentShell = kShellUninitialized;
    
    if (currentShell == kShellUninitialized)
    {
        /* All platforms other than win32 will get this shell */
        currentShell = kShellWin31;         
        
#if _NI_mswin32_
        /* win32 platforms get the new shell if the Windows version >= 4.0 */
        if (LOBYTE(LOWORD(GetVersion())) >= 4)
            currentShell = kShellWin95;
#endif
    }

    return currentShell;
}


static int GetToolbarButtonSize (void)
{
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
        return kWin40ToolButtonSize;
    else
#endif
        return kToolbarButtonSize;
}


static int GetToolbarButtonTop (void)
{
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
        return kWin40ToolButtonTop;
    else
#endif
        return kToolButtonTop;
}


static int GetToolbarButtonLeft (void)
{
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
        return kWin40ToolButtonLeft;
    else
#endif
        return kToolButtonLeft;
}


static int GetSystemTextColor (void)
{
    static int colorInit = FALSE, gColor;
    
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
    {
        if (!colorInit)
        {
            gColor = WinColorToCVIColor (GetSysColor (COLOR_BTNTEXT));
            colorInit = TRUE;
        }
            
        return gColor;
    }
    else
#endif
        return kSystemButtonTextColor;
}


static int GetToolbarHelpBGColor (void)
{
    static int colorInit = FALSE, gColor;
    
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
    {
        if (!colorInit)
        {
            gColor = WinColorToCVIColor (GetSysColor (COLOR_INFOBK));
            colorInit = TRUE;
        }
            
        return gColor;
    }
    else
#endif
        return kToolbarHelpPanelColor;
}


static int GetToolbarHelpTextColor (void)
{
    static int colorInit = FALSE, gColor;
    
#if _NI_mswin32_
    if (GetCurrentShell() == kShellWin95)
    {
        if (!colorInit)
        {
            gColor = WinColorToCVIColor (GetSysColor (COLOR_INFOTEXT));
            colorInit = TRUE;
        }
            
        return gColor;
    }
    else
#endif
        return kToolbarHelpTextColor;
}


#if _NI_mswin32_

static int WinColorToCVIColor(int c)
{
    unsigned char *colorPtr = (unsigned char *)&c;
    unsigned char blue;

    blue = colorPtr[0];
    colorPtr[0] = colorPtr[2];
    colorPtr[2] = blue;
    colorPtr[3] = 0x00;

    return (int)c;
}

#endif  /* _NI_mswin_ */

